/*
 * Copyright (c) 1999-2000 GZ 
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "services.h"

static struct akill *firstAkill = NULL;

void autoremoveakill(char *mask);

void listAkills(char *from, char type, char *masker)
{
	struct akill   *ak;
	char mask[NICKLEN + USERLEN + HOSTLEN + 2];
  	char length[16];
  	struct tm *t;
	char date[20];
	int found = 0;
	
	if (firstAkill)
	{
                for (ak = firstAkill; ak; ak = ak->next) {
			if (ak->type == type) {
				if (!found)
					sSend(":%s NOTICE %s :%-14s %-8s %-20s %-10s %s", OperServ, from, "SetTime", "Length", "Mask", "Set By", "Reason");
				found++;
				strcpy(mask, ak->nick);
				strcat(mask, "!");
				strcat(mask, ak->user);
				strcat(mask, "@");
				strcat(mask, ak->host);
				t=localtime(&(ak->set));
				strftime(date, 20, "%m/%d/%y-%I:%M%p", t);
				
				if(ak->unset)
					sprintf(length, "%luh", (long)(ak->unset - ak->set)/3600);
				else
					strcpy(length, "forever");

				if(masker) {
					if(!match(masker, mask))
						sSend(":%s NOTICE %s :%s %-8s %-20s %-10s %s", OperServ, from, date, length, mask, ak->setby, ak->reason);
				}
				else
					sSend(":%s NOTICE %s :%s %-8s %-20s %-10s %s", OperServ, from, date, length, mask, ak->setby, ak->reason);
			}
		}
	}

	if (found)
		sSend(":%s NOTICE %s :End of list. (%d total)", OperServ, from, found);
	else
		sSend(":%s NOTICE %s :There are no %s.", OperServ, from, (type == A_AKILL) ? "AutoKills" : "Ignores");
}

struct akill *isAKilled(char *nick, char *user, char *host)
{
	struct akill *ak;

	for (ak = firstAkill; ak; ak = ak->next) {
		if (!match(nick, ak->nick) && !match(user, ak->user) && !match(host, ak->host) && (ak->type == A_AKILL))
			return ak;
	}
	return NULL;
}

struct akill *isIgnored(char *nick, char *user, char *host)
{
	struct akill *ak;

	for (ak = firstAkill; ak; ak = ak->next) {
		if (!match(nick, ak->nick) && !match(user, ak->user) && !match(host, ak->host) && (ak->type == A_IGNORE))
			return ak;
	}
	return NULL;
}

void addakill(long length, char *mask, char *by, char type, char *reason)
{
	UserList	*from;
	struct akill   *ak;
	char           *akmask;
	char           *temp1;
	char           *user1 = NULL, *host1 = NULL;
	char		   sendfrom[256];
	char 		   subject[256];
	char temp[12];

	if (strlen(mask) < 5)
		return;

	ak = smalloc(sizeof(struct akill));

	akmask = strdup(mask);

	for (temp1 = akmask; *temp1; temp1++)
	{
		if (*temp1 == '!') {
			*temp1 = 0;
			user1 = temp1 + 1;
		}
		if (*temp1 == '@') {
			*temp1 = 0;
			host1 = temp1 + 1;
		}
	}

	if ((!user1) || (!host1))
	{
		sfree(akmask);
		return;
	}

	strcpy(ak->nick, akmask);
	strcpy(ak->user, user1);
	strcpy(ak->host, host1);

	if(strlen(ak->user) == 1 && strlen(ak->host) == 1) {
		from = getNickData(by);
		if(from && from->reg && !(from->reg->flags & NROOT)) {
			sSend(":%s NOTICE %s :Access denied - only SRA can do nick-akills.", NickServ, by);
			sfree(akmask); 
			return;
		}
	}

	if (firstAkill)
		firstAkill->prev = ak;

	ak->next = firstAkill;
	ak->prev = NULL;
	firstAkill = ak;

	strcpy(ak->setby, by);

	ak->set = time(NULL);

	if (length == 0)
		ak->unset = 0;
	else
		ak->unset = ak->set + length;

	ak->type = type;
	strcpy(ak->reason, reason);

	sprintf(sendfrom, "Operator Services - Kline <%s@%s>", OperServ, NETWORK);

	if (type == A_AKILL) {
		sSend(":%s GLOBOPS :%s set akill for %s for %s [%s]", OperServ, by, mask, expand_time(length), reason);
		sprintf(subject, "AKILL for %s", mask); 
		strcpy(temp,"autokilled");
	}
	else {
		sSend(":%s GLOBOPS :%s set ignore for %s for %s [%s]", OperServ, by, mask, expand_time(length), reason);
		sprintf(subject, "IGNORE for %s", mask);
		strcpy(temp,"ignored");
	}

	if (type == A_AKILL && !strcmp(ak->nick, "*"))
		sSend(":%s AKILL %s %s :K-lined for %s [%s]", myname, ak->host, ak->user, expand_time(length), ak->reason);

	if (ak->unset)
		ak->tid = timer(length, autoremoveakill, "autoremoveakill", mask);
	else
		ak->tid = 0;
  
	sendMail(AKILLMAILTO, sendfrom, subject, "%s was %s by %s for %s\nReason: %s\nSet at %s", mask, temp, ak->setby, expand_time(length), reason, ctime(&(ak->set)));

	sfree(akmask);
}

void removeakill(char *from, char *mask)
{
	char           *temp1;
	struct akill   *ak;
	char           *nick, *user = NULL, *host = NULL, *oldmask;
	char            sendfrom[256], subject[256];
	time_t          temp;
 
	if (strlen(mask) < 5)
		return;

	oldmask = strdup(mask);

	nick = oldmask;

	for (temp1 = oldmask; *temp1; temp1++)
	{
		if (*temp1 == '!') {
			*temp1 = 0;
                        user = temp1 + 1;
		}
		if (*temp1 == '@') {
			*temp1 = 0;
                        host = temp1 + 1;
		}
	}

	if ((!user) || (!host))
		return;

	sprintf(sendfrom, "Operator Services - Kline <%s@%s>", OperServ, NETWORK);

	for (ak = firstAkill; ak; ak = ak->next) {

		if (!strcmp(host, ak->host) && !strcmp(user, ak->user) && !strcmp(nick, ak->nick)) {

			if (ak->type == A_AKILL) {
                        sprintf(subject, "Removal of AKILL for %s", mask);
				sSend(":%s GLOBOPS :%s is removing akill %s.", OperServ, from, mask);
				logDump(operlog,"%s removed akill %s", from, mask);
			}
			else {
                        sprintf(subject, "Removal of IGNORE for %s", mask);
				sSend(":%s GLOBOPS :%s is removing ignore %s.", OperServ, from, mask);
				logDump(operlog,"%s removed ignore %s", from, mask);
                  }

			temp = time(NULL);
   
			sendMail(AKILLMAILTO, sendfrom, subject, "%s removed autokill/ignore %s at %s", from, mask, ctime(&temp));

			if (ak->prev)
				ak->prev->next = ak->next;
			if (ak->next)
				ak->next->prev = ak->prev;
			if (ak == firstAkill)
				firstAkill = ak->next;

			if (ak->type == A_AKILL && !strcmp(nick, "*"))
				sSend(":%s RAKILL %s %s", myname, ak->host, ak->user);
/*
 * Fun fact:  If OperServ is removing the autokill, the timer is up, and
 * it shouldn't be canceled, or two timers will be cancelled...
 */
			if (strcmp(from, OperServ))
			{
				sSend(":%s NOTICE %s :%s %s removed", OperServ, from, (ak->type == A_AKILL) ? "AutoKill" : "Ignore", mask);
				cancel_timer(ak->tid);
			}

			sfree(ak);
			sfree(oldmask);
			return;
		}
	}

	if (strcmp(from, OperServ))
		sSend(":%s NOTICE %s :Autokill/Ignore %s not found", OperServ, from, mask);

	sfree(oldmask);
}

int findakill(char *mask)
{
	char *temp1;
	struct akill *ak;
  	char *nick, *user = NULL, *host = NULL, *oldmask;
	
	oldmask = strdup(mask);

	nick = oldmask;

	for (temp1 = oldmask; *temp1; temp1++) {
		if (*temp1 == '!') {
			*temp1 = 0;
			user = temp1 + 1;
		}
		if (*temp1 == '@') {
			*temp1 = 0;
			host = temp1 + 1;
		}
	}

	for (ak = firstAkill; ak; ak = ak->next) {
		if (!strcmp(host, ak->host) && !strcmp(user, ak->user) && !strcmp(nick, ak->nick)) {
			sfree(oldmask);
			return 1;
		}
	}
	sfree(oldmask);
	return 0;
}

void autoremoveakill(char *mask)
{
	removeakill(OperServ, mask);
}

void saveakills()
{
	FILE           *aksave;
	struct akill   *ak;

	aksave = fopen(DATABASEDIR "/akill.db", "w");
	if (aksave == NULL) {
		Debug((DEBUG_ERROR,"Failed to open akill db"));
		return;
	}

	for (ak = firstAkill; ak; ak = ak->next) {
		fputs("-\n", aksave);
		fputs(ak->nick, aksave);
		fputs("\n", aksave);
		fputs(ak->user, aksave);
		fputs("\n", aksave);
		fputs(ak->host, aksave);
		fputs("\n", aksave);
		fprintf(aksave, "%lu %lu\n", (long) ak->set, (long) ak->unset);
		fputs(ak->setby, aksave);
		fputs("\n", aksave);
		fputs(ak->reason, aksave);
		fputs("\n", aksave);
		fprintf(aksave, "%d\n", ak->type);
	}
	fclose(aksave);
}

void loadakills()
{
	FILE           *akload;
	struct akill   *ak;
        char            buffer[256], mask[NICKLEN+USERLEN+HOSTLEN+2];
	int             i = 1, t, expired = 0, count = 0;
	long            length;
	time_t	    start;

	firstAkill = NULL;

	start = time(NULL);

	akload = fopen(DATABASEDIR "/akill.db", "r");

	if (!akload) {
		return;
	}

	while (sfgets(buffer, 256, akload)) {
/*
 * Check for sync 
 */
		if (strcmp(buffer, "-") != 0) {
			sSend(":%s GLOBOPS :Sync error in akill.db near line %d", OperServ, i);
			fprintf(corelog, "Sync error in akill.db near line %d\n", i);
			fclose(akload);
			return;
		}

		ak = smalloc(sizeof(struct akill));

		fscanf(akload, "%s\n", ak->nick);
		fscanf(akload, "%s\n", ak->user);
		fscanf(akload, "%s\n", ak->host);
/*		
		sfgets(ak->nick, NICKLEN, akload);
		sfgets(ak->user, USERLEN, akload);
		sfgets(ak->host, HOSTLEN, akload);
*/
		fscanf(akload, "%lu %lu\n", (long *) &(ak->set), (long *) &(ak->unset));
		sfgets(ak->setby, NICKLEN, akload);
		sfgets(ak->reason, 80, akload);
		fscanf(akload, "%d\n", &t);   
		ak->type = t;
		ak->tid = 0;

		strcpy(mask, ak->nick);
		strcat(mask, "!");
		strcat(mask, ak->user);
		strcat(mask, "@");
		strcat(mask, ak->host);

		count++;
		length = 0;

		if (ak->unset)
			length = ak->unset - ak->set;

		if ((start < ak->unset) || (ak->unset == 0)) {
			if(ak->unset) ak->tid = timer((ak->unset - start), autoremoveakill, "autoremovesakill", mask);

			if ((ak->type == A_AKILL) && (!strcmp(ak->nick,"*")))
				sSend(":%s AKILL %s %s :K-lined for %s [%s]", myname, ak->host, ak->user, expand_time(length), ak->reason);

			ak->next = firstAkill;
			ak->prev = NULL;
			if (ak->next)
				ak->next->prev = ak;
			firstAkill = ak;
		} 
		else {   /* IT'S EXPIRED, REALLY! */
			if(ak->type == A_AKILL)
				sSend(":%s RAKILL %s %s", myname, ak->host, ak->user);
      		expired++;
			sfree(ak);
		}
	}
	fclose(akload);
}
