/*
 * Copyright (c) 1999-2000 GZ
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
                            

#include "services.h"

/* Some small macro to check if a channel name is valid. -GZ */
#define IsChannelName(name) ((name) && ((*(name) == '#')))

void indexAkickItems(RegChanList *);
void indexOpItems(RegChanList *);

/* all these are ONLY for channel linked list funcs */
int isFounder(RegChanList *, UserList *);
void makeModeLockStr(RegChanList *, char *);
void banKick(ChanList *, UserList *, char *, ...);
int countRegChans(char *);
void remfromallchanops(char *);
char *showchanoplevel(int);

void createGhostChannel(char *);
void deleteGhostChannel(char *);

void doRegChanMode(RegChanList *);

/* This one will massdeop a channel, no questions asked */
void CSmassdeop(char *);

int check_wildcard(char *);
int check_mask(char *, int);

/* THIS FUNCTION MAKES SURE A HOSTMASK CONTAINS MORE THAN JUST WILDCARDS */
/* RETURNS 1 IF IT DOES, 0 IF IT DOES NOT */
/* INPUT STRING MUST BE NULL TERMINATED */

int check_wildcard(char *mask)
{
   char *temp1;
   int foundflag = 0;

   for (temp1 = mask; *temp1; temp1++)
   {
      if ((*temp1 != '.') && (*temp1 != '*') && (*temp1 != '?'))
      {
         foundflag = 1;
         break;
      }
   }
   return (foundflag);
}

/* THIS FUNCTION CHECKS A HOSTMASK FOR VALIDITY */
/* *!*@* IS NOT ALLOWED */
/* RETURNS 1 IF GOOD, 0 IF FAULT */

int check_mask(char *maskin, int allow_wild)
{
	char *nick;
	char *user = NULL;
	char *host = NULL;
	char *tmp;
	char mask[NICKLEN+USERLEN+HOSTLEN+2];
      int valid;

	if ((!maskin) || (!(*maskin)) || (strlen(maskin) > NICKLEN+USERLEN+HOSTLEN))
		return 0;

	strcpy(mask,maskin);

	if (!strchr(mask,'!') || !strchr(mask,'@'))
		return 0;

	nick = mask;

	for (tmp = mask; *tmp; tmp++)
	{
		if (*tmp == '!')
		{
			if ((!user) && (!host))
			{ 
				*tmp = 0;
				user = tmp+1;
			}
		}
		if (*tmp == '@')
		{
                        if ((!host) && (user))
			{
				*tmp = 0;
				host = tmp+1;
	    		}
      	}
	}

	if (!strlen(nick) || (!user) || (!host))
		return 0;

	if (strchr(user,'!') || strchr(user,'@') || strchr(host,'!') || strchr(host,'@'))
		return 0;

	if (allow_wild)
		return 1;
	else
	{	
		valid = check_wildcard(nick) + check_wildcard(user) + check_wildcard(host);

		if (!valid)
			return 0;
		else
			return 1;
	}
} 

/* The following bunch of functions all deal with channels and registered
 * ones, ignore them unless you need to fiddle with the linked lists for
 * channels. */
void addChan(ChanList *newchan)
{
	long hashEnt = getHashKey(newchan->name) % CHANHASHSIZE;
	
	if(getChanData(newchan->name)) {
		sfree(newchan);
		return;
	}
	
	if(!firstChan) {
		firstChan = newchan;
		newchan->previous = NULL;
	}
	else {
		lastChan->next = newchan;
		newchan->previous = lastChan;
	}
	lastChan = newchan;
	newchan->next = NULL;
	
	if(!ChanHash[hashEnt].chan) {
		ChanHash[hashEnt].chan = newchan;
		newchan->hashprev = NULL;
	}
	else {
		ChanHash[hashEnt].lastchan->hashnext = newchan;
		newchan->hashprev = ChanHash[hashEnt].lastchan;
	}
	
	ChanHash[hashEnt].lastchan = newchan;
	newchan->hashnext = NULL;
}

void addRegChan(RegChanList *newchan)
{
	long hashEnt = getHashKey(newchan->name) % CHANHASHSIZE;
	
	if(getRegChanData(newchan->name)) {
		sfree(newchan);
		return;
	}
	
	if(!firstRegChan) {
		firstRegChan = newchan;
		newchan->previous = NULL;
	}
	else {
		lastRegChan->next = newchan;
		newchan->previous = lastRegChan;
	}
	lastRegChan = newchan;
	newchan->next = NULL;
	
	if(!RegChanHash[hashEnt].chan) {
		RegChanHash[hashEnt].chan = newchan;
		newchan->hashprev = NULL;
	}
	else {
		RegChanHash[hashEnt].lastchan->hashnext = newchan;
		newchan->hashprev = RegChanHash[hashEnt].lastchan;
	}
	
	RegChanHash[hashEnt].lastchan = newchan;
	newchan->hashnext = NULL;
}

void addChanUser(ChanList *channel, cNickList *newnick)
{
   if(!channel->firstUser) {
		channel->firstUser = newnick;
		newnick->previous = NULL;
	}
	else {
		channel->lastUser->next = newnick;
		newnick->previous = channel->lastUser;
	}
	
	channel->lastUser = newnick;
	newnick->next = NULL;
  	
}


/* add someone to the channel ban list */
void addChanBan(ChanList *channel, cBanList *item)
{
	if(!channel->firstBan) {
		channel->firstBan = item;
		item->previous = NULL;
	}
	else {
		channel->lastBan->next = item;
		item->previous = channel->lastBan;
	}
	
	channel->lastBan = item;
	item->next = NULL;
}


/* add someone to the channel akick list */
void addChanAkick(RegChanList *channel, cAkickList *item)
{
	if(!channel->firstAkick) {
		channel->firstAkick = item;
		item->previous = NULL;
	}
	else {
		channel->lastAkick->next = item;
		item->previous = channel->lastAkick;
	}
	
	channel->lastAkick = item;
	item->next = NULL;
	channel->akicks++;
	indexAkickItems(channel);
}

/* add someone to the channel 'access list' */
void addChanOp(RegChanList *channel, cAccessList *item)
{
   if(!channel->firstOp) {
		channel->firstOp = item;
		item->previous = NULL;
	}
	else {
		channel->lastOp->next = item;
		item->previous = channel->lastOp;
	}
	
	channel->lastOp = item;
	item->next = NULL;
	
	channel->ops++;
	indexOpItems(channel);
}

/* remove a channel from the channel list, and close it up where necessary */
void delChan(ChanList *killme)
{
	long hashEnt;
	ChanList *tmpchan;
	
	if(killme->previous)
	  killme->previous->next = killme->next;
	else
	  firstChan = killme->next;
	if(killme->next)
	  killme->next->previous = killme->previous;
	else
	  lastChan = killme->previous;
	
	hashEnt = getHashKey(killme->name) % CHANHASHSIZE;
	for(tmpchan = ChanHash[hashEnt].chan;tmpchan;tmpchan=tmpchan->hashnext) {
		if(tmpchan == killme) {
			if(killme->hashprev)
			  killme->hashprev->hashnext = killme->hashnext;
			else
			  ChanHash[hashEnt].chan = killme->hashnext;
			if(killme->hashnext)
			  killme->hashnext->hashprev = killme->hashprev;
			else
			  ChanHash[hashEnt].lastchan = killme->hashprev;
		}
	}
	
	while(killme->firstUser)
	  delChanUser(killme, killme->lastUser);
	while(killme->firstBan)
	  delChanBan(killme, killme->lastBan);
	
	sfree(killme);
}

void delRegChan(RegChanList *killme)
{
	long hashEnt;
	RegChanList *tmpchan;
	
	if(killme->previous)
	  killme->previous->next = killme->next;
	else
	  firstRegChan = killme->next;
	if(killme->next)
	  killme->next->previous = killme->previous;
	else
	  lastRegChan = killme->previous;
	
	hashEnt = getHashKey(killme->name) % CHANHASHSIZE;
	for(tmpchan = RegChanHash[hashEnt].chan;tmpchan;tmpchan=tmpchan->hashnext) {
		if(tmpchan == killme) {
			if(killme->hashprev)
			  killme->hashprev->hashnext = killme->hashnext;
			else
			  RegChanHash[hashEnt].chan = killme->hashnext;
			if(killme->hashnext)
			  killme->hashnext->hashprev = killme->hashprev; /* AAAAAHHHHAAAA!!! */
			else
			  RegChanHash[hashEnt].lastchan = killme->hashprev;
		}
	}
	while(killme->firstOp)
	  delChanOp(killme, killme->lastOp);
	while(killme->firstAkick)
	  delChanAkick(killme, killme->lastAkick);
	
	sfree(killme);
}

/* now delete the user from the channel, gee is this easy or what? no. :> */
void delChanUser(ChanList *channel, cNickList *nick)
{
	if(!nick)
		return;
	if(nick->previous)
		nick->previous->next = nick->next;
	else
		channel->firstUser = nick->next;
	if(nick->next)
		nick->next->previous = nick->previous;
	else
		channel->lastUser = nick->previous;
	
	sfree(nick);
}

/* remove channel ban */
void delChanBan(ChanList *channel, cBanList *item)
{
	if(item->previous)
	  item->previous->next = item->next;
	else
	  channel->firstBan = item->next;
	if(item->next)
	  item->next->previous = item->previous;
	else
	  channel->lastBan = item->previous;
	
	sfree(item);
}


/* remove channel akick */
void delChanAkick(RegChanList *channel, cAkickList *item)
{
	if(item->previous)
	  item->previous->next = item->next;
	else
	  channel->firstAkick = item->next;
	if(item->next)
	  item->next->previous = item->previous;
	else
	  channel->lastAkick = item->previous;
	
	sfree(item);
	channel->akicks--;
	indexAkickItems(channel);
}

/* remove channel ban */
void delChanOp(RegChanList *channel, cAccessList *item)
{
	if(item->previous)
	  item->previous->next = item->next;
	else
	  channel->firstOp = item->next;
	if(item->next)
	  item->next->previous = item->previous;
	else
	  channel->lastOp = item->previous;
	
	sfree(item);
	channel->ops--;
	indexOpItems(channel);
}



/* find a channel in the list, return NULL if its not found */
ChanList *getChanData(char *name)
{
	long hashEnt;
	ChanList *tmpchan;

	if(!name)
	  return NULL;
	
	hashEnt = getHashKey(name) % CHANHASHSIZE;

	if(ChanHash[hashEnt].chan) {
		for(tmpchan = ChanHash[hashEnt].chan;tmpchan;tmpchan=tmpchan->hashnext) {
			if(!strcasecmp(tmpchan->name, name))
			  return tmpchan;
		}
	}
	return NULL;
}

RegChanList *getRegChanData(char *name)
{
	long hashEnt;
	RegChanList *tmpchan;

	if(!name)
	  return NULL;
	
	hashEnt = getHashKey(name) % CHANHASHSIZE;

	if(RegChanHash[hashEnt].chan) {
		for(tmpchan = RegChanHash[hashEnt].chan;tmpchan;tmpchan=tmpchan->hashnext) {
			if(!strcasecmp(tmpchan->name, name))
			  return tmpchan;
		}
	}

	/* HASH FAILED - SEARCH THE HARD WAY 
		OH, WE PRAY WE DON'T NEED THIS ANYMORE 

	for (tmpchan = firstRegChan;tmpchan;tmpchan = tmpchan->next) {
		if (!strcasecmp(tmpchan->name, name))
			return tmpchan;
	}

	*/
	return NULL;	
}


cNickList *getChanUserData(ChanList *chan, UserList *data)
{
	if(!data || !chan)
		return NULL;	
	else {
		cNickList *tmpnick;
		for(tmpnick = chan->firstUser;tmpnick;tmpnick=tmpnick->next) {
			if(tmpnick->person == data)
			  return tmpnick;
		}
		return NULL;
	}
}

int getChanOp(RegChanList *chan, char *nick)
{
	int highest;
	UserList *tmp;
	cAccessList *tmpnick;
        nLinkList *tmp2;

	if(!nick || !chan)
	  return 0;
	
	tmp = getNickData(nick);

	if(tmp == NULL)
	  return 0;

      highest = 0;

	if(chan->flags & CIDENT) {
		UserList *check = getNickData(nick);
		if(check && check->reg) {
			if(check->caccess == 3) {
				tmpnick = getChanOpData(chan, nick);
				if(tmpnick)
				  highest = tmpnick->uflags;
			}
		}
	}
	else {
		if(tmp->caccess > 1 && (tmpnick = getChanOpData(chan, nick)))
		  highest = tmpnick->uflags;
		else {
			for(tmpnick = chan->firstOp;tmpnick;tmpnick=tmpnick->next) {
				if(checkAccess(tmp->user, tmp->host, getRegNickData(tmpnick->nick)) == 1) {
					if(highest < tmpnick->uflags)
					  highest = tmpnick->uflags;
				}
			}
		}
	}
	
	/* And now for the remote ID's -GZ */

	if(tmp->linked) {
	        for(tmp2=tmp->firstLink;tmp2;tmp2=tmp2->next) {
			        tmpnick = getChanOpData(chan, tmp2->nick);
                                if(tmpnick)
					if(tmpnick->uflags > highest)
	                                	highest = tmpnick->uflags;
	        }
	}
	
	if(!highest) {
		cAkickList *tmpak;
		char hostmask[HOSTLEN+USERLEN+NICKLEN];
		sprintf(hostmask, "%s!%s@%s", tmp->nick, tmp->user, tmp->host);
		for(tmpak = chan->firstAkick;tmpak;tmpak=tmpak->next) {
			if(!match(hostmask, tmpak->mask)) {
				highest = -1;
				break;
			}
		}
	}
	return highest;
}

cAccessList *getChanOpData(RegChanList *chan, char *nick)
{
	if(!nick || !chan)
	  return NULL;
	
	else {
		cAccessList *tmpnick;
		for(tmpnick = chan->firstOp;tmpnick;tmpnick=tmpnick->next) {
			if(!strcasecmp(tmpnick->nick, nick)) {
				return tmpnick;
			}
		}
		return NULL;
	}
}

cBanList *getChanBan(ChanList *chan, char *ban)
{
	cBanList *tmp;
	
	if(!chan || !ban)
	  return NULL;
	
	for(tmp=chan->firstBan;tmp;tmp=tmp->next)
	  {
		  if(!strcasecmp(ban, tmp->ban))
			 return tmp;
	  }
	return NULL;
}

cAkickList *getChanAkick(RegChanList *chan, char *akick)
{
	cAkickList *tmp;
	
	if(!chan || !akick)
	  return NULL;
	for(tmp=chan->firstAkick;tmp;tmp=tmp->next) {
		if(!strcasecmp(akick, tmp->mask))
		  return tmp;
	}	
	return NULL;
}

void indexAkickItems(RegChanList *chan) {
	cAkickList *tmp;
	int i=1;
	for(tmp=chan->firstAkick;tmp;tmp=tmp->next) {
		tmp->index=i;
		i++;
	}
}

void indexOpItems(RegChanList *chan) {
	cAccessList *tmp;
	int i=1;
	for(tmp=chan->firstOp;tmp;tmp=tmp->next) {
		tmp->index=i;
		i++;
	}
}
 
int isFounder(RegChanList *chan, UserList *nick)
{
       char *from = nick->nick;

/* RETURNS 0 ON FAIL, 1 FOR TRUE FOUNDER MATCH, 2 FOR SERVICES OVERRIDE */

	if (nick->caccess < 3)
		return 0;

	if ((!strcasecmp(chan->founder,from)) && (nick->caccess > 2))
		return 1;
	else if (isRoot(nick))
		return 2;
	else
		return 0;
}

int countRegChans(char *nick)
{
/* THIS FUNCTION COUNTS THE TOTAL CHANNELS REGISTERED TO <NICK> */
/* USES LITTLE MORE CPU, BUT IT'S WORTH THE EXPENSE */

	RegChanList *tmp;
	int total_count = 0;

	for(tmp=firstRegChan;tmp;tmp=tmp->next)
	{
		if (!strcasecmp(tmp->founder,nick))
			total_count++;
	}
	return total_count;
} 

void remfromallchanops(char *nick)
{
/* THIS FUNCTION WILL REMOVED A SPECIFIED NICK FROM ALL CHANNEL ACCESS LISTS */
/* IF NICK IS CHANNEL FOUNDER, CHANNEL WILL BE EXPIRED */

	RegChanList *tmpchan, *nextchan;
	cAccessList *tmpop, *nextop;
	int need_new_founder;
	int got_new_founder;
	RegNickList *new_founder;

	for(tmpchan=firstRegChan;tmpchan;tmpchan=nextchan)
	{
		nextchan = tmpchan->next;

		need_new_founder = 0;
		got_new_founder = 0;
		tmpchan->ops = 0;

		if (!strcasecmp(tmpchan->founder,nick)) /* LOST THE FOUNDER, FIND ANOTHER OR EXPIRE */
			need_new_founder = 1;

		for (tmpop = tmpchan->firstOp;tmpop;tmpop = nextop)
		{
			nextop = tmpop->next;
			if (!strcasecmp(tmpop->nick, nick)) /* FOUND HIM */
				delChanOp(tmpchan,tmpop);
			else
			{
				if ((need_new_founder) && (tmpop->uflags == FOUNDER))
				{
					new_founder = getRegNickData(tmpop->nick);
					if ((new_founder) && !(new_founder->flags & NACTIVE) && (new_founder->chans < ChanLimit))
					{
						got_new_founder = 1;
						strcpy(tmpchan->founder,tmpop->nick);
						tmpop->uflags = FOUNDER;
						new_founder->chans++;
						logDump(chanlog, "%s FOUNDER LOST, TRANSFERRED TO %s.", tmpchan->name, tmpop->nick);
					}
				}
				tmpchan->ops++;
			}					
		}

		if ((!tmpchan->ops) || ((need_new_founder) && (!got_new_founder)))
		{
			ChanList *ick;
			ick = getChanData(tmpchan->name);
			if(ick)
				ick->reg = NULL;
			delRegChan(tmpchan);
			logDump(chanlog, "%s EXPIRE (FOUNDER EXPIRED OR DELETED)", tmpchan->name);
			mostchans--;
		}
	}
}
			
char *showchanoplevel(int level)
{
/* SHOWS A FRIENDLY TEXT VERSION OF THE INTERNAL CHANOP LEVELS */
/* RETURNS POINTER TO TEMP STRING - DESTROYED ON CONSECUTIVE CALLS */

	static char retstr[20];
	char tempstr[6];
	
	strcpy(retstr,"AKICK");
	if (level >= OPNONE)
		strcpy(retstr,"NORMAL");
	if (level >= MAOP)
		strcpy(retstr,"VOICE");
#ifdef UNREAL
	if (level >= HOP)
		strcpy(retstr, "HOP");
#endif
	if (level >= AOP)
		strcpy(retstr,"AOP");
	if (level >= MSOP)
		strcpy(retstr,"MSOP");
	if (level >= SOP)
		strcpy(retstr,"SOP");
	if (level >= MFOUNDER)
		strcpy(retstr,"COFOUNDER");
	if (level == FOUNDER)
		strcpy(retstr,"FOUNDER");

	sprintf(tempstr," (%i)",level);
	strcat(retstr,tempstr);

	return((char *)retstr);
}

void initChanData(ChanList *chan) {
	bzero(chan->name, CHANLEN);
	chan->firstBan = chan->lastBan = NULL;
	chan->firstUser = chan->lastUser = NULL;
	chan->modes = 0;
	chan->reg = NULL;
	chan->next = chan->previous = NULL;
}

void initRegChanData(RegChanList *chan) {
	bzero(chan->name, CHANLEN);
	bzero(chan->founder, NICKLEN);
	chan->facc = 0;
	bzero(chan->desc, 101);
	bzero(chan->topic, 307);
	bzero(chan->tsetby, NICKLEN);
	bzero(chan->url, 128);
	bzero(chan->password, 16);
	chan->ttimestamp = 0;
	chan->mlock = 0;
	chan->flags = 0;
	chan->timereg = 0;
	chan->timestamp = 0;
	bzero(chan->key, 23);
	chan->limit = 0;
	chan->ops = 0;
	chan->akicks = 0;
	chan->tlocklevel = 0;
	chan->restrictlevel = 0;
	chan->firstOp = chan->lastOp = NULL;
	chan->firstAkick = chan->lastAkick = NULL;
}

void sendToChanOps(ChanList *chan, char *format, ...)
{
	char buffer[512];
	va_list stuff;
	cNickList *tmp;

	if(!chan || (chan->reg && chan->reg->flags & CQUIET))
		return;
	
	va_start(stuff, format);
	vsprintf(buffer, format, stuff);
	
	for(tmp=chan->firstUser;tmp;tmp=tmp->next) {
		if(tmp->op) {
			uSend(ChanServ, tmp->person, GetReply(RPL_TOCHANOPS, tmp->person), chan->name, buffer);
		}
	}
}

void banKick(ChanList *chan, UserList *nick, char *format, ...)
{
	char user[USERLEN], host[HOSTLEN], theirmask[USERLEN+HOSTLEN+3];
	char buffer[256];
	va_list stuff;
	cBanList *tmpban;
	
	strcpy(user, nick->user);
	strcpy(host, nick->host);
	
	mask(user, host, 1, theirmask);

	sSend(":%s MODE %s +b %s", ChanServ, chan->name, theirmask);
	tmpban = smalloc(sizeof(cBanList));
	strcpy(tmpban->ban, theirmask);
	addChanBan(chan, tmpban);
#ifdef CDEBUG
	sSend(":%s GNOTICE :Added ban %s on %s", myname, theirmask, chan->name);
#endif
	va_start(stuff, format);
	vsprintf(buffer, format, stuff);
	
	sSend(":%s KICK %s %s :%s", ChanServ, chan->name, nick->nick, buffer);
	remUserFromChan(nick, chan->name);
}
	

void createGhostChannel(char *chan) {
	sSend(":%s JOIN %s", ChanServ, chan);
	sSend(":%s MODE %s +o %s", ChanServ, chan, ChanServ);
	sSend(":%s MODE %s +sil 1", ChanServ, chan);
}

void deleteGhostChannel(char *chan) {
	sSend(":%s MODE %s -sil", ChanServ, chan);
	sSend(":%s PART %s", ChanServ, chan);
}

void makeModeLockStr(RegChanList *chan, char *modelock) {
	char *p = modelock;
	memset(modelock, ' ', 20);
	
	if(chan->mlock & (PM_I|PM_L|PM_K|PM_M|PM_N|PM_P|PM_S|PM_T|PM_C|PM_R)) {
		*p++ = '+';
		if(chan->mlock & PM_I)
		  *p++ = 'i';
		if(chan->mlock & PM_L)
		  *p++ = 'l';
		if(chan->mlock & PM_K)
		  *p++ = 'k';
		if(chan->mlock & PM_M)
		  *p++ = 'm';
		if(chan->mlock & PM_N)
		  *p++ = 'n';
		if(chan->mlock & PM_P)
		  *p++ = 'p';
		if(chan->mlock & PM_S)
		  *p++ = 's';
		if(chan->mlock & PM_T)
		  *p++ = 't';
		if(chan->mlock & PM_C)
		  *p++ = 'c';
                if(chan->mlock & PM_R)
                  *p++ = 'R';
	}
	if(chan->mlock & (MM_I|MM_L|MM_K|MM_M|MM_N|MM_P|MM_S|MM_T|MM_C|MM_R)) {
		*p++ = '-';
		if(chan->mlock & MM_I)
		  *p++ = 'i';
		if(chan->mlock & MM_L)
		  *p++ = 'l';
		if(chan->mlock & MM_K)
		  *p++ = 'k';
		if(chan->mlock & MM_M)
		  *p++ = 'm';
		if(chan->mlock & MM_N)
		  *p++ = 'n';
		if(chan->mlock & MM_P)
		  *p++ = 'p';
		if((chan->mlock & MM_S) && (!((isForced(chan->name) || (chan->flags&CFORCES)) && !(chan->flags&CFEXEMPT))))
		  *p++ = 's';
		if(chan->mlock & MM_T)
		  *p++ = 't';
		if(chan->mlock & MM_C)
		  *p++ = 'c';
                if(chan->mlock & MM_R)
                  *p++ = 'R';
	}
	
	*p++ = 0;
}

/* This will determine if a user is a scanbot -GZ */

int 	IsScanBot(UserList *nick) {
	
	char	*s;
	
	s = nick->nick;

	if(nick->reg)
		return 0;

	if(!strcmp(nick->nick, nick->realname) || strlen(nick->realname) == 0) {
		while(*s) {
			if(!islower(*s))
				return 0;
			s++;
		}
		return 1;
	}

	return 0;
}

/* add a user to a channel, fun...really */
void addUserToChan(UserList *nick, char *channel)
{
	char chan[CHANLEN];
	int a=0;
	ChanList *tmp;
	cNickList *person;
        RegNickList *webbie, *aol;
	
	if(*channel == ':')
		channel++;

	while(*channel) {
		bzero(chan, CHANLEN);
		a=0;
		while ((*channel) && (*channel != ',')) {
			chan[a] = *channel;
			a++;
			channel++;
		}
		chan[a] = 0;
		if(*channel)
			channel++;
		  
		if(chan[0] == '0')
			remFromAllChans(nick);
		else if(chan[0] != '+') {
			tmp = getChanData(chan);

			if((!tmp) || (!tmp->firstUser)) {
				/* FIRST USER IN CHANNEL */

				/* INSERT NEW CHANNEL INTO INTERNAL LIST */
				if(!tmp) {
					tmp = smalloc(sizeof(ChanList));
					initChanData(tmp);
					strncpy(tmp->name,chan, CHANLEN);
					addChan(tmp);
				}

				person = smalloc(sizeof(cNickList));
				person->person = nick;
				person->op = 0;
					 
				addChanUser(tmp, person);

				if (isBanChannel(tmp->name)) {
					if (!isServop(nick)) {
						createGhostChannel(tmp->name);
						timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
						sSend(":%s MODE %s +b *!*@*", ChanServ, chan);
						sSend(":%s KICK %s %s :Channel is not permitted on this network.", ChanServ, chan, nick->nick);
						remUserFromChan(nick, chan);
						return;
					}
				}

#ifdef CDEBUG
				sSend(":%s GNOTICE :Created channel %s, added %s to it", myname, chan, nick->nick);
#endif
					 
				tmp->reg = getRegChanData(chan);

                               /* new closed channels check -GZ */
                        
                                if(tmp->reg && tmp->reg->flags & CCLOSED) {
                                                if (!isServop(nick)) {
                                                        createGhostChannel(tmp->name);
                                                        timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                                                        sSend(":%s MODE %s +b *!*@*", ChanServ, chan);
                                                        sSend(":%s KICK %s %s :This channel has been \002CLOSED\002 by the " NETWORKNAME " administration.", ChanServ, chan, nick->nick);
                                                        remUserFromChan(nick, chan);
                                                        return;
                                                }
                                }

				if((isForced(tmp->name) || (tmp->reg && tmp->reg->flags&CFORCES)) && !(tmp->reg && tmp->reg->flags&CFEXEMPT)) {
					sSend(":%s MODE %s +s", ChanServ, tmp->name);
					uSend(ChanServ, nick, GetReply(RPL_AUTOSET_S, nick));
					if(tmp->reg)
						tmp->reg->flags|=CPRIVATE;
				}
				/* IF THE CHANNEL IS REGISTERED -- CHECK OUT THE REGISTRATION DATA */

				if(tmp->reg) {
					char mode[20];
                                               
					/* OPS/BANS LIST */
					a=getChanOp(tmp->reg, nick->nick);

                                    /* Thou shall not akick servops */

                                        if(!isServop(nick)) {
					 char reason[500];
					 if(a == -1) {  /* IS AKICKED? */
						cAkickList *blah;
						char userhost[NICKLEN + USERLEN + HOSTLEN];
						
						sprintf(userhost, "%s!%s@%s", nick->nick, nick->user, nick->host);
						for(blah = tmp->reg->firstAkick;blah;blah=blah->next) {
							if(!match(userhost, blah->mask)) {
								cBanList *newban = smalloc(sizeof(cBanList));
								sSend(":%s MODE %s +b %s", ChanServ, chan, blah->mask);
								strcpy(newban->ban, blah->mask);
								sprintf(reason, "You have been permanently banned by ");
								strcat(reason, blah->reason);
								addChanBan(tmp, newban);
								break;
							}
						}
						createGhostChannel(tmp->name);
						timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
						sSend(":%s KICK %s %s :%s.", ChanServ, chan, nick->nick, reason);
						remUserFromChan(nick, chan);
						return;
					 }
                                        }
                                        
                                         /* Temporary AOL protection */

                                        if(tmp->reg->flags & CNOAOL) {
                                             if(!match(nick->host, "*.aol.com")) {
                                                 aol = getRegNickData(nick->nick);
                                                 if(!aol) {
							createGhostChannel(tmp->name);
							timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                                                        sSend(":%s KICK %s %s :You need a registered nick to join this channel", ChanServ, chan, nick->nick);
							remUserFromChan(nick, chan);
							return;
                                                 }
                                             }
                                        }
				
					/* Scanbot protection -GZ */

                                        if(tmp->reg->flags & CSCAN && a < 3) {
						if(IsScanBot(nick)) {
							createGhostChannel(tmp->name);
							timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                	                                sSend(":%s KICK %s %s :This channel does not allow Scanbots.", ChanServ, chan, nick->nick);
							remUserFromChan(nick, chan);
							return;
						}
                                        }
     
					if((tmp->reg->restrictlevel != 0) && ((a < tmp->reg->restrictlevel) || (!a && ((tmp->reg->mlock&PM_K)||(tmp->reg->mlock&PM_I))))) {
						/* IS BELOW RESTRICT LEVEL */
						createGhostChannel(tmp->name);
						timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
						if(!isServop(nick)) { 
//                                                    banKick(tmp, nick, "This channel is restricted to level %i and above", tmp->reg->restrictlevel);
						      banKick(tmp, nick, "Sorry. This is a restricted access channel.");
                                                }
						return;
					}

#ifdef UNREAL
					if(a && a < HOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
#else
					if(a && a < AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
#endif
						/* IS A VOICED OP +v */
						sSend(":%s MODE %s +v %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANVOICE;
						tmp->reg->timestamp = CTime;
					}
#ifdef UNREAL
					else
					if(a && a < HOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						/* IS A VOICED OP +v */
						sSend(":%s MODE %s +h %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANHALFOP;
						tmp->reg->timestamp = CTime;
					}
#endif
					else if(a && a >= AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						/* IS AUTO-OPPED */
						sSend(":%s MODE %s +o %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANOP;
						tmp->reg->timestamp = CTime;
                                        /* This will let Webbies know they are getting ops */
                                                webbie = getRegNickData(nick->nick);
                                                if(webbie) {
                                                    if (webbie->flags2 & NWEBTV) 
                                                    sSend(":%s PRIVMSG %s :You are now a channel operator in %s.", WebServ, nick->nick, tmp->name);
                                                }
					}

					if ((tmp->reg->flags & CKTOPIC) && (tmp->reg->topic[0]) && !(tmp->reg->flags & CFREEZE))
						/* IF KEEPTOPIC - RESTORE SAVED TOPIC */
						sSend(":%s TOPIC %s %s %lu :%s", ChanServ, tmp->name, tmp->reg->tsetby, tmp->reg->ttimestamp, tmp->reg->topic);

                               /* ChanServ will now enforce that absolutely nobody but SO's get ops */

                                        if ((tmp->reg->flags & CFREEZE) && !isServop(nick)) {
                                             uSend(ChanServ, nick, GetReply(RPL_NOOPSFROZEN, nick));
                                             sSend(":%s MODE %s -o %s", ChanServ, tmp->name, nick->nick);
                                        }

   					makeModeLockStr(tmp->reg, mode);
					doRegChanMode(tmp->reg);

					if ((tmp->reg->flags & CNOTICE) && (tmp->reg->desc[0])) {
					     uSend(ChanServ, nick, GetReply(RPL_CHANDESC, nick), tmp->name, tmp->reg->desc);
					}
				}
			}
			else {
				/* NOT FIRST USER IN CHANNEL */			

				person = getChanUserData(tmp, nick);

				tmp->reg = getRegChanData(chan);

				if (person)
					delChanUser(tmp, person);

		 		person = smalloc(sizeof(cNickList));
				person->person = nick;
				person->op = 0;
				addChanUser(tmp, person);

                               /* new closed channels check -GZ */
                        
                                if(tmp->reg && tmp->reg->flags & CCLOSED) {
                                                if (!isServop(nick)) {
                                                        createGhostChannel(tmp->name);
                                                        timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                                                        sSend(":%s MODE %s +b *!*@*", ChanServ, chan);
                                                        sSend(":%s KICK %s %s :This channel has been \002CLOSED\002 by the " NETWORKNAME " administration.", ChanServ, chan, nick->nick);
                                                        remUserFromChan(nick, chan);
                                                        return;
                                                }
                                }

				if(tmp->reg) {
					a=getChanOp(tmp->reg, nick->nick);
					/* IS AKICKED?? */
                                      /* Thou shall not akick servops */
                                        if(!isServop(nick)) {
					 if(a == -1) {
						cAkickList *blah;
						char userhost[NICKLEN + USERLEN + HOSTLEN];
						char reason[500];
							 
						sprintf(userhost, "%s!%s@%s", nick->nick, nick->user, nick->host);
						for(blah = tmp->reg->firstAkick;blah;blah=blah->next) {
							if(!match(userhost, blah->mask)) {
								cBanList *newban = smalloc(sizeof(cBanList));
								sSend(":%s MODE %s +b %s", ChanServ, chan, blah->mask);
								strcpy(newban->ban, blah->mask);
								sprintf(reason, "You have been permanently banned by ");
								strcat(reason, blah->reason);
								addChanBan(tmp, newban);
								break;
							}
						}
						sSend(":%s KICK %s %s :%s", ChanServ, chan, nick->nick, reason);
						remUserFromChan(nick,chan);
						return;
					 }
                                        }

                                         /* Temporary AOL protection */

                                        if(tmp->reg->flags & CNOAOL) {
                                             if(!match(nick->host, "*.aol.com")) {
                                                 aol = getRegNickData(nick->nick);
                                                 if(!aol) {
                                                      sSend(":%s KICK %s %s :You need a registered nick to join this channel", ChanServ, chan, nick->nick);
                                                      sSend(":%s MODE %s +b *!*%s@%s", ChanServ, chan, nick->user, nick->host);
                                                      remUserFromChan(nick, chan);
					              return;
                                                 }
                                             }
                                        }
     
					/* Scanbot protection -GZ */

                                        if(tmp->reg->flags & CSCAN && a < 3) {
						if(IsScanBot(nick)) {
                                                        sSend(":%s MODE %s +b *!*@%s", ChanServ, chan, nick->host);
                	                                sSend(":%s KICK %s %s :This channel does not allow Scanbots.", ChanServ, chan, nick->nick);
							remUserFromChan(nick, chan);
							return;
						}
                                        }
     

					if(a < tmp->reg->restrictlevel && tmp->reg->restrictlevel != 0) {
						/* BELOW RESTRICT LEVEL */
                                                if (!isServop(nick)) {
//						    banKick(tmp, nick, "This channel is restricted to level %i+", tmp->reg->restrictlevel);
						    banKick(tmp, nick, "Sorry. This is a restricted access channel.");

                                                }
						return;
					}
#ifdef UNREAL
					if(a && a < HOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
#else
					if(a && a < AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
#endif
						/* AUTO VOICE */
						sSend(":%s MODE %s +v %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANVOICE;
						tmp->reg->timestamp = CTime;
					}
#ifdef UNREAL
					else
					if(a && a < AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						/* AUTO HALFOP */
						sSend(":%s MODE %s +h %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANHALFOP;
						tmp->reg->timestamp = CTime;
					}
#endif
					else if(a && a >= AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						/* AUTO OP */
						sSend(":%s MODE %s +o %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANOP;
						tmp->reg->timestamp = CTime;

                                        /* This will let Webbies know they are getting ops */
                                                webbie = getRegNickData(nick->nick);
                                                if(webbie) {
                                                    if (webbie->flags2 & NWEBTV)
                                                    sSend(":%s PRIVMSG %s :You are now a channel operator in %s.", WebServ, nick->nick, tmp->name); 
                                                }
					}
                                        
                                        /* Make sure nobody gets ops */

                                        if ((tmp->reg->flags & CFREEZE) && !isServop(nick)) {
                                             uSend(ChanServ, nick, GetReply(RPL_NOOPSFROZEN, nick));
                                             sSend(":%s MODE %s -o %s", ChanServ, tmp->name, nick->nick);
                                        }

					if ((tmp->reg->flags & CNOTICE) && (tmp->reg->desc[0])) {
					      uSend(ChanServ, nick, GetReply(RPL_CHANDESC, nick), tmp->name, tmp->reg->desc);
					}
				}
#ifdef CDEBUG
					 sSend(":%s GNOTICE :Added %s to channel %s", myname, nick->nick, chan);
#endif
			}
		}
	}
}

void remUserFromChan(UserList *nick, char *channel)
{
	char chan[CHANLEN];
	int a=0; 
	ChanList *tmp;
	cNickList *person;
	
	if(*channel == ':')
	  	channel++;
	while(*channel)
	{
		bzero(chan, CHANLEN);
		a=0;
		while(*channel != ',' && *channel != 0) {
			chan[a] = *channel;
			a++;
			channel++;
		}
		chan[a] = 0;
		a = 0;

		if(*channel)
			channel++;
		  
		if(chan[0] != '+')
		{
			tmp = getChanData(chan);
			if(!tmp)
				return;
			person=getChanUserData(tmp,nick);
			if(!person)
				return;

			delChanUser(tmp, person);
#ifdef CDEBUG
			sSend(":%s GNOTICE :Removed %s from %s", myname, nick->nick, chan);
#endif			  
				 
			if(!tmp->firstUser) {						 
				if(tmp->reg && (tmp->reg->flags & CCLOSED))
					sSend(":%s PART %s", ChanServ, tmp->name);
				delChan(tmp);
#ifdef CDEBUG
				sSend(":%s GNOTICE :Removed channel %s", myname, chan);
#endif
			}
		}
	}
}


void doKick(UserList *from, UserList *to, char *channel) 
{
	ChanList *chan;

	if ((!from) || (!to) || (!channel) || (channel[0] != '#'))
		return;

	chan = getChanData(channel);

	if (!chan)
		return;

	if (chan->reg)
	{
		if (chan->reg->flags & CPROTOP)
		{
			cAccessList *fromop, *toop;
			int fromuflags = 0, touflags = 0;

			fromop = getChanOpData(chan->reg, from->nick);
			if (fromop)
				fromuflags = fromop->uflags;

			toop = getChanOpData(chan->reg, to->nick);
			if (toop)
				touflags = toop->uflags;

			if ((touflags > fromuflags) && (!isOper(from)))
	  			sSend(":%s KICK %s %s :It's bad manners to kick people with higher access than you.", ChanServ, channel, from->nick);
		}
	}
	remUserFromChan(to, channel);
}

void remFromAllChans(UserList *nick)
{
	ChanList *tmp = NULL;
#ifdef CDEBUG
	sSend(":%s GNOTICE :Clearing %s from ALL channels", myname, nick->nick);
#endif
	for(tmp=firstChan;tmp;tmp=tmp->next) {
		if(tmp) {
			delChanUser(tmp, getChanUserData(tmp, nick));
			if(!tmp->firstUser)
		  		delChan(tmp);
		}
	}
}

void changeNickOnAllChans(UserList *oldnick, UserList *newnick)
{
	ChanList *tmp;
	cNickList *oldent, *newent;
#ifdef CDEBUG
	  sSend(":%s GNOTICE :Changing nick for %s->%s", myname, oldnick->nick, newnick->nick);
#endif
	for(tmp=firstChan;tmp;tmp=tmp->next) {
		if((oldent=getChanUserData(tmp, oldnick))) {
			newent = smalloc(sizeof(cNickList));
			newent->op = oldent->op;
			newent->person = newnick;
			delChanUser(tmp, oldent);
			addChanUser(tmp, newent);
		}
	}
}


/* set chan modes, called on /os mode, when a user sets a mode, etc etc */

void setChanMode(char **args, int numargs)
{
	int on=1, onarg = 4, i=0;
	ChanList *tmp = getChanData(args[2]);
	cBanList *tmpban;
	cNickList *tmpnick;
	char mode[20];

/* This is for the webbies, so they'll know if they get (de)opped  -GZ */

        UserList *webbie;
        RegNickList *regwebbie;

	int minusS=0;
	
	if(!tmp)
	  return;
	
	if (*args[0] == ':')
		args[0]++;

	for(i=0;args[3][i];i++)
	  {
		  switch(args[3][i]) {
			case '+':
			  on=1;
			  break;
			case '-':
			  on=0;
			  break;
			  
			case 'b':
			  if(on) {
				  tmpban = smalloc(sizeof(cBanList));
				  strcpy(tmpban->ban, args[onarg]);
				  addChanBan(tmp, tmpban);
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Added ban %s on %s", myname, args[onarg], tmp->name);
#endif
			  }
			  else {
				  tmpban = getChanBan(tmp, args[onarg]);
				  if (tmpban) {
#ifdef CDEBUG
					  sSend(":%s GNOTICE :Removed ban %s on %s", myname, args[onarg], tmp->name);
#endif
					  delChanBan(tmp, tmpban);
				  }
			  }
			  onarg++;
			  break;

			case 'c':
			  if(on) {
				  tmp->modes |= PM_C;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +c", myname, tmp->name);
#endif		 
			  }
			  else {
				  tmp->modes -= PM_C;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -c", myname, tmp->name);
#endif
			  }
			  break;
			  
			case 'i':
			  if(on) {
				  tmp->modes |= PM_I;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +i", myname, tmp->name);
#endif
			  }
			  else {
				  tmp->modes -= PM_I;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -i", myname, tmp->name);
#endif
			  }
			  break;
			case 'l':
			  if(onarg == numargs)
				break;
			  if(on) {
				  tmp->modes |= PM_L;
				  if(tmp->reg && !(tmp->reg->mlock & PM_L))
					 tmp->reg->limit = atol(args[onarg]);
				  onarg++;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +l %lu", myname, tmp->name, tmp->reg->limit);
#endif
			  }
			  else {
				  tmp->modes -= PM_L;
				  if(tmp->reg)
					 tmp->reg->limit = 0;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -l", myname, tmp->name, tmp->reg->limit);
#endif
			  }
			  break;
			case 'k':
			  if(onarg == numargs)
				break;
			  if(on) {
				  tmp->modes |= PM_K;
				  if(tmp->reg && !(tmp->reg->mlock & PM_K))
					 strcpy(tmp->reg->key, args[onarg]);
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +k %s", myname, tmp->name, tmp->reg->key);
#endif
			  }
			  else {
				  tmp->modes -= PM_K;
				  if (tmp->reg)
					 tmp->reg->key[0] = 0;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -k %s", myname, tmp->reg->name, args[onarg]);
#endif
			  }
			  onarg++;
			  break;
			case 'm':
			  if(on) {
				  tmp->modes |= PM_M;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +m", myname, tmp->name);
#endif
			  }
			  else {
				  tmp->modes -= PM_M;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -m", myname, tmp->name);
#endif
			  }
			  break;
			case 'n':
			  if(on) {
				  tmp->modes |= PM_N;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +n", myname, tmp->name);
#endif		 
			  }
			  else {
				  tmp->modes -= PM_N;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -m", myname, tmp->name);
#endif
			  }
			  break;
			case 'o':

               /* Create data for WebTV users so they get notified when the receive ops 
                * this will be used to notify Webbies if they get ops (see below) -GZ
                */
                                
                                webbie = getNickData(args[onarg]);
                                regwebbie = getRegNickData(args[onarg]);

				if(on) {
					tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
					if (tmpnick) {
						if(tmp->reg) {
							int newoplevel = getChanOp(tmp->reg, args[onarg]);
							if(((newoplevel < AOP) && (tmp->reg->flags & COPGUARD || tmp->firstUser == tmp->lastUser)) && !isOper(getNickData(args[onarg]))) {
								uSend(ChanServ, webbie, GetReply(RPL_NOOPS, webbie));
								sSend(":%s MODE %s -o %s", ChanServ, tmp->name, args[onarg]);
							}
							else {
								tmpnick->op |= CHANOP;
                                                        
                                                        if(regwebbie) {
                                                        if(regwebbie->flags2 & NWEBTV) 
                                                                sSend(":%s PRIVMSG %s :You are now a channel operator in %s.", WebServ, webbie->nick, tmp->name); 
                                                        }
#ifdef CDEBUG
								sSend(":%s GNOTICE :Oped %s in %s", myname, args[onarg], tmp->name);
#endif
							}
						}
						else {
							tmpnick->op |= CHANOP;
                                                        if(regwebbie) {
                                                        if(regwebbie->flags2 & NWEBTV) 
                                                                sSend(":%s PRIVMSG %s :You are now a channel operator in %s.", WebServ, webbie->nick, tmp->name); 
                                                        }
#ifdef CDEBUG
							sSend(":%s GNOTICE :Oped %s in %s", myname, args[onarg], tmp->name);
#endif
						}
			  		}
				}
				else {
					tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
				  	if(tmpnick) {
						if(tmp->reg) {
							if((getChanOp(tmp->reg, args[onarg]) > getChanOp(tmp->reg, args[0])) && tmp->reg->flags & CPROTOP)
								sSend(":%s MODE %s +o %s", ChanServ, tmp->name, args[onarg]);
						  	else {
								tmpnick->op &= ~CHANOP;
                                                        if(regwebbie) {
                                                        if(regwebbie->flags2 & NWEBTV)
                                                                sSend(":%s PRIVMSG %s :You are no longer a channel operator in %s", WebServ, webbie->nick, tmp->name);
                                                        }
#ifdef CDEBUG
								sSend(":%s GNOTICE :DeOped %s in %s", myname, args[onarg], tmp->name);
#endif
							}
					  	}
					  	else {
							tmpnick->op &= ~CHANOP;
                                                        if(regwebbie) {
                                                        if(regwebbie->flags2 & NWEBTV)
                                                                sSend(":%s PRIVMSG %s :You are no longer a channel operator in %s", WebServ, webbie->nick, tmp->name);
                                                        }
#ifdef CDEBUG
							sSend(":%s GNOTICE :DeOped %s in %s", myname, args[onarg], tmp->name);
#endif
						}
					}
				}
				onarg++;
				break;
			case 'p':
			  if(on) {
				  tmp->modes |= PM_P;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +p", myname, tmp->name);
#endif		 
			  }
			  else {
				  tmp->modes -= PM_P;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -p", myname, tmp->name);
#endif
			  }
			  break;
			case 's':
			  if(on) {
				  tmp->modes |= PM_S;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +s", myname, tmp->name);
#endif		 
			  }
			  else {
				  tmp->modes -= PM_S;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -s", myname, tmp->name);
#endif
				  minusS=1;
			  }
			  break;
			case 't':
			  if(on) {
				  tmp->modes |= PM_T;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s +t", myname, tmp->name);
#endif		 
			  }
			  else {
				  tmp->modes -= PM_T;
#ifdef CDEBUG
				  sSend(":%s GNOTICE :Set %s -t", myname, tmp->name);
#endif
			  }
			  break;
			  
			case 'v':
			  if(on) {
				  tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
				  if(tmpnick) {
					  tmpnick->op |= CHANVOICE;
#ifdef CDEBUG
					  sSend(":%s GNOTICE :Oped %s in %s", myname, args[onarg], tmp->name);
#endif
				  }
			  }
			  else { 
				  tmpnick = getChanUserData(tmp,getNickData(args[onarg]));
				  if(tmpnick) {
					  tmpnick->op &= ~CHANVOICE;
#ifdef CDEBUG
					  sSend(":%s GNOTICE :DeOped %s in %s", myname, args[onarg], tmp->name);
#endif	  
				  }
			  }
			  onarg++;
			  break;
		  }
		  
	  }
	
	if(tmp->reg) {
		makeModeLockStr(tmp->reg, mode);
		doRegChanMode(tmp->reg);
	}
	if (minusS) {
		if ((isForced(tmp->name) || (tmp->reg && tmp->reg->flags&CFORCES)) && !(tmp->reg && tmp->reg->flags&CFEXEMPT)) {
			sSend(":%s MODE %s +s", ChanServ, tmp->name);
			sSend(":%s NOTICE %s :This channel is auto set to mode +s as a precaution for   our younger users. This will not affect the channel in any way, other than to   make it invisible on a channel list.", ChanServ, tmp->name);
		}
	}
}

void setChanTopic(char **args, int numargs)
{
	RegChanList *tmp;
	UserList *changeuser;
	int i=6;
	tmp = getRegChanData(args[2]);
	
	if(!tmp)
	  return;

	changeuser = getNickData(args[3]);

	if ((getChanOp(tmp, args[3]) < tmp->tlocklevel) && (!isOper(changeuser))) {
		if(tmp->topic[0])
		  sSend(":%s TOPIC %s %s %lu :%s", ChanServ, tmp->name, tmp->tsetby, tmp->ttimestamp, tmp->topic);
	}
	
	else {
		strcpy(tmp->tsetby, args[3]);
		tmp->ttimestamp = atol(args[4]);
		
		args[5]++;
		strcpy(tmp->topic, args[5]);
		
		while(i<numargs) {
			strcat(tmp->topic, " ");
			strcat(tmp->topic, args[i]);
			i++;
		}
	}
}

void sendToChanServ(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;

	/* small "hack" to make services work in either /cs #channel cmd
	 * or /cs cmd #channel way */
	char crud[40];
		
	if(index(args[0], '#') && args[1]) {
		strcpy(crud, args[0]);
		strcpy(args[0], args[1]);
		strcpy(args[1], crud);
	}
	
	if(!strcmp(args[0], "\001PING"))
	  {
		  if(numargs<3)
                       uSend(ChanServ, nick, GetReply(RPL_PING1, nick), args[1]);
		  else
                       uSend(ChanServ, nick, GetReply(RPL_PING2, nick), args[1], args[2]);
		  return;
	  }	
					  
  	else if(!strcasecmp(args[0], "HELP") && numargs == 1)
	  {
		  sendHelp(from, "chanserv", NULL, 0);
	  }

	else if(!strcasecmp(args[0], "HELP"))
	  {
		  sendHelp(from, "chanserv", args, numargs);
	  }

	/* Parse and adjust some commands before sending it to the message tables */

	else if(!strcasecmp(args[0], "CHANOP")) {
		int i = 3;
		if(args[1] && args[2]) {
			if(!strcasecmp(args[2], "add"))
			  strcpy(args[0], "addop");
			else if(!match(args[2], "del*"))
			  strcpy(args[0], "delop");
			else if(!strcasecmp(args[2], "list"))
			  strcpy(args[0], "listop");
			else {
				uSend(ChanServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[2]);
				uSend(ChanServ, nick, GetReply(RPL_MOREHELPCMD, nick), ChanServ, "ChanOp");
				return;
			}
		}
		else {
			uSend(ChanServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(ChanServ, nick, GetReply(RPL_MOREHELPCMD, nick), ChanServ, "ChanOp");
			return;
		}
		while(i<numargs) {
			strcpy(args[i-1], args[i]);
			i++;
		}
		numargs--;
		sendToChanServ(nick, args, numargs);
		return;
	} 
	else if(!strcasecmp(args[0], "AKICK")) {
		int i = 3;
		if(args[1] && args[2]) {
			if(!strcasecmp(args[2], "add"))
			  strcpy(args[0], "addak");
			else if(!match(args[2], "del*"))
			  strcpy(args[0], "delak");
			else if(!strcasecmp(args[2], "list"))
			  strcpy(args[0], "listak");
			else {
				uSend(ChanServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[2]);
				uSend(ChanServ, nick, GetReply(RPL_MOREHELPCMD, nick), ChanServ, "Akick");
				return;
			}
		}
		else {
			uSend(ChanServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(ChanServ, nick, GetReply(RPL_MOREHELPCMD, nick), ChanServ, "Akick");
			return;
		}
		while(i<numargs) {
			strcpy(args[i-1], args[i]);
			i++;
		}
		numargs--;
		sendToChanServ(nick, args, numargs);
		return;
	} 

	else 
		msgtabCheck(ChanServ, nick, args, numargs); /* New messagetables to replace old code -GZ */

#ifdef SHOWSYNCHALL
	uSend(ChanServ, nick, GetReply(RPL_NEXTDBSYNC, nick), ((nextCsync - CTime) / 60));
#endif
}

void CinfoCmd(UserList *nick, char **args, int numargs)
{
/* says unused
	char *from = nick->nick; */
	RegChanList *chan;
	char timejunk[80];
	char modestring[20];
	
	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}
	
	uSend(ChanServ, nick, GetReply(RPL_CHANINFO1, nick), chan->name);
	uSend(ChanServ, nick, GetReply(RPL_CHANINFO2, nick), chan->founder);
	if(chan->desc[0])
		uSend(ChanServ, nick, GetReply(RPL_CHANINFO3, nick), chan->desc);
	if(chan->topic[0])
		uSend(ChanServ, nick, GetReply(RPL_CHANINFO4, nick), chan->topic, chan->tsetby);
	if(chan->flags & CIDENT || chan->flags & CPROTOP || chan->flags & CKTOPIC || 
	  	chan->flags & COPGUARD || chan->flags & CQUIET || chan->flags & CCLASSIC ||
	  	chan->flags & CNOTICE)

		uSend(ChanServ, nick, GetReply(RPL_CHANINFO5, nick), 
			chan->flags & CIDENT ? "Ident. " : "",
			chan->flags & CPROTOP ? "Protected Ops. " : "",
			chan->flags & CKTOPIC ? "Preserve Topic. " : "",
			chan->flags & COPGUARD ? "Op Guard. " : "",
			chan->flags & CQUIET ? "Quiet Changes. " : "",
			chan->flags & CCLASSIC ? "No AutoOping." : "",
			chan->flags & CNOTICE ? "Join Notice." : "",
			chan->flags & CSCAN ? "Scan Guard." : ""
		);

	strftime(timejunk, 80, "%a %d/%m/%Y %T", gmtime(&chan->timestamp));
	uSend(ChanServ, nick, GetReply(RPL_CHANINFO8, nick), timejunk);
	strftime(timejunk, 80, "%a %d/%m/%Y %T", gmtime(&chan->timereg));
	uSend(ChanServ, nick, GetReply(RPL_CHANINFO7, nick), timejunk);
	if(isOper(nick)) {
		uSend(ChanServ, nick, GetReply(RPL_CHANINFO9, nick),
			chan->flags & CCLOSED ? "Closed. " : "",
			chan->flags & CNOAOL ? "NoAol. " : "",
			chan->flags & CFORCES ? "Forces. " : "",
			chan->flags & CHOLD ? "Held. " : "",
			chan->flags & CFREEZE ? "Frozen. " : "",
			chan->ops, chan->akicks
		);
	}
	makeModeLockStr(chan, modestring);
	if(modestring[0] == 0)
		strcpy(modestring, "(none)");
	uSend(ChanServ, nick, GetReply(RPL_CHANINFO6, nick), chan->tlocklevel, chan->restrictlevel, modestring);

	uSend(ChanServ, nick, GetReply(RPL_CHANINFO10, nick));
}


void CaccessCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	RegChanList *tmp;
	int test_level;

	tmp = getRegChanData(args[1]);
	
	if(!tmp) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

	if(numargs > 2)
	{
		test_level = getChanOp(tmp, args[2]);
		uSend(ChanServ, nick, GetReply(RPL_CHANACCESS1, nick), args[2], showchanoplevel(test_level), tmp->name);
	}
	else
	{
		test_level = getChanOp(tmp, from);
		uSend(ChanServ, nick, GetReply(RPL_CHANACCESS2, nick), showchanoplevel(test_level), tmp->name);
	}
	return;
}


void CregisterCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	cNickList *tmp;
	cAccessList *founder;
	ChanList *chan;
	RegChanList *reg;
	int i;

	if (!(nick->reg) || !(nick->reg->flags & NACTIVE)) {
		uSend(ChanServ, nick, GetReply(ERR_NEEDREGNICK, nick));
		return;
	}

	if(!(nick->caccess > 2)) {
		uSend(ChanServ, nick, GetReply(ERR_NEEDID, nick));	
		return;
	}

	chan=getChanData(args[1]);
	tmp=getChanUserData(chan, nick);

	if(!chan || !tmp || !tmp->op) {
		uSend(ChanServ, nick, GetReply(ERR_NOTOPEDINCHAN, nick));
		return;
	}

	reg = getRegChanData(args[1]);

	if(reg) {
		uSend(ChanServ, nick, GetReply(ERR_CHANALREADYREG, nick));
		return;
	}

	/* PROPER FIX - THIS VERIFIES ACTUAL CHANNELS REGISTERED TO NICK */

	nick->reg->chans = countRegChans(from);

	if((nick->reg->chans >= ChanLimit && !isServop(nick)) && (!isOper(nick)) && strcasecmp(from, "" NETWORKNAME "")) {
		uSend(ChanServ, nick, GetReply(ERR_TOMANYREGCHAN1, nick));
		sSend(":%s GLOBOPS :%s is trying to register too many channels", ChanServ, from);
		return;
	}

	nick->reg->chans++;
	reg = smalloc(sizeof(RegChanList));
	initRegChanData(reg);
	chan->reg = reg;
	strcpy(reg->name, chan->name);
	strcpy(reg->founder, nick->reg->nick);
	for(i=0;i<8;i++)
		reg->password[i] = (((u_short)rand())%26)+97;
	reg->timereg = CTime;
	reg->timestamp = CTime;
	sendToChanOps(chan, "%s just registered %s", from, reg->name);

	uSend(ChanServ, nick, GetReply(RPL_CHANNOWREG, nick), args[1], from);

	uSend(ChanServ, nick, GetReply(RPL_CHANREGNOTE, nick));
	uSend(ChanServ, nick, GetReply(RPL_MOREHELP, nick), ChanServ);

	if(isForced(reg->name)) {
		sSend(":%s MODE %s +rs", ChanServ, reg->name);
		uSend(ChanServ, nick, GetReply(RPL_AUTOSET_S, nick));
		reg->flags|=CPRIVATE;
		sSend(":%s GLOBOPS :%s registered %s (forced +s)", ChanServ, from, reg->name);
	}
	else {
		sSend(":%s GLOBOPS :%s registered %s", ChanServ, from, reg->name);
	        sSend(":%s MODE %s +r", ChanServ, reg->name);
	}

	founder = smalloc(sizeof(cAccessList));
	strcpy(founder->nick, nick->reg->nick);
	founder->uflags = FOUNDER;
	
	addChanOp(reg, founder);
	reg->flags |= CKTOPIC;
	reg->flags |= CSCAN;
	reg->mlock |= PM_N;
	reg->mlock |= PM_T;
	reg->mlock |= MM_K;
	addRegChan(reg);
	mostchans++;
	logDump(chanlog, "%s!%s@%s REGISTER (%s)", nick->nick, nick->user, nick->host, reg->name);
}

void CfounderCmd(UserList *nick, char **args, int numargs)
{
      char *from = nick->nick;
	cAccessList *newitem;
	RegChanList *chan;
	RegNickList *add;
	int founder_control;

	if(numargs<4) {
		uSend(ChanServ, nick, GetReply(ERR_BADPARAMS, nick));
		return;
	}
	
	chan = getRegChanData(args[1]);

	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	founder_control = isFounder(chan,nick);

	if(!founder_control) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

	add = getRegNickData(args[3]);
	if (!add || !(add->flags & NACTIVE))
	{
		uSend(ChanServ, nick, GetReply(ERR_NICKNOTREGGED, nick), args[3]);
		return;
	}

        if ((add->flags & NNOOP) && !isRoot(nick)) {
		
		uSend(ChanServ, nick, GetReply(ERR_NOOPON, nick), add->nick);
		return;
	}

	add->chans = countRegChans(add->nick);

	if ((add->chans >= ChanLimit) && (founder_control !=2) && strcasecmp(add->nick, "" NETWORKNAME ""))
	{

		uSend(ChanServ, nick, GetReply(ERR_TOMANYREGCHAN2, nick), add->nick);
		return;
	}

	newitem = getChanOpData(chan, add->nick);

	if (!newitem)
	{
		newitem = smalloc(sizeof(cAccessList));
		strcpy(newitem->nick, add->nick);
		addChanOp(chan, newitem);
	}

	newitem->uflags = FOUNDER;

	strcpy(chan->founder,newitem->nick);

	uSend(ChanServ, nick, GetReply(RPL_CHANTRANS, nick), chan->name,newitem->nick);
	sendToChanOps(getChanData(chan->name),"%s transferred founder status to %s.",from,newitem->nick);
	logDump(chanlog, "%s!%s@%s TRANSFER (%s) TO %s", nick->nick, nick->user, nick->host, chan->name, newitem->nick);
	add->chans++;

	if (founder_control !=2)
		nick->reg->chans--;
}

void CaddopCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	cAccessList *olditem, *newitem;
	RegChanList *chan;
	RegNickList *add;
	int access_level,a;
	int host_access_level;
	
	chan = getRegChanData(args[1]);
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
	
	host_access_level = getChanOp(chan,from);

	if((SOP > host_access_level) && (!isRoot(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_CHANAXTOOLOW, nick));
		return;
	}
	
	if(chan->ops >= OpLimit) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOMOREOPS, nick), chan->name);
		return;
	}

	add = getRegNickData(args[2]);
	if(!add || !(add->flags & NACTIVE)) {
		uSend(ChanServ, nick, GetReply(ERR_NICKNOTREGGED, nick), args[2]);
		return;
	}

	olditem = getChanOpData(chan, args[2]);
	if(olditem) {

		access_level = OPNONE;

		if (numargs < 4)
			access_level = AOP;

		else if (isalpha(args[3][0]))
		{
			if (!strcasecmp(args[3],"founder"))
				access_level = FOUNDER;
			else if (!strcasecmp(args[3],"cofounder"))
				access_level = MFOUNDER;
			else if (!strcasecmp(args[3],"sop"))
				access_level = SOP;
			else if (!strcasecmp(args[3],"msop"))
				access_level = MSOP;
			else if (!strcasecmp(args[3],"aop"))
				access_level = AOP;
			else if (!strcasecmp(args[3],"voice"))
				access_level = MAOP;
			else if (!strcasecmp(args[3],"hop"))
				access_level = HOP;
		} else
                        access_level = atoi(args[3]);

		if ((MAOP > access_level) || (FOUNDER < access_level))
		{
			uSend(ChanServ, nick, GetReply(ERR_UNKAXLEVEL, nick), args[4]);
			return;
		}	

		if(((access_level >= MFOUNDER && FOUNDER > host_access_level) || (access_level >= MSOP && MFOUNDER > host_access_level) || host_access_level < SOP) && (!isRoot(nick)))
		{
			uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
			return;
		}

		if (!strcasecmp(chan->founder,args[2]))
		{
			uSend(ChanServ, nick, GetReply(ERR_NOAXCGFOUNDER, nick), args[2], chan->name);
			return;
		}

		a=getChanOp(chan, from);

		if((olditem->uflags >= MFOUNDER && FOUNDER > a) || 
		  (olditem->uflags >= MSOP && MFOUNDER > a) ||
		  ((a < SOP) && !isRoot(nick) && strcasecmp(from,args[2]))) {
			uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
			return;
		}

		delChanOp(chan, olditem);

	        newitem = smalloc(sizeof(cAccessList));

        	newitem->uflags = access_level;
	        strcpy(newitem->nick, add->nick);
		newitem->uflags = access_level;

		uSend(ChanServ, nick, GetReply(RPL_CHANAXCHANGE, nick), newitem->nick, showchanoplevel(newitem->uflags));
		sendToChanOps(getChanData(chan->name), "%s changed access for op %s to %s", from, newitem->nick, showchanoplevel(newitem->uflags));
		addChanOp(chan, newitem);
		return;
	}

	access_level = OPNONE;

	if (numargs < 4)
		access_level = AOP;
	else
	{
		if (isalpha(args[3][0]))
		{
			if (!strcasecmp(args[3],"founder"))
				access_level = FOUNDER;
			else if (!strcasecmp(args[3],"cofounder"))
				access_level = MFOUNDER;
			else if (!strcasecmp(args[3],"sop"))
				access_level = SOP;
			else if (!strcasecmp(args[3],"msop"))
				access_level = MSOP;
			else if (!strcasecmp(args[3],"aop"))
				access_level = AOP;
			else if (!strcasecmp(args[3],"voice"))
				access_level = MAOP;
		} else
                        access_level = atoi(args[3]);
	}

	if ((MAOP > access_level) || (FOUNDER < access_level))
	{
		uSend(ChanServ, nick, GetReply(ERR_UNKAXLEVEL, nick), args[4]);
		return;
	}	

	if(((access_level >= MFOUNDER && FOUNDER > host_access_level) || (access_level >= MSOP && MFOUNDER > host_access_level) || host_access_level < SOP) && (!isRoot(nick)))
	{
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

	if (add->flags & NNOOP && !isRoot(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_NOOPON, nick), add->nick);
		return;
	}

	newitem = smalloc(sizeof(cAccessList));

	newitem->uflags = access_level;
	strcpy(newitem->nick, add->nick);

	uSend(ChanServ, nick, GetReply(RPL_CHANOPADDED, nick), chan->ops, newitem->nick, showchanoplevel(newitem->uflags));
	sendToChanOps(getChanData(chan->name), "%s added op %s as %s", from, newitem->nick, showchanoplevel(newitem->uflags));
	addChanOp(chan, newitem);
}

void CaddakCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	cAkickList *newitem;
	RegChanList *chan;
	int i;
	
	chan = getRegChanData(args[1]);
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
	
	if((MSOP > getChanOp(chan, from)) && (!isServop(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_CHANAXTOOLOW, nick));
		return;
	}
	
	if(chan->ops >= AkickLimit) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOMOREAKICK, nick), chan->name);
		return;
	}
	newitem = getChanAkick(chan, args[2]);
	if(newitem) {
		uSend(ChanServ, nick, GetReply(ERR_THISEXISTS, nick), args[2], "Akick");
		return;
	}

	if(!check_mask(args[2],isServop(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_INVALIDAKICK, nick), args[2]);
		return;
	}
	
	newitem = smalloc(sizeof(cAkickList));
	strcpy(newitem->mask, args[2]);
	sprintf(newitem->reason, "%s:", from);
	for(i=3;i<numargs && (strlen(newitem->reason) + strlen(args[i]) <= NICKLEN+51);i++) {
		strcat(newitem->reason, " ");
		strcat(newitem->reason, args[i]);
	}
	newitem->added = CTime;
	addChanAkick(chan, newitem);
	sendToChanOps(getChanData(chan->name), "%s added akick %s (%s)", from, newitem->mask, newitem->reason);
	uSend(ChanServ, nick, GetReply(RPL_CHANAKICKADDED, nick), newitem->mask, chan->name);
}
		
void CdelopCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	RegChanList *chan;
	cAccessList *delop;
	int a, x;

	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick));
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
	
	if((x = atoi(args[2]))) {
		for(delop=chan->firstOp;delop;delop=delop->next) {
			if(delop->index == x)
			  break;
		}
	}
	else
		delop = getChanOpData(chan, args[2]);
	
	if(!delop) {
		uSend(ChanServ, nick, GetReply(ERR_NOTINOPSLIST, nick), args[2], chan->name);
		return;
	}
	
	if (!strcasecmp(chan->founder,args[2]))
	{
		uSend(ChanServ, nick, GetReply(ERR_NOAXCGFOUNDER, nick), chan->founder, chan->name);
		return;
	}

	a=getChanOp(chan, from);
	if(((delop->uflags >= MFOUNDER && FOUNDER > a) || (delop->uflags >= MSOP && MFOUNDER > a) || a < SOP) && (!isRoot(nick)) && strcasecmp(from,args[2])) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	sendToChanOps(getChanData(chan->name), "%s deleted op %s [%s]", from, delop->nick, showchanoplevel(delop->uflags));
	uSend(ChanServ, nick, GetReply(RPL_CHANOPDELED, nick), delop->nick, chan->name);
	delChanOp(chan, delop);
}

void CdelakCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	RegChanList *chan;
	cAkickList *delak;
	int x;

	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	if((x = atoi(args[2]))) {
		for(delak=chan->firstAkick;delak;delak=delak->next) {
			if(delak->index == x)
			  break;
		}
	}
	else
		delak = getChanAkick(chan, args[2]);
	
	if(!delak) {
		uSend(ChanServ, nick, GetReply(ERR_AKICKNOTEXIST, nick), args[2], chan->name);
		return;
	}
	
	if((MSOP > getChanOp(chan, from)) && (!isServop(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	sendToChanOps(getChanData(chan->name), "%s deleted akick %s", from, delak->mask);
	uSend(ChanServ, nick, GetReply(RPL_CHANAKICKDELED, nick), delak->mask, chan->name);
	delChanAkick(chan, delak);
}

void ClistopCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	
	RegChanList *tmpchan;
	int levelsearch=0, i=0, x=0;
	char namesearch[NICKLEN] = "\0";
	cAccessList *tmp;
	
	if(numargs == 2) {
		levelsearch = 0;
		namesearch[0] = 0;
	}
	if(numargs == 3) {
		uSend(ChanServ, nick, GetReply(ERR_BADPARAMS, nick));
		return;
	}
	if(numargs == 4) {
		if(!strcasecmp(args[2], "-l"))
		  levelsearch = atoi(args[3]);
		else if(!strcasecmp(args[2], "-n"))
		  strcpy(namesearch, args[3]);
		else {
			uSend(ChanServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[2]);
			return;
		}
	}

	tmpchan = getRegChanData(args[1]);
	
	if((tmpchan == NULL)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

	if(!(getChanOp(tmpchan, from) > 0) && !isOper(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	uSend(ChanServ, nick, GetReply(RPL_CHANOPSLISTHEAD1, nick), tmpchan->name);
	uSend(ChanServ, nick, GetReply(RPL_CHANOPSLISTHEAD2, nick));

	for(tmp=tmpchan->firstOp;tmp;tmp=tmp->next,i++) {
		if(levelsearch) {
			if(tmp->uflags == levelsearch) {
				uSend(ChanServ, nick, GetReply(RPL_CHANOPSLISTDATA, nick), tmp->index, tmp->nick, showchanoplevel(tmp->uflags));
				x++;
			}
		}
		else if(namesearch[0]) {
			if(!match(tmp->nick, namesearch)) {
				uSend(ChanServ, nick, GetReply(RPL_CHANOPSLISTDATA, nick), tmp->index, tmp->nick, showchanoplevel(tmp->uflags));
				x++;
			}
		}
		else {
			uSend(ChanServ, nick, GetReply(RPL_CHANOPSLISTDATA, nick), tmp->index, tmp->nick, showchanoplevel(tmp->uflags));
			x++;
		}
	}

	if(namesearch[0] || levelsearch) {
		uSend(ChanServ, nick, GetReply(RPL_LISTSEARCHNUM, nick), x, i);
	}
	uSend(ChanServ, nick, GetReply(RPL_CHANOPSLISTEND, nick));
}

void ClistakCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	char aktime[30];
	RegChanList *tmpchan;
	int i=0, x=0;
	char namesearch[NICKLEN+USERLEN+HOSTLEN] = "\0";
	cAkickList *tmp;
	
	if(numargs == 2) {
		namesearch[0] = 0;
	}

	if(numargs == 3) {
		uSend(ChanServ, nick, GetReply(ERR_BADPARAMS, nick));
		return;
	}

	if(numargs == 4) {
		if(!strcasecmp(args[2], "-n"))
		  strcpy(namesearch, args[3]);
		else {
			uSend(ChanServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[2]);
			return;
		}
	}
	
	tmpchan = getRegChanData(args[1]);
	
	if(!tmpchan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}
	
	if(!getChanOp(tmpchan, from) && !isOper(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTHEAD1, nick), args[1]);
	uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTHEAD2, nick));
	uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTHEAD3, nick));

	for(tmp=tmpchan->firstAkick;tmp;tmp=tmp->next,i++) {
		if(namesearch[0]) {
			if(!match(tmp->mask, namesearch)) {
				strftime(aktime, 15, "%a %d/%m/%Y", gmtime(&tmp->added));
				uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTDATA1, nick), tmp->index, tmp->mask);
				uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTDATA2, nick), aktime, tmp->reason);
				x++;
			}
			uSend(ChanServ, nick, GetReply(RPL_LISTSEARCHNUM, nick), x, i);
		}
		else {
			strftime(aktime, 15, "%a %d/%m/%Y", gmtime(&tmp->added));
			uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTDATA1, nick), tmp->index, tmp->mask);
			uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTDATA2, nick), aktime, tmp->reason);
			x++;
		}
	}

	uSend(ChanServ, nick, GetReply(RPL_CHANAKLISTEND, nick));
}

void CdropCmd(UserList *nick, char **args, int numargs)
{
	RegChanList *tmp;
	ChanList *chan;
	char *from = nick->nick;
        int oper_drop;
	
	tmp = getRegChanData(args[1]);

	if(!tmp) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((tmp->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

        /* Stop users from getting rid of this stuff so easily*/

        if(((tmp->flags & CFORCES) || (tmp->flags & CCLOSED)) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
                return;
        }          
	
	if(!isFounder(tmp, nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

        /* DON'T DECREMENT OPERS CHAN COUNT */
        oper_drop = 0;
        if (!match(tmp->founder,from))
		oper_drop = 1;

	sendToChanOps(getChanData(tmp->name), "%s just dropped %s", from, args[1]);
	uSend(ChanServ, nick, GetReply(RPL_CHANDROP, nick), args[1]);

	chan = getChanData(args[1]);
	if(chan)
		chan->reg = NULL;
	logDump(chanlog, "%s!%s@%s DROP (%s)", nick->nick, nick->user, nick->host, tmp->name);
	delRegChan(tmp);
       
        if (!oper_drop)
        {
           /* DON'T LET NICKSERV GO NEGATIVE */
           if (nick->reg->chans > 0)
	      nick->reg->chans--;
        }
        return;
}

void CopCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	int a;
	RegChanList *tmp;
	ChanList *chan;
	cNickList *tmpnick;
        RegNickList *webbie;
	
	tmp = getRegChanData(args[1]);
	chan = getChanData(args[1]);

	if(!tmp) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((tmp->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
	
	if(numargs > 2)
		tmpnick = getChanUserData(chan, getNickData(args[2]));
	else
		tmpnick = getChanUserData(chan, nick);
	
	if(!tmpnick) {
		uSend(ChanServ, nick, GetReply(ERR_NICKNOTINCHAN, nick), (numargs == 3) ? args[2] : from, args[1]);
		return;
	}

	a = getChanOp(tmp, (numargs > 2) ? args[2] : from);
	
	if((!a || !getChanOp(tmp, from) || a < 3) && !isRoot(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	sendToChanOps(getChanData(tmp->name), "%s used \002OP\002 on %s", from, (numargs > 2) ? args[2] : from);
	if(a < AOP && !isRoot(nick)) {
		sSend(":%s MODE %s +v %s", ChanServ, args[1], (numargs > 2) ? args[2] : from);
		tmpnick->op |= CHANVOICE;
		return;
	}
	else {
		sSend(":%s MODE %s +o %s", ChanServ, args[1], (numargs > 2) ? args[2] : from);
		tmpnick->op |= CHANOP;

                /* Inform webbies they are opped now */

	        if(numargs > 2)
	                webbie = getRegNickData(args[2]);
	        else
	                webbie = getRegNickData(from);     

                if(!webbie) return;

                if (webbie->flags2 & NWEBTV) {
			uSend(ChanServ, nick, GetReply(RPL_NOWCHANOPERATOR, nick), args[1]);
                }
		return;
	}
}

void CdeopCmd(UserList *nick, char **args, int numargs)
{
	char *from = nick->nick;
	int a;
	RegChanList *tmp;
	ChanList *chan;
	cNickList *tmpnick;
	RegNickList *webbie;

	chan = getChanData(args[1]);
	tmp = getRegChanData(args[1]);

	if(!chan) {
                uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}
		
	if(!tmp) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((tmp->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
	
	if(numargs > 2)
		tmpnick = getChanUserData(chan, getNickData(args[2]));
	else
		tmpnick = getChanUserData(chan, nick);
	
	if(!tmpnick) {
		uSend(ChanServ, nick, GetReply(ERR_NICKNOTINCHAN, nick), (numargs == 3) ? args[2] : from, args[1]);
		return;
	}

	a = getChanOp(tmp, from);
	
	if(!a || !getChanOp(tmp, from) || a < 3) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	sendToChanOps(getChanData(tmp->name), "%s used \002DEOP\002 on %s", from, (numargs > 2) ? args[2] : from);
	sSend(":%s MODE %s -ov %s %s", ChanServ, args[1], (numargs == 3) ? args[2] : from, (numargs > 2) ? args[2] : from);
	tmpnick->op &= ~CHANOP;

        /* Inform webbies they are de-opped now */

        if(numargs > 2)
                webbie = getRegNickData(args[2]);
        else
                webbie = getRegNickData(from);       

        if(!(webbie)) return;

        if (webbie->flags2 & NWEBTV) {
		uSend(ChanServ, nick, GetReply(RPL_NOTCHANOPERATOR, nick), (numargs == 3) ? args[2] : from, args[1]);
        }
	return;                                                

}
	
void CclistCmd(UserList *nick, char **args, int numargs)
{
/* unused
	char *from = nick->nick; */
	ChanList *chan;
	
	if(!isOper(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	chan = getChanData(args[1]);

	if (chan) {
		cNickList *tmp;
		int totusers = 0;

		for (tmp=chan->firstUser;tmp;tmp=tmp->next) {
			totusers++;
			uSend(ChanServ, nick, GetReply(RPL_CHANCLISTDATA, nick), chan->name, tmp->person->nick, (tmp->op) ? " [OP]" : "");
		}
		uSend(ChanServ, nick, GetReply(RPL_NUMUSERSINCHAN, nick), totusers, chan->name);
	}
	else 
		uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
}


void CmodelockCmd(UserList *nick, char **args, int numargs) {
	int onargs = 3;
	int on=1;
	RegChanList *chan;
	ChanList *tmp;
	char *from = nick->nick;
	int i;
	char mode[20];

	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}
	
        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	if((MFOUNDER > getChanOp(chan, from)) && (!isRoot(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	chan->mlock = 0;
	for(i=0;i<strlen(args[2]);i++) {
		switch(args[2][i]) {
		 case '+':
			on = 1;
			break;
		 case '-':
			on = 0;
			break;
                 case 'R':
                        if(on) {
                                chan->mlock |= PM_R;
                                chan->mlock &= ~MM_R;
                        }
                        else {
                                chan->mlock |= MM_R;
                                chan->mlock &= ~PM_R;
                       }
                       break;
		 case 'c':
			if(on) {
				chan->mlock |= PM_C;
				chan->mlock &= ~MM_C;
			}
			else {
				chan->mlock |= MM_C;
				chan->mlock &= ~PM_C;
			}
			break;
		 case 'i':
			if(on) {
				chan->mlock |= PM_I;
				chan->mlock &= ~MM_I;
			}
			else {
				chan->mlock |= MM_I;
				chan->mlock &= ~PM_I;
			}
			break;
		 case 'k':
			if(on) {
				if(numargs<(onargs+1))
					uSend(ChanServ, nick, GetReply(ERR_MISSINGKEY, nick));
				else {
					strcpy(chan->key, args[onargs]);
					onargs++;
					chan->mlock |= PM_K;
					chan->mlock &= ~MM_K;
				}
			}
			else {
				chan->mlock |= MM_K;
				chan->mlock &= ~PM_K;
			}
			break;
		 case 'l':
			if(on) {
				if(numargs<(onargs+1))
					uSend(ChanServ, nick, GetReply(ERR_MISSINGLIMIT, nick));
				else {
					chan->limit = atoi(args[onargs]);
					onargs++;
					chan->mlock |= PM_L;
					chan->mlock &= ~MM_L;
				}
			}
			else {
				chan->mlock |= MM_L;
				chan->mlock &= ~PM_L;
			}
			break;
		 case 'm':
			if(on) {
				chan->mlock |= PM_M;
				chan->mlock &= ~MM_M;
			}
			else {
				chan->mlock |= MM_M;
				chan->mlock &= ~PM_M;
			}
			break;
		 case 'n':
			if(on) {
				chan->mlock |= PM_N;
				chan->mlock &= ~MM_N;
			}
			else {
				chan->mlock |= MM_N;
				chan->mlock &= ~PM_N;
			}
			break;
		 case 'p':
			if(on) {
				chan->mlock |= PM_P;
				chan->mlock &= ~MM_P;
			}
			else {
				chan->mlock |= MM_P;
				chan->mlock &= ~PM_P;
			}
			break;
		 case 's':
			if(on) {
				chan->mlock |= PM_S;
				chan->mlock &= ~MM_S;
			}
 
                        else {
    
			    tmp = getChanData(args[1]);               
                                
			    if ((tmp && isForced(tmp->name)) || 
				((tmp && tmp->reg && tmp->reg->flags&CFORCES) && !(tmp->reg && tmp->reg->flags&CFEXEMPT)))  {
				uSend(ChanServ, nick, GetReply(RPL_AUTOSET_S, nick));
				return;
			    }

		            else {
					chan->mlock |= MM_S;
					chan->mlock &= ~PM_S;
			     }	
			}
			break;
		 case 't':
			if(on) {
				chan->mlock |= PM_T;
				chan->mlock &= ~MM_T;
			}
			else {
				chan->mlock |= MM_T;
				chan->mlock &= ~PM_T;
			}
			break;
		}
	}
	
	makeModeLockStr(chan, mode);
	sendToChanOps(getChanData(chan->name), "%s set modelock (%s)", from, mode);
	doRegChanMode(chan);
	uSend(ChanServ, nick, GetReply(RPL_MLOCKSETTO, nick), chan->name, mode);
}


void CrestrictCmd(UserList *nick, char **args, int numargs) {
	RegChanList *chan;
	char *from = nick->nick;
	int reslevel;
	
	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	if((MFOUNDER > getChanOp(chan, from)) && (!isRoot(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	reslevel = atoi(args[2]);
	if((reslevel > getChanOp(chan, from)) && (!isRoot(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	chan->restrictlevel = reslevel;
	
	sendToChanOps(getChanData(chan->name), "%s has set restriction to %i+ levels", from, reslevel);
	uSend(ChanServ, nick, GetReply(RPL_RESTRICTSETTO, nick), chan->name, reslevel);
}

void CtopiclockCmd(UserList *nick, char **args, int numargs) {
	RegChanList *chan;
	char *from = nick->nick;
	int reslevel;
	
	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}
	
	if((MFOUNDER > getChanOp(chan, from)) && (!isRoot(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

        if((chan->tlocklevel > getChanOp(chan,from)) && (!isRoot(nick))) {	
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

	reslevel = atoi(args[2]);
	if((reslevel > getChanOp(chan, from)) && (!isRoot(nick))) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	chan->tlocklevel = reslevel;
	
	sendToChanOps(getChanData(chan->name), "%s has set topic restriction to %i+ levels", from, reslevel);
	uSend(ChanServ, nick, GetReply(RPL_TOPICLOCKSETTO, nick), chan->name, reslevel);
}

void CsetCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	RegChanList *chan;
	ChanList *tmp;
		
	chan = getRegChanData(args[1]);
	tmp = getChanData(args[1]);

	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);		
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
	
	if(MFOUNDER > getChanOp(chan, from) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

	if(!strcasecmp(args[2], "NFPRIV") && isServop(nick)) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CFEXEMPT;
				chan->flags &= ~CFORCES;
				sSend(":%s GLOBOPS :%s made %s exempt from +s forcing", ChanServ, from, chan->name);
			}
			else if(!strcasecmp(args[3],"OFF")) {
				chan->flags&= ~CFEXEMPT;
				sSend(":%s GLOBOPS :%s removed %s from +s force exemption", ChanServ, from, chan->name);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NFPRIV", chan->name, (chan->flags&CFEXEMPT) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "HOLD") && isServop(nick)) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CHOLD;
				sSend(":%s GLOBOPS :%s made %s held.", ChanServ, from, chan->name);
			}
			else if(!strcasecmp(args[3],"OFF")) {
				chan->flags&= ~CHOLD;
				sSend(":%s GLOBOPS :%s removed hold from %s.", ChanServ, from, chan->name);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "HELD", chan->name, (chan->flags&CHOLD) ? "ON" : "OFF");
	}
	else if (!strcasecmp(args[2], "FORCES") && isServop(nick)) {
		if (numargs>3) {
			if (!strcasecmp(args[3], "ON")) {
				chan->flags |= CFORCES;
				chan->flags &= ~CFEXEMPT;
				sSend(":%s MODE %s +s", ChanServ, chan->name);
				sSend(":%s GLOBOPS :%s forces %s as +s", ChanServ, from, chan->name);
			}
			else if(!strcasecmp(args[3],"OFF")) {
				chan->flags &= ~CFORCES;
				sSend(":%s GLOBOPS :%s removed +s force from %s", ChanServ, from, chan->name);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "FORCES", chan->name, (chan->flags&CFORCES) ? "ON" : "OFF");
	}
	else if (!strcasecmp(args[2], "FREEZE") && isServop(nick)) {
		if (numargs>3) {
			if (!strcasecmp(args[3], "ON")) {
				chan->flags |= CFREEZE;
                                CSmassdeop(chan->name);
				sSend(":%s GLOBOPS :%s freezes %s", ChanServ, from, chan->name);
                                logDump(chanlog, "%s FREEZES %s",from,chan->name);
			}
			else if(!strcasecmp(args[3],"OFF")) {
				chan->flags &= ~CFREEZE;
				sSend(":%s GLOBOPS :%s defrosted %s", ChanServ, from, chan->name);
                                logDump(chanlog, "%s UNFREEZES %s",from,chan->name);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), chan->name, (chan->flags&CFREEZE) ? "FROZEN" : "UNFROZEN");
	}

	else if(!strcasecmp(args[2], "PRIVATE")) {
		if(numargs>3) {
			if(!strcasecmp(args[3],"ON")) {
				chan->flags|=CPRIVATE;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "PRIVATE", chan->name, "ON");
				sendToChanOps(tmp,"%s set private mode on", from);
			}
			else if(!strcasecmp(args[3],"OFF")) {
				chan->flags&= ~CPRIVATE;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "PRIVATE", chan->name, "OFF");
				sendToChanOps(tmp,"%s set private mode off",from);
			}
		}
		else 
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "PRIVATE", chan->name, (chan->flags&CPRIVATE) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "OPGUARD")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= COPGUARD;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "OPGUARD", chan->name, "ON");
				sendToChanOps(tmp, "%s set op guard on", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~COPGUARD;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "OPGUARD", chan->name, "OFF");
				sendToChanOps(tmp, "%s set op guard off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "OPGUARD", chan->name, (chan->flags & COPGUARD) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "NOAOL")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CNOAOL;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOAOL", chan->name, "ON");
				sendToChanOps(tmp, "%s set AOL guard on", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CNOAOL;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOAOL", chan->name, "OFF");
				sendToChanOps(tmp, "%s set AOL guard off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOAOL", chan->name, (chan->flags & CNOAOL) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "NOSCAN")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CSCAN;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOSCAN", chan->name, "ON");
				sendToChanOps(tmp, "%s set SCAN guard on", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CSCAN;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOSCAN", chan->name, "OFF");
				sendToChanOps(tmp, "%s set SCAN guard off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOSCAN", chan->name, (chan->flags & CSCAN) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "KEEPTOPIC")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CKTOPIC;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "KEEPTOPIC", chan->name, "ON");
				sendToChanOps(tmp, "%s set topic holding on", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CKTOPIC;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "KEEPTOPIC", chan->name, "OFF");
				sendToChanOps(tmp, "%s set topic holding off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "KEEPTOPIC", chan->name, (chan->flags & CKTOPIC) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "NAOP")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CCLASSIC;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NAOP", chan->name, "ON");
				sendToChanOps(tmp, "%s set CS AutoOp OFF", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CCLASSIC;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NAOP", chan->name, "OFF");
				sendToChanOps(tmp, "%s set CS AutoOp ON", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NAOP", chan->name, (chan->flags & CCLASSIC) ? "OFF" : "ON");
	}
	else if(!strcasecmp(args[2], "IDENT")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CIDENT;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "IDENT", chan->name, "ON");
				sendToChanOps(tmp, "%s set ID requirement on", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CIDENT;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "IDENT", chan->name, "OFF");
				sendToChanOps(tmp, "%s set ID requirement off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "IDENT", chan->name, (chan->flags & CIDENT) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "PROTOP")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				chan->flags |= CPROTOP;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "PROTOP", chan->name, "ON");
				sendToChanOps(tmp, "%s set Op protection on", from);
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CPROTOP;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "PROTOP", chan->name, "OFF");
				sendToChanOps(tmp, "%s set op protection off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "PROTOP", chan->name, (chan->flags & CPROTOP) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2], "QUIET")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "QUIET", chan->name, "ON");
				sendToChanOps(tmp, "%s set Quiet changes mode on, this is the last you'll hear from me", from);
				chan->flags |= CQUIET;
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CQUIET;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "QUIET", chan->name, "OFF");
				sendToChanOps(tmp, "%s set quiet changes mode off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "QUIET", chan->name, (chan->flags & CQUIET) ? "ON" : "OFF");
	}
	else if(!strcasecmp(args[2],"FOUNDER"))
      	CfounderCmd(nick, args, numargs);
	else if(!match(args[2], "desc*")) {
		int i;
		chan->desc[0] = '\0';
		if(numargs >= 4) {
			if(strlen(args[3]) <= 99) {
				strcpy(chan->desc, args[3]);
				for(i=4;i<numargs && strlen(chan->desc) + strlen(args[i]) <= 99;i++) {
					strcat(chan->desc, " ");
					strcat(chan->desc, args[i]);
				}
			}
		}
		if (chan->desc[0])
			uSend(ChanServ, nick, GetReply(RPL_CHANSETDESC, nick), chan->name, chan->desc);
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETDESCCLEAR, nick), chan->name);
	}
	else if(!strcasecmp(args[2], "NOTICE")) {
		if(numargs>3) {
			if(!strcasecmp(args[3], "ON")) {
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOTICE", chan->name, "ON");
				sendToChanOps(tmp, "%s set Notice On Join mode ON", from);
				chan->flags |= CNOTICE;
			}
			else if(!strcasecmp(args[3], "OFF")) {
				chan->flags &= ~CNOTICE;
				uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOTICE", chan->name, "OFF");
				sendToChanOps(tmp, "%s set Notice On Join mode off", from);
			}
		}
		else
			uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "NOTICE", chan->name, (chan->flags & CQUIET) ? "ON" : "OFF");
	}
	else
		uSend(ChanServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[2]);
}

void expireChans(void)
{
	time_t timestart, timeend;
	RegChanList *tmp, *next;
#ifdef BACKUPDATABASE
	char backup[30];
	char cmd[500];
#endif
	int i=0;

	mostchans = 0;
	timestart = time(NULL);
#ifdef BACKUPDATABASE
	strftime(backup, 30, "chanserv.db%d%m%Y", localtime(&timestart));
	sprintf(cmd, "cp %s/chanserv.db %s/%s", 
			DATABASEDIR, DATABASEDIR, backup);
	system(cmd);
#endif
	for(tmp=firstRegChan;tmp;tmp=next) {
		next=tmp->next;

		if (tmp->timestamp > timestart) /* NEGATIVE TIME? */
			tmp->timestamp = timestart;  /* FIX THIS NOW */

		if (((timestart - tmp->timestamp) >= CHANDROPTIME) && !(tmp->flags&CHOLD))
		{
			ChanList *ick;
			ick = getChanData(tmp->name);
			if(ick)
				ick->reg = NULL;
			logDump(chanlog, "%s EXPIRE (AGE)",tmp->name);
			delRegChan(tmp);
			i++;
		}
		mostchans++;
	}
	timeend = time(NULL);
#ifdef SHOWSYNCH
	sSend(":%s GLOBOPS :ChanServ EXPIRE(%i/%lu) %lu seconds", ChanServ, i, mostchans, (timeend - timestart));
#endif
	timer(2.5*3600, doTfunc, "dotfunc", "1");
}

void CsaveCmd(UserList *tmp, char **args, int numargs) {
	char *from = tmp->nick;
	
	sSend(":%s GLOBOPS :Synchronising \002ChanServ database\002 to disk at the request of \002%s\002.", ChanServ, from);
	saveChanData(firstRegChan);
}

void CinviteCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	RegChanList *chan;
	
	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED , nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }
										  
	if((1 > getChanOp(chan, from)) && (!isOper(nick)))
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
	else 
		sSend(":%s INVITE %s :%s", ChanServ, from, chan->name);
}

void CunbanCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	RegChanList *chan;
	ChanList *tmp;
	cBanList *tmpban;
	int a, domodes = 0;
	char mode[512] = "";
	char userhost[NICKLEN + USERLEN + HOSTLEN];
	char cloakedhost[NICKLEN + USERLEN + HOSTLEN];
	
	if(!(chan = getRegChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

	if(!(tmp = getChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}

        if((chan->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	a = getChanOp(chan, from);
		
	if(0>=a) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}
	
	if(!strcasecmp(args[2], "me")) {
		sprintf(userhost, "%s!%s@%s", nick->nick, nick->user, nick->host);
		sprintf(cloakedhost, "%s!%s@%s", nick->nick, nick->user, maskme(nick->host));
		for(tmpban=tmp->firstBan;tmpban;tmpban=tmpban->next) {
			if((!match(userhost, tmpban->ban)) || (!match(cloakedhost, tmpban->ban ))) {
				strcat(mode, tmpban->ban);
				strcat(mode, " ");
				domodes++;
				delChanBan(tmp, tmpban);
				if(domodes == 6 || !tmpban->next) {
					uSend(ChanServ, nick, GetReply(RPL_REMCHANBANS, nick), mode);
					sSend(":%s MODE %s -bbbbbb %s", ChanServ, args[1], mode);
					domodes = 0;
					mode[0] = 0;
				}
			}
		}
	}
	else if(!strcasecmp(args[2], "all")) {
		if(a < MSOP && !isOper(nick))
			uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		else {
			for(tmpban=tmp->firstBan;tmpban;tmpban=tmpban->next) {
				strcat(mode, tmpban->ban);
				strcat(mode, " ");
				domodes++;
				delChanBan(tmp, tmpban);
				if(domodes == 6 || !tmpban->next) {
					uSend(ChanServ, nick, GetReply(RPL_REMCHANBANS, nick), mode);
					sSend(":%s MODE %s -bbbbbb %s", ChanServ, args[1], mode);
					domodes = 0;
					mode[0] = 0;
				}
			}
		}
	}
	else {
		if(a < MSOP && !isOper(nick))
			uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		else {
			for(tmpban=tmp->firstBan;tmpban;tmpban=tmpban->next) {
				if(!match(tmpban->ban,args[2])) {
					strcat(mode, tmpban->ban);
					strcat(mode, " ");
					domodes++;
					delChanBan(tmp, tmpban);
					if(domodes == 6 || !tmpban->next) {
						uSend(ChanServ, nick, GetReply(RPL_REMCHANBANS, nick), mode);
						sSend(":%s MODE %s -bbbbbb %s", ChanServ, args[1], mode);
						domodes = 0;
						mode[0] = 0;
					}
				}
			}
		}
	}
}

/* defines for flags, not really important but I like them */
#define CLF_FOUNDER 0x1
#define CLF_CHNAME 0x2
#define CLF_NUMOPG 0x4
#define CLF_NUMOPL 0x8
#define CLF_NUMAKG 0x10
#define CLF_NUMAKL 0x20

void ClistCmd(UserList *tmp, char **args, int numargs) {
/* unused
	char *from = tmp->nick; */
	char mask[CHANLEN];
	char fmask[NICKLEN];
	int sno,sna;
	int i,x,m;
	int cf=0;
	char *s;
	int onargs=2;
	RegChanList *reg;
	sno=sna=0;
	i=x=0;
	
	s=args[1];
	while(*s && args[onargs]) {
		switch(*s) {
		 case 'm':
			cf|=CLF_CHNAME;
			strncpy(mask,args[onargs],CHANLEN);
			onargs++;
			break;
		 case 'f':
			cf|=CLF_FOUNDER;
			strncpy(fmask,args[onargs],NICKLEN);
			onargs++;
			break;
		 case 'o':
			s++;
			if(*s=='<')
			  cf|=CLF_NUMOPL;
			else if(*s=='>')
			  cf|=CLF_NUMOPG;
			else
			  s--;
			sno=atoi(args[onargs]);
			onargs++;
			break;
		 case 'a':
			s++;
			if(*s=='<')
			  cf|=CLF_NUMAKL;
			else if(*s=='>')
			  cf|=CLF_NUMAKG;
			else
			  s--;
			sna=atoi(args[onargs]);
			onargs++;
			break;
		}
		s++;
	}
	for(reg=firstRegChan;reg;reg=reg->next) {
		m=0;
		if(cf&CLF_FOUNDER) {
			if(!match(reg->founder,fmask))
			  m=1;
			else
			  m=0;
		}
		if(cf&CLF_CHNAME) {
			if(!match(reg->name,mask))
			  m=1;
			else
			  m=0;
		}
		if(cf&CLF_NUMOPG) {
			if(reg->ops>sno)
			  m=1;
			else
			  m=0;
		}
		if(cf&CLF_NUMOPL) {
			if(reg->ops<sno)
			  m=1;
			else
			  m=0;
		}
		if(cf&CLF_NUMAKG) {
			if(reg->akicks>sna)
			  m=1;
			else
			  m=0;
		}
		if(cf&CLF_NUMAKL) {
			if(reg->akicks<sna)
			  m=1;
			else
			  m=0;
		}
		if(m) {
			i++;
			if(!(reg->flags&CPRIVATE) || isOper(tmp))
				uSend(ChanServ, tmp, GetReply(RPL_CHANLIST, tmp), reg->name, reg->founder, reg->ops, reg->akicks);
		}
		x++;
	}
	uSend(ChanServ, tmp, GetReply(RPL_LISTSEARCHNUM, tmp), i, x);
}

void CdeleteCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	ChanList *unf;
	RegChanList *moo;
	
	moo = getRegChanData(args[1]);
	if(!moo) {
		uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}
	nick->reg->chans--;
	unf = getChanData(args[1]);
	if(unf)
		unf->reg = NULL;
	uSend(ChanServ, nick, GetReply(RPL_CHANSETSTATUS, nick), "REGISTRATION", moo->name, "DROPPED");
	delRegChan(moo);
	mostchans--;
	logDump(chanlog,"%s!%s@%s deleted channel %s", from, nick->user, nick->host, args[1]);
	
	sSend(":%s GLOBOPS :%s deleted channel %s", ChanServ, from, args[1]);
}
	


void syncChanData(time_t next) {
	nextCsync = next;
	saveChanData(firstRegChan);
}

void doRegChanMode(RegChanList *chan) {
	char mode[80];
	char fmt[80];
	strcpy(fmt,":%s MODE %s r%s");
	if(chan->mlock&PM_L)
	  sprintf(&fmt[strlen(fmt)], " %lu", chan->limit);
	strcat(fmt," %s");
	
	makeModeLockStr(chan,mode);
	sSend(fmt, ChanServ, chan->name, mode, (chan->mlock&PM_K) ? chan->key : "");
}

void CmassdeopCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	cNickList *x;
	ChanList *c;
	RegChanList *rc;
	int l,ml,w;
	
	char deoplist[512];
	char devclist[512];
	int dop,dvc;
	
	if(!(rc=getRegChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

        if(!(c = getChanData(args[1]))) {
                uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
                return;
        }

        if((rc->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
        }
	
	l=getChanOp(rc,from);
	
	w=ml=0;
	for(x=c->firstUser;x;x=x->next) {
		w=getChanOp(rc,x->person->nick);
		if(w>ml) {
			ml=w;
		}
	}
	if((ml>l || !l) && !isOper(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		sSend(":%s NOTICE %s :%s tried to massdeop, but failed.", ChanServ, c->name, from);
		return;
	}
	else {
		bzero(deoplist,512);
		bzero(devclist,512);
		dop=dvc=0;
		for(x=c->firstUser;x;x=x->next) {
			if(dop==6) {
				sSend(":%s MODE %s -oooooo %s", ChanServ, c->name,deoplist);
				uSend(ChanServ, nick, GetReply(RPL_CHANMASSDEOP, nick), deoplist);
				dop=0;
				bzero(deoplist,512);
			}
			if(dvc==6) {
				sSend(":%s MODE %s -vvvvvv %s", ChanServ, c->name,devclist);
				uSend(ChanServ, nick, GetReply(RPL_CHANMASSDEVOICE, nick), devclist);
				dvc=0;
				bzero(devclist,512);
			}
			if(x->op&CHANOP) {
				dop++;
				strcat(deoplist, " ");
				strcat(deoplist, x->person->nick);
				x->op&= ~CHANOP;
			}
			else if(x->op&CHANVOICE) {
				dvc++;
				strcat(devclist," ");
				strcat(devclist,x->person->nick);
				x->op &= ~CHANVOICE;
			}
		}
		if(dop) {
			sSend(":%s MODE %s -oooooo %s", ChanServ, c->name,deoplist);
			uSend(ChanServ, nick, GetReply(RPL_CHANMASSDEOP, nick), deoplist);
			dop=0;
			bzero(deoplist,512);
		}
		if(dvc) {
			sSend(":%s MODE %s -vvvvvv %s", ChanServ, c->name,devclist);
			uSend(ChanServ, nick, GetReply(RPL_CHANMASSDEVOICE, nick), devclist);
			dvc=0;
			bzero(devclist,512);
		}
	}
}


void CmasskickCmd(UserList *nick, char **args, int numargs) {
	char *from=nick->nick;
	cNickList *x, *nextnick = NULL;
	ChanList *c;
	RegChanList *rc;
	int ml,l,w;
	int kicked;
	
	if(!(rc=getRegChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}

	if(!(c = getChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}
	
        if((rc->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	l=getChanOp(rc,from);
	
	w=ml=0;
	for(x=c->firstUser;x;x=x->next) {
		w=getChanOp(rc,x->person->nick);
		if(w>ml) {
			ml=w;
		}
	}
	if((ml>l || !l) && !isOper(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		sendToChanOps(c, "%s tried to masskick, but failed.", from);
		return;
	}
	else {
		kicked=0;
		
		createGhostChannel(c->name);
		timer(10,deleteGhostChannel, "deleteGhostChannel", c->name);
		for(x=c->firstUser;x;x=nextnick) {

			nextnick = x->next;

			if(x->person != nick && !isRoot(x->person)) {
				kicked++;
				sSend(":%s KICK %s %s :Masskick from \002%s\002", ChanServ, c->name, x->person->nick, from);
				remUserFromChan(x->person,c->name);
			}
		}
		
		uSend(ChanServ, nick, GetReply(RPL_CHANMASSKICK, nick), kicked, args[1]);
	}
}

void CmodeclearCmd(UserList *nick, char **args, int numargs) {
	char *from=nick->nick;
	cNickList *x;
	ChanList *c;
	RegChanList *rc;
	int ml,l,w;
	
	if(!(rc=getRegChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[1]);
		return;
	}
	
	if(!(c=getChanData(args[1]))) {
		uSend(ChanServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}

	if((rc->flags & CFREEZE) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_CHANFROZEN, nick));
		return;
        }

	l=getChanOp(rc,from);

	w=ml=0;
	for(x=c->firstUser;x;x=x->next) {
		w=getChanOp(rc,x->person->nick);
		if(w>ml) {
			ml=w;
		}
	}

        /* Fix users removing force +s with modeclear */

        if((rc->flags & CFORCES) && !isServop(nick)) {
		uSend(ChanServ, nick, GetReply(RPL_AUTOSET_S, nick));
               return;
        }

	if((ml>l || !l) && !isOper(nick)) {
		uSend(ChanServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		sSend(":%s NOTICE %s :%s tried to modeclear, but failed.", ChanServ, c->name, from);
		return;
	}
	else {
		sSend(":%s MODE %s -iklmnpstcR cowsRleet", ChanServ, c->name);
		uSend(ChanServ, nick, GetReply(RPL_CHANMODECLEAR, nick), c->name);
	}
}


ForceList *firstForce=NULL;
ForceList *lastForce=NULL;

void addForce(char *mask) {
	ForceList *nf;
	if(!mask)
	  return;
	
	nf=smalloc(sizeof(ForceList));
	strncpy(nf->mask,mask,CHANLEN);
	
	if(!firstForce) {
		firstForce=nf;
		nf->prev=NULL;
	}
	else {
		nf->prev = lastForce;
		lastForce->next = nf;
	}
	nf->next=NULL;
	lastForce=nf;
}

int isForced(char *mask) {
	ForceList *f;
	
	for(f=firstForce;f;f=f->next) {
		if(!match(mask,f->mask))
		  return 1;
	}
	return 0;
}

BanChannelList *firstBanChannel=NULL;
BanChannelList *lastBanChannel=NULL;

void addBanChannel(char *mask) {
	BanChannelList *newban;
	if(!mask)
	  return;
	
	newban=smalloc(sizeof(BanChannelList));
	strncpy(newban->mask,mask,CHANLEN);
	
	if(!firstBanChannel) {
		firstBanChannel=newban;
		newban->prev=NULL;
	}
	else {
		newban->prev = lastBanChannel;
		lastBanChannel->next = newban;
	}
	newban->next=NULL;
	lastBanChannel=newban;
}

int isBanChannel(char *mask) {
	BanChannelList *tmp;
	
	for(tmp=firstBanChannel;tmp;tmp=tmp->next) {
		if(!match(mask,tmp->mask))
		  return 1;
	}
	return 0;
}

void CSmassdeop(char *channame) {
	cNickList *x;
	ChanList *c;
	RegChanList *rc;
	int ml,w;
	
	char deoplist[512];
	char devclist[512];
	int dop,dvc;
	
        if(!(c=getChanData(channame))) {
		return; 
        }

	rc=getRegChanData(channame);
	
	w=ml=0;
	for(x=c->firstUser;x;x=x->next) {
		w=getChanOp(rc,x->person->nick);
		if(w>ml) {
			ml=w;
		}
	}
        {
		bzero(deoplist,512);
		bzero(devclist,512);
		dop=dvc=0;
		for(x=c->firstUser;x;x=x->next) {
			if(dop==6) {
				sSend(":%s MODE %s -oooooo %s", ChanServ, c->name,deoplist);
				dop=0;
				bzero(deoplist,512);
			}
			if(dvc==6) {
				sSend(":%s MODE %s -vvvvvv %s", ChanServ, c->name,devclist);
				dvc=0;
				bzero(devclist,512);
			}
			if(x->op&CHANOP) {
				dop++;
				strcat(deoplist, " ");
				strcat(deoplist, x->person->nick);
				x->op&= ~CHANOP;
			}
			else if(x->op&CHANVOICE) {
				dvc++;
				strcat(devclist," ");
				strcat(devclist,x->person->nick);
				x->op &= ~CHANVOICE;
			}
		}
		if(dop) {
			sSend(":%s MODE %s -oooooo %s", ChanServ, c->name,deoplist);
			dop=0;
			bzero(deoplist,512);
		}
		if(dvc) {
			sSend(":%s MODE %s -vvvvvv %s", ChanServ, c->name,devclist);
			dvc=0;
			bzero(devclist,512);
		}
	}
}

/* This will parse SJOINS and convert them to regular joins -gz */

void ParseSJOIN(char *channel, char *string) {

	 int i,c;
	 char *nick[256];
         char stringlist[256][NICKLEN*2];
	 char buffer[1024];
	 UserList *tmp;

	 for(i=0;i<256;i++)
         	nick[i] = stringlist[i];

	 c = breakString(255, string, nick, " ");

	 for(i=0;i<c;i++) {

		if(nick[i][0] == ':') {
			nick[i]++;
		}
		if(nick[i][0] == '@') {
	 
			nick[i]++;
			if(nick[i][0] == '+')
				nick[i]++;
			tmp=getNickData(nick[i]);
			if(!tmp) {
				sSend(":%s KILL %s :%s!%s (Desynch, type: [%s])", NickServ, nick[i], services[1].host, NickServ, "SJOIN");
				continue;
			}
			addUserToChanSJ(tmp,channel,2);
			bzero(buffer, 1024);
			printf(buffer, ":%s MODE %s +o %s", myname, channel,nick[i]);
			parseLine(buffer);
		}
		else if(nick[i][0] == '+') {
			nick[i]++;
			if(!tmp) {
				sSend(":%s KILL %s :%s!%s (Desynch, type: [%s])", NickServ, nick[i], services[1].host, NickServ, "SJOIN");
				continue;
			}
			tmp=getNickData(nick[i]);
			addUserToChanSJ(tmp,channel,1);
			bzero(buffer, 1024);
			sprintf(buffer, ":%s MODE %s +v %s", myname, channel, nick[i]);
			parseLine(buffer);
		}
		
		else {
			tmp=getNickData(nick[i]);
			if(!tmp) {
				sSend(":%s KILL %s :%s!%s (Desynch, type: [%s])", NickServ, nick[i], services[1].host, NickServ, "SJOIN");
				continue;
			}
			addUserToChanSJ(tmp,channel,0);	
		}
	 }
	return;
}

/* add a user to a channel, fun...really */
void addUserToChanSJ(UserList *nick, char *channel, int status)
{
	char chan[CHANLEN];
	int a=0;
	ChanList *tmp;
	cNickList *person;
        RegNickList *aol;
	

	if (nick == NULL || channel == NULL)
	  {
	    sSend("GLOBOPS :addUserToChanSJ() called with null arguments! [%p %p %d]",
		  nick, channel, status);
	    return;
	  }

	if(*channel == ':')
		channel++;

	while(*channel) {
		bzero(chan, CHANLEN);
		a=0;
		while ((*channel) && (*channel != ',')) {
			chan[a] = *channel;
			a++;
			channel++;
		}
		chan[a] = 0;
		if(*channel)
			channel++;
		  
		if(chan[0] == '0')
			remFromAllChans(nick);
		else if(chan[0] != '+') {
			tmp = getChanData(chan);

			if((!tmp) || (!tmp->firstUser)) {
				/* FIRST USER IN CHANNEL */

				/* INSERT NEW CHANNEL INTO INTERNAL LIST */
				if(!tmp) {
					tmp = smalloc(sizeof(ChanList));
					initChanData(tmp);
					strncpy(tmp->name,chan, CHANLEN);
					addChan(tmp);
				}

				person = smalloc(sizeof(cNickList));
				person->person = nick;
				person->op = 0;
					 
				addChanUser(tmp, person);

				if (isBanChannel(tmp->name)) {
					if (!isServop(nick)) {
						createGhostChannel(tmp->name);
						timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
						sSend(":%s MODE %s +b *!*@*", ChanServ, chan);
						sSend(":%s KICK %s %s :Channel is not permitted on this network.", ChanServ, chan, nick->nick);
						remUserFromChan(nick, chan);
						return;
					}
				}

#ifdef CDEBUG
				sSend(":%s GNOTICE :Created channel %s, added %s to it", myname, chan, nick->nick);
#endif
					 
				tmp->reg = getRegChanData(chan);

				/* new closed channels check -GZ */

				if(tmp->reg && tmp->reg->flags & CCLOSED) {
                                                if (!isServop(nick)) {
                                                        createGhostChannel(tmp->name);
                                                        timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                                                        sSend(":%s MODE %s +b *!*@*", ChanServ, chan);
                                                        sSend(":%s KICK %s %s :This channel has been \002CLOSED\002 by the " NETWORKNAME " administration.", ChanServ, chan, nick->nick);
                                                        remUserFromChan(nick, chan);
                                                        return;
                                                }
				}

				if((isForced(tmp->name) || (tmp->reg && tmp->reg->flags&CFORCES)) && !(tmp->reg && tmp->reg->flags&CFEXEMPT)) {
					sSend(":%s MODE %s +s", ChanServ, tmp->name);
					if(tmp->reg)
						tmp->reg->flags|=CPRIVATE;
				}
				/* IF THE CHANNEL IS REGISTERED -- CHECK OUT THE REGISTRATION DATA */

				if(tmp->reg) {
					char mode[20];
                                               
					/* OPS/BANS LIST */
					a=getChanOp(tmp->reg, nick->nick);

                                    /* Thou shall not akick servops */

                                        if(!isServop(nick)) {
					 if(a == -1) {  /* IS AKICKED? */
						cAkickList *blah;
						char userhost[NICKLEN + USERLEN + HOSTLEN];
						
						sprintf(userhost, "%s!%s@%s", nick->nick, nick->user, nick->host);
						for(blah = tmp->reg->firstAkick;blah;blah=blah->next) {
							if(!match(userhost, blah->mask)) {
								cBanList *newban = smalloc(sizeof(cBanList));
								sSend(":%s MODE %s +b %s", ChanServ, chan, blah->mask);
								strcpy(newban->ban, blah->mask);
								addChanBan(tmp, newban);
								break;
							}
						}
						createGhostChannel(tmp->name);
						timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
						sSend(":%s KICK %s %s :You have been permanently banned from this channel.", ChanServ, chan, nick->nick);
						remUserFromChan(nick, chan);
						return;
					 }
                                        }
                                        
                                         /* Temporary AOL protection */

                                        if(tmp->reg->flags & CNOAOL) {
                                             if(!match(nick->host, "*.aol.com")) {
                                                 aol = getRegNickData(nick->nick);
                                                 if(!aol) {
							createGhostChannel(tmp->name);
							timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                                                      	sSend(":%s KICK %s %s :You need a registered nick to join this channel", ChanServ, chan, nick->nick);
							remUserFromChan(nick, chan);
							return;
                                                 }
                                             }
                                        }

				
					/* Scanbot protection -GZ */

                                        if(tmp->reg->flags & CSCAN && a < 3) {
						if(IsScanBot(nick)) {
							createGhostChannel(tmp->name);
							timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                	                                sSend(":%s KICK %s %s :This channel does not allow Scanbots.", ChanServ, chan, nick->nick);
							remUserFromChan(nick, chan);
							return;
						}
                                        }

     
					if((tmp->reg->restrictlevel != 0) && ((a < tmp->reg->restrictlevel) || (!a && ((tmp->reg->mlock&PM_K)||(tmp->reg->mlock&PM_I))))) {
						/* IS BELOW RESTRICT LEVEL */
						createGhostChannel(tmp->name);
						timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
						if(!isServop(nick)) { 
//                                                    banKick(tmp, nick, "This channel is restricted to level %i and above", tmp->reg->restrictlevel);
						    banKick(tmp, nick, "Sorry. This is a restricted access channel.");

                                                }
						return;
					}
/*
					if(a && a < AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						if(status != 1) sSend(":%s MODE %s +v %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANVOICE;
						tmp->reg->timestamp = CTime;
					}
					else if(a && a >= AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						if(status != 2) sSend(":%s MODE %s +o %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANOP;
						tmp->reg->timestamp = CTime; 
                                                webbie = getRegNickData(nick->nick);
                                                if(webbie) {
                                                    if (webbie->flags2 & NWEBTV) 
                                                    sSend(":%s PRIVMSG %s :You are now a channel operator in %s.", WebServ, nick->nick, tmp->name);
                                                }
					} */

					if ((tmp->reg->flags & CKTOPIC) && (tmp->reg->topic[0]) && !(tmp->reg->flags & CFREEZE))
						/* IF KEEPTOPIC - RESTORE SAVED TOPIC */
						sSend(":%s TOPIC %s %s %lu :%s", ChanServ, tmp->name, tmp->reg->tsetby, tmp->reg->ttimestamp, tmp->reg->topic);

                               /* ChanServ will now enforce that absolutely nobody but SO's get ops */

                                        if ((tmp->reg->flags & CFREEZE) && !isServop(nick) && status == 2) {
                                             sSend(":%s MODE %s -o %s", ChanServ, tmp->name, nick->nick);
                                        }

   					makeModeLockStr(tmp->reg, mode);
					doRegChanMode(tmp->reg);
				}
			}
			else {
				/* NOT FIRST USER IN CHANNEL */			

				tmp->reg = getRegChanData(chan);

				person = getChanUserData(tmp, nick);
				if (person)
					delChanUser(tmp, person);

		 		person = smalloc(sizeof(cNickList));
				person->person = nick;
				person->op = 0;
				addChanUser(tmp, person);
				
				/* New CLOSED channel check. -GZ */

                                if(tmp->reg && tmp->reg->flags & CCLOSED) {
                                                if (!isServop(nick)) {
                                                        createGhostChannel(tmp->name);
                                                        timer(30, deleteGhostChannel, "deleteGhostChannel", tmp->name);
                                                        sSend(":%s MODE %s +b *!*@*", ChanServ, chan);
                                                        sSend(":%s KICK %s %s :This channel has been \002CLOSED\002 by the " NETWORKNAME " administration.", ChanServ, chan, nick->nick);
                                                        remUserFromChan(nick, chan);
                                                        return;
                                                }
                                }

				if(tmp->reg) {
					a=getChanOp(tmp->reg, nick->nick);
					/* IS AKICKED?? */
                                      /* Thou shall not akick servops */
                                        if(!isServop(nick)) {
					 if(a == -1) {
						cAkickList *blah;
						char userhost[NICKLEN + USERLEN + HOSTLEN];
							 
						sprintf(userhost, "%s!%s@%s", nick->nick, nick->user, nick->host);
						for(blah = tmp->reg->firstAkick;blah;blah=blah->next) {
							if(!match(userhost, blah->mask)) {
								cBanList *newban = smalloc(sizeof(cBanList));
								sSend(":%s MODE %s +b %s", ChanServ, chan, blah->mask);
								strcpy(newban->ban, blah->mask);
								addChanBan(tmp, newban);
								break;
							}
						}
						sSend(":%s KICK %s %s :You have been permanently banned from this channel.", ChanServ, chan, nick->nick);
						remUserFromChan(nick,chan);
						return;
					 }
                                        }

                                         /* Temporary AOL protection */

                                        if(tmp->reg->flags & CNOAOL) {
                                             if(!match(nick->host, "*.aol.com")) {
                                                 aol = getRegNickData(nick->nick);
                                                 if(!aol) {
                                                      	sSend(":%s MODE %s +b *!*%s@%s", ChanServ, chan, nick->user, nick->host);
                                                      	sSend(":%s KICK %s %s :You need a registered nick to join this channel", ChanServ, chan, nick->nick);
							remUserFromChan(nick,chan);
							return;
                                                 }
                                             }
                                        }

				
					/* Scanbot protection -GZ */

                                        if(tmp->reg->flags & CSCAN && a < 3) {
						if(IsScanBot(nick)) {
                                                      	sSend(":%s MODE %s +b *!*@%s", ChanServ, chan, nick->host);
                	                                sSend(":%s KICK %s %s :This channel does not allow Scanbots.", ChanServ, chan, nick->nick);
							remUserFromChan(nick, chan);
							return;
						}
                                        }

					if(a < tmp->reg->restrictlevel && tmp->reg->restrictlevel != 0) {
						/* BELOW RESTRICT LEVEL */
                                                if (!isServop(nick)) {
//						    banKick(tmp, nick, "This channel is restricted to level %i+", tmp->reg->restrictlevel);
						    banKick(tmp, nick, "Sorry. This is a restricted access channel.");
                                                }
						return;
					}
/*
					if(a && a < AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						if (status != 1) sSend(":%s MODE %s +v %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANVOICE;
						tmp->reg->timestamp = CTime;
					}
					else if(a && a >= AOP && !(tmp->reg->flags & CCLASSIC) && !(tmp->reg->flags & CFREEZE)) {
						if(status != 2) sSend(":%s MODE %s +o %s", ChanServ, tmp->name, nick->nick);
						person->op |= CHANOP;
						tmp->reg->timestamp = CTime;

                                                webbie = getRegNickData(nick->nick);
                                                if(webbie) {
                                                    if (webbie->flags2 & NWEBTV)
                                                    sSend(":%s PRIVMSG %s :You are now a channel operator in %s.", WebServ, nick->nick, tmp->name); 
                                                }
					}
                                        */

                                        /* Make sure nobody gets ops */

                                        if ((tmp->reg->flags & CFREEZE) && !isServop(nick) && status == 2) {
						uSend(ChanServ, nick, GetReply(RPL_NOOPSFROZEN, nick));
						sSend(":%s MODE %s -o %s", ChanServ, tmp->name, nick->nick);
                                        }

				}
#ifdef CDEBUG
					 sSend(":%s GNOTICE :Added %s to channel %s", myname, nick->nick, chan);
#endif
			}
		}
	}
}

void CcloseCmd(UserList *nick, char **args, int numargs) {
	ChanList			*channel;
	RegChanList		*regchan;
	int		i;
	cAccessList		*founder;
	cNickList		*x, *nextnick = NULL;
	char			reason[512];
	if(!strcasecmp(args[1], "add")) {
	
		if(!IsChannelName(args[2])) {
			uSend(ChanServ, nick, GetReply(ERR_BADCHANNELNAME, nick), args[2]);
			return;
		}

		channel = getChanData(args[2]);
		regchan = getRegChanData(args[2]);
		*reason = '\0';
		for (i = 3; i < numargs; i++)
		{
			if ((strlen(reason) + strlen(args[i])) > 510)
				break;
			strcat(reason, args[i]);
			if ((i + 1) != numargs)
				strcat(reason, " ");
		}

		if (!*reason)
		{
			sSend(":%s NOTICE %s :Syntax error: \2/CS CLOSE ADD #Channel <reason>\2", ChanServ,nick->nick);
			return;
		}
		if (!regchan) {
			regchan = smalloc(sizeof(RegChanList));
			initRegChanData(regchan);

			if(channel) 
        		channel->reg = regchan;

        		strcpy(regchan->name, args[2]);
        		strcpy(regchan->founder, "" NETWORKNAME "");

        		for(i=0;i<8;i++)
                		regchan->password[i] = (((u_short)rand())%26)+97;

        		regchan->timereg = CTime;
        		regchan->timestamp = CTime;
        
        		founder = smalloc(sizeof(cAccessList));
	        	strcpy(founder->nick, "" NETWORKNAME "");
        		founder->uflags = FOUNDER;
				addChanOp(regchan, founder);
				regchan->mlock |= PM_N;
				regchan->mlock |= PM_T;
				regchan->mlock |= PM_I;
				regchan->mlock |= PM_S;
				addRegChan(regchan);
				mostchans++;
		}


		regchan->flags |= CCLOSED;
		regchan->flags |= CKTOPIC;
		strcpy(regchan->topic, "This channel has been \002CLOSED\002 by the " NETWORKNAME " administration.");
		
		if(channel) {			
			createGhostChannel(channel->name);
			timer(30,deleteGhostChannel, "deleteGhostChannel", channel->name);
			for(x=channel->firstUser;x;x=nextnick) {
				nextnick = x->next;

				if(!isOper(x->person)) {
					sSend(":%s KICK %s %s :This channel has been \002CLOSED\002 by the " NETWORKNAME " administration.", ChanServ, channel->name, x->person->nick);
					remUserFromChan(x->person,channel->name);
				}
			}
			sSend(":%s TOPIC %s :%s", ChanServ, channel->name, regchan->topic);
		}

		uSend(ChanServ, nick, GetReply(RPL_CHANCLOSED, nick), regchan->name);
		sSend(":%s GLOBOPS :%s has CLOSED the channel \002%s\002 (%s) - founder: %s.", ChanServ, nick->nick, regchan->name, reason, regchan->founder);
		sSend(":%s NOTICE @#closed :(Ops/#Closed) %s has CLOSED the channel \002%s\002 (%s) - founder: %s", ChanServ, nick->nick, regchan->name, reason, regchan->founder);
		logDump(chanlog, "%s has CLOSED the channel %s (%s) - founder: %s.", nick->nick, regchan->name, reason, regchan->founder);
		/* int sendMail(char *to, char *from, char *subject, char *format, ...) */					
		sendMail("" CLOSUREADDY "", "Services@" NETWORK "", "Closure of channel", "%s has closed channel %s - founder is %s and reason is \"%s\"",
			nick->nick, regchan->name, regchan->founder, reason);
	}

	else if(!strcasecmp(args[1], "del")) {

		regchan = getRegChanData(args[2]);

        	if(!regchan) {
	                uSend(ChanServ, nick, GetReply(ERR_CHANNOTREGGED, nick), args[2]);
                	return;         
        	}		
		
		if(!regchan->flags & CCLOSED) {
	                uSend(ChanServ, nick, GetReply(ERR_CHANNOTCLOSED, nick), regchan->name);
			return;
		}
		
		regchan->flags &= ~CCLOSED;
	        uSend(ChanServ, nick, GetReply(RPL_CHANUNCLOSED, nick), regchan->name);
		sSend(":%s GLOBOPS :%s has REOPENED the channel \002%s\002 (founder: %s).", ChanServ, nick->nick, regchan->name, regchan->founder);
		logDump(chanlog, "%s has REOPENED the channel %s (founder: %s).", nick->nick, regchan->name, regchan->founder);
		sSend(":%s NOTICE @#closed :(Ops/#Closed) %s has REOPENED the channel \002%s\002 - founder: %s", ChanServ, nick->nick, regchan->name, regchan->founder);

		sendMail("" CLOSUREADDY "", "Services@" NETWORK, "Reopening of channel", "%s has reopened channel %s - founder: %s",
			nick->nick, regchan->name, regchan->founder);

		if(!strcmp(regchan->founder, "" NETWORKNAME "")) {
			delRegChan(regchan);
		        mostchans--;
		}
	}
	
	else {
		uSend(ChanServ, nick, GetReply(ERR_BADPARAMS, nick));
		uSend(ChanServ, nick, GetReply(RPL_MOREHELPCMD, nick), ChanServ, "CLOSE");
		return;
	}

	return;
}


