/*
 * Copyright (c) 1999-2000 GZ
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "services.h"

/* NickServ -- Nicknames. save/read/delete */

void saveNickData(RegNickList *first)
{
	RegNickList *nl = first;
	nAccessList *accitem;
	int i, x=0;
	db.ns = fopen(DATABASEDIR "/nickserv.db", "w");
	if (db.ns == NULL) {
		Debug((DEBUG_ERROR,"Failed to open nickserv db"));
		return;
	}
   while(nl != NULL) {
		i=0;
		/* new nick insertion point */
		fprintf(db.ns, "nick ");
		fprintf(db.ns, "%s %s %s ", nl->nick, nl->user, nl->host);
		fprintf(db.ns, "%s %s ", nl->password, nl->password2);
		fprintf(db.ns, "%s ", nl->email);
		fprintf(db.ns, "%lu ", (unsigned long)nl->timestamp);
		fprintf(db.ns, "%lu ", (unsigned long)nl->timereg);
		fprintf(db.ns, "%i %i :%s\n", nl->flags, nl->chans, nl->realname);
		if (nl->lastpwchange > nl->timereg)
			fprintf(db.ns, "expire %s %lu\n", nl->nick, (unsigned long)nl->lastpwchange);
		if(nl->url[0])
		  	fprintf(db.ns, "url %s %s\n", nl->nick, nl->url);
		for(accitem=nl->firstAcc;accitem;accitem=accitem->next) {
			fprintf(db.ns, "access %s %s\n", nl->nick, accitem->mask);
		}
		if(nl->info[0])
		  fprintf(db.ns, "info %s x x x x x x x x x :%s\n", nl->nick, nl->info);
		if(nl->flags2)
		  fprintf(db.ns, "flags %s %i\n", nl->nick, nl->flags2);
		x++;
		fprintf(db.ns, "language %s %i\n", nl->nick, nl->language);
		fflush(db.ns);
#ifdef DBDEBUG
		sSend("GNOTICE :Saved nick data (%s)", nl->nick);
#endif
		nl = nl->next;
	}
	fprintf(db.ns, "done\n");
	fflush(db.ns);
	fclose(db.ns);
#ifdef SHOWSYNCH
	sSend(":%s GLOBOPS :Completed save(%i)", NickServ, x);
#endif
}


void readNickData()
{
	RegNickList *read;
	char line[1024];
	char *tmp = line;
	char args[13][129];
	int done = 0;
	int a, x, teller;
/* LET'S TRY WITHOUT THIS SILLYNESS 
	readRegData();
*/
	db.ns = fopen(DATABASEDIR "/nickserv.db", "r");
	if(!db.ns) {
	  Debug((DEBUG_ERROR,"Failed to open nickserv db"));
	  return;
	}
	while(!done) {
		memset(args, 0, 13*129);
		if (!(sfgets(line, 1024, db.ns))) {
			done = 1;
			fclose(db.ns);
			return;
		}
		tmp = line;
		x = a = 0;
		while(*tmp && x < 12) {
			if(x == 11) {
				tmp++;
				while(*tmp) {
					args[x][a] = *tmp;
					a++;
					tmp++;
				}
			}
			else {
				while(*tmp != ' ' && *tmp) {
					args[x][a] = *tmp;
					a++;
					tmp++;
				}
			}
			
			args[x][a] = 0;
			x++;
			while(*tmp == ' ')
			  tmp++;
			a = 0;
		}
		teller++;
		if(!strcmp(args[0], "nick")) {
			read = smalloc(sizeof(RegNickList));
			memset(read, 0, sizeof(RegNickList));
			strcpy(read->nick, args[1]);
			strcpy(read->user, args[2]);
			strcpy(read->host, args[3]);
			strcpy(read->password, args[4]);
			strcpy(read->password2, args[5]);
			strcpy(read->email, args[6]);
			read->timestamp = (time_t)atol(args[7]);
			read->timereg = (time_t)atol(args[8]);
			read->lastpwchange = read->timereg;  /* THIS WAY TO SAVE STORAGE */
			read->amasks = 0; /* atoi(args[9]);  <--- not needed... */
			read->flags = atoi(args[9]);
			read->chans = atoi(args[10]);
			strcpy(read->realname, args[11]);
			read->flags &= ~NNETADM;
			read->flags &= ~NROOT;
			addRegNick(read);
		}
		else if(!strcmp(args[0], "expire")) {
			read = getRegNickData(args[1]);
			read->lastpwchange = (time_t)atol(args[2]);
		}
		else if(!strcmp(args[0], "url")) {
			read = getRegNickData(args[1]);
			strcpy(read->url, args[2]);
		}
		else if(!strcmp(args[0], "info")) {
			read=getRegNickData(args[1]);
			strcpy(read->info,args[11]);
		}
		else if(!strcmp(args[0], "flags")) {
			read=getRegNickData(args[1]);
			read->flags2 = atoi(args[2]);
		}
		else if(!strcmp(args[0], "access")) {
			read = getRegNickData(args[1]);
			addAccItem(read, args[2]);
		}
		else if(!strcmp(args[0], "language")) {
			read = getRegNickData(args[1]);
			read->language = atoi(args[2]);
		}
		else if(!strcmp(args[0], "done"))
		  done = 1;
		else {
		  printf("failed.\n");
		  printf("Error reading: line: %s (%i)\n", line, teller);
		  sshutdown(-1);
		}
#ifdef DBDEBUG
		  sSend("GNOTICE :Read nick data (%s)", line);
#endif
	}
	fclose(db.ns);
	readMemoData();
}

/* ChanServ, all things ChanServ, and the like.... */

void saveChanData(RegChanList *first)
{
	RegChanList *cl = first;
	cAccessList *accitem;
	cAkickList *akitem;
	int x=0;
	db.cs = fopen(DATABASEDIR "/chanserv.db", "w");
	if (db.cs == NULL) {
		Debug((DEBUG_ERROR,"Failed to open chanserv db"));
		return;
	}
   while(cl) {
		/* new chan insertion point */
		fprintf(db.cs, "channel ");
		fprintf(db.cs, "%s %s ", cl->name, cl->founder);
		fprintf(db.cs, "%lu %lu %s ", cl->mlock, cl->flags, cl->password);
		fprintf(db.cs, "%lu ", (unsigned long)cl->timereg);
		fprintf(db.cs, "%lu ", (unsigned long)cl->timestamp);
		fprintf(db.cs, "%s %lu ", cl->key, cl->limit);
		fprintf(db.cs, "%i %i :%s\n", cl->tlocklevel, cl->restrictlevel, cl->desc);
		if(cl->topic[0])
		  fprintf(db.cs, "topic %s %s %lu :%s\n", cl->name, cl->tsetby, cl->ttimestamp, cl->topic);
		if(cl->firstOp) {
			for(accitem=cl->firstOp;accitem;accitem=accitem->next) {
				fprintf(db.cs, "op %s %s ", cl->name, accitem->nick);
				fprintf(db.cs, "%i\n", accitem->uflags);
			}
		}
		
		if(cl->firstAkick) {
			for(akitem=cl->firstAkick;akitem;akitem = akitem->next) {
				fprintf(db.cs, "akick %s %s ", cl->name, akitem->mask);
				fprintf(db.cs, "%lu :%s\n", (unsigned long)akitem->added, akitem->reason);
			}
		}
		x++;
		fflush(db.cs);
#ifdef DBDEBUG
		sSend("GNOTICE :Saved chan data (%s)", cl->name);
#endif
		cl = cl->next;
	}
	fprintf(db.cs, "done\n");
	fflush(db.cs);
	fclose(db.cs);
#ifdef SHOWSYNCH
	sSend(":%s GLOBOPS :Completed save(%i)", ChanServ, x);
#endif
}

void readChanData()
{
	RegChanList *read;
	char line[2048];
	char *tmp = line;
	char args[15][512];
	int done = 0;
	int x,a, teller;
	db.cs = fopen(DATABASEDIR "/chanserv.db", "r");
	if(!db.cs) {
		Debug((DEBUG_ERROR,
			"Failed to open \"" DATABASEDIR "/chanserv.db\"!"));
		return;
	}
	while(!done) {
		read=NULL;
		bzero(line,2048);
		tmp = line;
		memset(args, 0, 15*512);
		
		if (!(sfgets(line, 2048, db.cs))) {
			done = 1;
			fclose(db.cs);
 		}
		tmp = line;
		x = a = 0;
		while(*tmp && x < 13) {
			if((x == 4 || x == 12) && (*tmp == ':')) {
				tmp++;
				while(*tmp) {
					args[x][a] = *tmp;
					a++;
					tmp++;
				}
			}
			else {
				while(*tmp != ' ' && *tmp) {
					args[x][a] = *tmp;
					a++;
					tmp++;
				}
			}
			
			args[x][a] = 0;
			x++;
			while(*tmp == ' ')
			  tmp++;
			a = 0;
		}
		teller++;
		if(!strcmp(args[0], "channel")) {
			read = smalloc(sizeof(RegChanList));
			initRegChanData(read);
			strcpy(read->name, args[1]);
			strcpy(read->founder, args[2]);
			read->mlock = atol(args[3]);
			read->flags = atol(args[4]);
			strcpy(read->password, args[5]);
			read->timereg = (time_t)atol(args[6]);
			read->timestamp = (time_t)atol(args[7]);
			strcpy(read->key, args[8]);
			read->limit = atol(args[9]);
			read->tlocklevel = atoi(args[10]);
			read->restrictlevel = atoi(args[11]);
			strcpy(read->desc, args[12]);
			addRegChan(read);
			mostchans++;
		}
		else if(!strcmp(args[0], "topic")) {
			char *s1;
			s1 = args[4];
			while (*s1 == ' ')
				s1++;
			if (*s1)
			{
				read = getRegChanData(args[1]);
				strcpy(read->tsetby, args[2]);
				read->ttimestamp = (time_t)atol(args[3]);
				strcpy(read->topic, s1);
			}
		}
		else if(!strcmp(args[0], "op")) {
			cAccessList *lame;
			lame = smalloc(sizeof(cAccessList));
			read = getRegChanData(args[1]);
			strcpy(lame->nick, args[2]);
			lame->uflags = atoi(args[3]);
			if(getRegNickData(args[2]))
			  addChanOp(read, lame);
			else
			  free(lame);
		}
		else if(!strcmp(args[0], "akick")) {
			cAkickList *lame;
			lame = smalloc(sizeof(cAkickList));
			read = getRegChanData(args[1]);
			strcpy(lame->mask, args[2]);
			lame->added = (time_t)atol(args[3]);
			strcpy(lame->reason, args[4]);
			addChanAkick(read, lame);
		}
		else if(!strcmp(args[0], "done"))
		  done = 1;
		else {
                  printf("failed.\n");
                  printf("Error reading: line: %s (%i)\n", args[0], teller);
		  sshutdown(-1);
		}
		
#ifdef DBDEBUG
		sSend("GNOTICE :Read chan data (%s)", line);
#endif
	}
	fclose(db.cs);
}


void saveMemoData(RegNickList *first) {
	RegNickList *nl = first;
	MemoBox *tmp;
	MemoList *memo;
	int x=0;
	db.ms = fopen(DATABASEDIR "/memoserv.db", "w");
	if (db.ms == NULL) {
		Debug((DEBUG_ERROR,"Failed to open memoserv db"));
		return;
	}

   while(nl != NULL) {
		sprintf(debugbuff, "Memo save %s", nl->nick); 
		tmp = nl->memos;
		if(tmp) {
			fprintf(db.ms, "data %s %i %i %i %i %lu\n", nl->nick, tmp->memos, tmp->sent, tmp->flags, tmp->max, (long)tmp->lastsent); 
			if((tmp->flags & MFORWARDED) && (tmp->redirect[0]))
			  fprintf(db.ms, "redirect %s %s\n", nl->nick, tmp->redirect);
			for(memo=tmp->firstMemo;memo;memo=memo->next)
			  fprintf(db.ms, "memo %s %i %i %lu %s :%s\n", nl->nick, memo->index, memo->flags, (long)memo->sent, memo->from, memo->memo);
			x++;
#ifdef DBDEBUG
			sSend("GNOTICE :Saved memo data (%s)", nl->nick);
#endif
		}
		else {
			fprintf(db.ms, "data %s 0 0 0 %i\n", nl->nick, MS_DEF_RCV_MAX);
			x++;
		}
		fflush(db.ms);
		strcat(debugbuff," (success)");
		nl = nl->next;
	}
	fprintf(db.ms, "done\n");
	fflush(db.ms);
	fclose(db.ms);
#ifdef SHOWSYNCH
	sSend(":%s GLOBOPS :Completed save(%i)", MemoServ, mostmemos);
#endif
}

void readMemoData()
{
	RegNickList *nick;
	MemoBox *newbox;
	MemoList *newmemo;
	char line[2048];
	char *tmp = line;
	char args[15][MEMOLEN];
	int done = 0;
	int x,a;

	mostmemos = 0;
	db.ms = fopen(DATABASEDIR "/memoserv.db", "r");
	if(!db.ms) {
		Debug((DEBUG_ERROR,"Failed to open memoserv db"));
	  return;
	}
	while(!done) {
		if (!(sfgets(line, 2048, db.ms))) {
			done = 1;
			fclose(db.ms);
			return;
 		}
		for (x = 0; x < 7; x++)
			args[x][0] = 0;

		tmp = line;
		x = a = 0;
		while(*tmp && x < 7) {
			if((x == 4 || x == 6) && *tmp == ':') {
				tmp++;
				while(*tmp) {
					args[x][a] = *tmp;
					a++;
					tmp++;
				}
			}
			else {
				while(*tmp != ' ' && *tmp) {
					args[x][a] = *tmp;
					a++;
					tmp++;
				}
			}
			
			args[x][a] = 0;
			x++;
			while(*tmp == ' ')
			  tmp++;
			a = 0;
		}
		
		if(!strcmp(args[0], "data")) {
			newbox = smalloc(sizeof(MemoBox));
			nick = getRegNickData(args[1]);
			if(nick) {
				nick->memos = newbox;
				newbox->memos = atoi(args[2]);
				newbox->sent = atoi(args[3]);
				newbox->flags = atoi(args[4]);
				newbox->max = atoi(args[5]);
				newbox->lastsent = (time_t)atol(args[6]);
			}
			else {
				free(newbox);
			}
		}
		else if(!strcmp(args[0], "redirect")) {
			RegNickList *fwd_nick;
			nick = getRegNickData(args[1]);
			fwd_nick = getRegNickData(args[2]);
			if (!fwd_nick)
				nick->memos->flags &= ~ MFORWARDED;
			else
				strcpy(nick->memos->redirect, args[2]);
		}
		else if (!strcmp(args[0], "memo")) {
			newmemo = smalloc(sizeof(MemoList));
			nick = getRegNickData(args[1]);
			if(nick) {
				newmemo->index = atoi(args[2]);
				newmemo->flags = atoi(args[3]);
				newmemo->sent = (time_t)atol(args[4]);
				strcpy(newmemo->from, args[5]);
				strcpy(newmemo->memo, args[6]);
				addMemo(nick->memos, newmemo);
			}
			else
			  sfree(newmemo);
		}
		else if(!strcmp(args[0], "done"))
		  done = 1;
		else {
                  printf("failed.\n");
                  printf("Error reading: line: %s\n", line);
		  sshutdown(-1);
		}
		
#ifdef DBDEBUG
		sSend("GNOTICE :Read memo data (%s)", line);
#endif
	}
	fclose(db.ms);
}

void sync_cfg(char *type) {
	type++;
	switch(*type) {
	 case '1':
		syncNickData((time_t)(CTime + (SYNCTIME * 3)));
		nextCsync = (CTime + SYNCTIME);
		nextMsync = (CTime + (SYNCTIME * 2));
		nextNsync = (CTime + (SYNCTIME * 3));
		timer(SYNCTIME, doTfunc, "doTfunc", "42");
		break;
	 case '2':
		syncChanData((time_t)(CTime + (SYNCTIME * 2)));
		nextMsync = (CTime + SYNCTIME);
		nextNsync = (CTime + (SYNCTIME * 2));
		nextCsync = (CTime + (SYNCTIME * 2));
		timer(SYNCTIME, doTfunc, "doTfunc", "43");
		break;
	 case '3':
		syncMemoData((time_t)(CTime + SYNCTIME));
		nextNsync = nextCsync = nextMsync = (CTime + SYNCTIME);
		timer(SYNCTIME, doTfunc, "doTfunc", "44");
		break;
	 case '4':
		syncNickData((time_t)(CTime + SYNCTIME));
		syncChanData((time_t)(CTime + (SYNCTIME * 2)));
		syncMemoData((time_t)(CTime + (SYNCTIME * 3)));
		nextNsync = (CTime + SYNCTIME);
		nextCsync = (CTime + SYNCTIME * 2);
		nextMsync = (CTime + SYNCTIME * 3);
		timer(SYNCTIME, doTfunc, "doTfunc", "41");
		break;
	}
}

void disk_synch() {
                saveChanData(firstRegChan);
                saveNickData(firstRegNick);
		saveMemoData(firstRegNick);
}

void do_remaining_backup() {

	char backup[50];
	char cmd[512];
	time_t timestart;

	timestart = time(NULL);

        strftime(backup, 30, "akill.db%d%m%Y", localtime(&timestart));
        sprintf(cmd, "cp %s/akill.db %s/%s",
                        DATABASEDIR, DATABASEDIR, backup);
        system(cmd);

        strftime(backup, 30, "dekill.db%d%m%Y", localtime(&timestart));
        sprintf(cmd, "cp %s/dekill.db %s/%s",
                        DATABASEDIR, DATABASEDIR, backup);
        system(cmd);

        strftime(backup, 30, "hostjinx.db%d%m%Y", localtime(&timestart));
        sprintf(cmd, "cp %s/hostjinx.db %s/%s",
                        DATABASEDIR, DATABASEDIR, backup);
        system(cmd);

        strftime(backup, 30, "memoserv.db%d%m%Y", localtime(&timestart));
        sprintf(cmd, "cp %s/memoserv.db %s/%s",
                        DATABASEDIR, DATABASEDIR, backup);
        system(cmd);

        strftime(backup, 30, "permignore.db%d%m%Y", localtime(&timestart));
        sprintf(cmd, "cp %s/permignore.db %s/%s",
                        DATABASEDIR, DATABASEDIR, backup);
        system(cmd);

        strftime(backup, 30, "permjinx.db%d%m%Y", localtime(&timestart));
        sprintf(cmd, "cp %s/permjinx.db %s/%s",
                        DATABASEDIR, DATABASEDIR, backup);
        system(cmd);

	timer(2*3600, doTfunc, "doTfunc", "6");
}
