/*
 * Copyright (c) 1999-2000 GZ
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include "services.h"

/* This will load the permjinxes from disk */

void savePermJinxes() {
	FILE *permjinx_file;

	aPJinx *pjinxptr;
    	aPJinx *pjinxptr2;

	permjinx_file = fopen(DATABASEDIR "/permjinx.db","w");

	if (permjinx_file == NULL) {
		Debug((DEBUG_ERROR,"Failed to open permjinx file"));
		return;
	}

	if (PJinxList == NULL)
        	return;

	for (pjinxptr = PJinxList; pjinxptr; pjinxptr = pjinxptr2) {
        	pjinxptr2 = pjinxptr->next;

		fprintf(permjinx_file,"-\n");
		fprintf(permjinx_file,"%s\n",pjinxptr->nick);
		fprintf(permjinx_file,"%s\n",pjinxptr->by);
		fprintf(permjinx_file,"%s\n",pjinxptr->reason);
		fprintf(permjinx_file,"%lu\n", (long) pjinxptr->time);
	}

	fclose(permjinx_file);
}

/* This will load the hostjinxes from disk */

void loadHostJinxes(void) {

    FILE *hostjinx_file;
    char buffer[512];

    char host[100];
    char by[50];
    char reason[250];
    long timestamp;

    bzero(host,100);
    bzero(by, 50);
    bzero(reason, 250);

    hostjinx_file = fopen(DATABASEDIR "/hostjinx.db","r");

    if (!hostjinx_file) {
	return;
    }

    while (sfgets(buffer, 512, hostjinx_file)) {
        if (strcmp(buffer, "-") != 0) {
             sSend(":%s GLOBOPS :Sync error in hostjinx.db", OperServ);
             fclose(hostjinx_file);
             return;
         }
         fscanf(hostjinx_file,"%s\n", (char *) host);
         fscanf(hostjinx_file,"%s\n",by);
	 sfgets(reason, 250, hostjinx_file);
         fscanf(hostjinx_file,"%lu\n", &timestamp);

	 AddHJinxTime(host, by, reason, timestamp);

     }
     fclose(hostjinx_file);
}

/* This will load the permjinxes from disk */

void loadPermJinxes(void) {

    FILE *permjinx_file;
    char buffer[512];

    char host[100];
    char by[50];
    char reason[250];
    long timestamp;

    bzero(host,100);
    bzero(by, 50);
    bzero(reason, 250);

    permjinx_file = fopen(DATABASEDIR "/permjinx.db","r");

    if (!permjinx_file) {
	return;
    }

    while (sfgets(buffer, 512, permjinx_file)) {
        if (strcmp(buffer, "-") != 0) {
             sSend(":%s GLOBOPS :Sync error in permjinx.db", OperServ);
             fclose(permjinx_file);
             return;
         }
         fscanf(permjinx_file,"%s\n", (char *) host);
         fscanf(permjinx_file,"%s\n",by);
	 sfgets(reason, 250, permjinx_file);
         fscanf(permjinx_file,"%lu\n", &timestamp);

	 AddPJinxTime(host, by, reason, timestamp);

     }
     fclose(permjinx_file);
}

void saveHostJinxes() {
	FILE *hostjinx_file;

	aHJinx *hjinxptr;
    	aHJinx *hjinxptr2;

	hostjinx_file = fopen(DATABASEDIR "/hostjinx.db","w");

	if (hostjinx_file == NULL) {
		Debug((DEBUG_ERROR,"Failed to open hostjinx file"));
		return;
	}

	if (HJinxList == NULL)
        	return;

	for (hjinxptr = HJinxList; hjinxptr; hjinxptr = hjinxptr2) {
        	hjinxptr2 = hjinxptr->next;

		fprintf(hostjinx_file,"-\n");
		fprintf(hostjinx_file,"%s\n",hjinxptr->host);
		fprintf(hostjinx_file,"%s\n",hjinxptr->by);
		fprintf(hostjinx_file,"%s\n",hjinxptr->reason);
		fprintf(hostjinx_file,"%lu\n", (long) hjinxptr->time);
	}

	fclose(hostjinx_file);
}

void SynchHJinxes() {
    aHJinx *hjinxptr;
    aHJinx *hjinxptr2;

    if (HJinxList == NULL)
        return;

    for (hjinxptr = HJinxList; hjinxptr; hjinxptr = hjinxptr2) {
        hjinxptr2 = hjinxptr->next;
	sSend(":%s JINX %s", myname, hjinxptr->host);
    }

    return;
}	
void checkPermJinx(char *nick) {
	
	if(CheckPJinx(nick)) {
		if(!CheckJinx(nick))
			AddJinx(nick, "OperServ", "Permanent JINX");
		sSend(":%s GLOBOPS :Re-activating JINX for %s.", OperServ, nick);
		sSend(":%s JINX %s", myname, nick);
	}

	return;
}

void CheckHostJinx(char *server, char *host, int action) {
	
	if(action) {
		if(!CheckHJinx(host)) {
			sSend(":%s GLOBOPS :Removing unknown jinx %s introduced by %s", OperServ, host, server);
			sSend(":%s UNJINX %s", myname, host);
		}	
	return;
	}
	else {
		if(CheckHJinx(host)) {
			sSend(":%s GLOBOPS :Restoring jinx %s because of removal by %s", OperServ, host, server);
			sSend(":%s JINX %s", myname, host);
		}
	}			
	
	return;
}

void AddJinx(char *nick, char *by, char *reason) {

        aJinx *Jinxptr;

	Jinxptr = make_jinx();
        strncpy(Jinxptr->nick, nick, sizeof(Jinxptr->nick));
	strncpy(Jinxptr->by, by, sizeof(Jinxptr->by));
	Jinxptr->time = CTime;
	strncpy(Jinxptr->reason, reason, sizeof(Jinxptr->reason));

    if (JinxList == NULL)
	JinxList = Jinxptr;
    else {
	JinxList->prev = Jinxptr;
	Jinxptr->next = JinxList;
	JinxList = Jinxptr;
    }

    return;
}

void AddPJinx(char *nick, char *by, char *reason) {

        aPJinx *PJinxptr;

	PJinxptr = make_pjinx();
        strncpy(PJinxptr->nick, nick, sizeof(PJinxptr->nick));
	strncpy(PJinxptr->by, by, sizeof(PJinxptr->by));
	PJinxptr->time = CTime;
	strncpy(PJinxptr->reason, reason, sizeof(PJinxptr->reason));

    if (PJinxList == NULL)
	PJinxList = PJinxptr;
    else {
	PJinxList->prev = PJinxptr;
	PJinxptr->next = PJinxList;
	PJinxList = PJinxptr;
    }

    return;
}

void AddHJinx(char *host, char *by, char *reason) {

        aHJinx *HJinxptr;

	HJinxptr = make_hjinx();
        strncpy(HJinxptr->host, host, sizeof(HJinxptr->host));
	strncpy(HJinxptr->by, by, sizeof(HJinxptr->by));
	HJinxptr->time = CTime;
	strncpy(HJinxptr->reason, reason, sizeof(HJinxptr->reason));

    if (HJinxList == NULL)
	HJinxList = HJinxptr;
    else {
	HJinxList->prev = HJinxptr;
	HJinxptr->next = HJinxList;
	HJinxList = HJinxptr;
    }

    return;
}
void AddHJinxTime(char *host, char *by, char *reason, long time) {

        aHJinx *HJinxptr;

	Debug((DEBUG_NOTICE, "Adding HJINX for %s %s %s %lu", host, by, reason, time));
	HJinxptr = make_hjinx();
        strncpy(HJinxptr->host, host, sizeof(HJinxptr->host));
	strncpy(HJinxptr->by, by, sizeof(HJinxptr->by));
	HJinxptr->time = time;
	strncpy(HJinxptr->reason, reason, sizeof(HJinxptr->reason));

    if (HJinxList == NULL)
	HJinxList = HJinxptr;
    else {
	HJinxList->prev = HJinxptr;
	HJinxptr->next = HJinxList;
	HJinxList = HJinxptr;
    }

    return;
}

void AddPJinxTime(char *nick, char *by, char *reason, long time) {

        aPJinx *PJinxptr;

	PJinxptr = make_pjinx();
        strncpy(PJinxptr->nick, nick, sizeof(PJinxptr->nick));
	strncpy(PJinxptr->by, by, sizeof(PJinxptr->by));
	PJinxptr->time = time;
	strncpy(PJinxptr->reason, reason, sizeof(PJinxptr->reason));

    if (PJinxList == NULL)
	PJinxList = PJinxptr;
    else {
	PJinxList->prev = PJinxptr;
	PJinxptr->next = PJinxList;
	PJinxList = PJinxptr;
    }

    return;
}

aPJinx *make_pjinx(){
    aPJinx *PJinxptr;

    PJinxptr = (aPJinx *) smalloc(sizeof(aPJinx));

    PJinxptr->nick[0]='\0';
    PJinxptr->by[0]='\0';
    PJinxptr->time = 0;
    PJinxptr->reason[0]='\0';
    PJinxptr->prev = PJinxptr->next = NULL;

    return PJinxptr;
}

void free_pjinx(pjinxptr)
    aPJinx *pjinxptr;
{
    free((char *) pjinxptr);
}

aHJinx *make_hjinx(){
    aHJinx *HJinxptr;

    HJinxptr = (aHJinx *) smalloc(sizeof(aHJinx));

    HJinxptr->host[0]='\0';
    HJinxptr->by[0]='\0';
    HJinxptr->time = 0;
    HJinxptr->reason[0]='\0';
    HJinxptr->prev = HJinxptr->next = NULL;

    return HJinxptr;
}

void free_hjinx(hjinxptr)
    aHJinx *hjinxptr;
{
    free((char *) hjinxptr);
}

aJinx *make_jinx(){
    aJinx *Jinxptr;

    Jinxptr = (aJinx *) smalloc(sizeof(aJinx));

    Jinxptr->nick[0]='\0';
    Jinxptr->by[0]='\0';
    Jinxptr->time = 0;
    Jinxptr->reason[0]='\0';
    Jinxptr->prev = Jinxptr->next = NULL;

    return Jinxptr;
}

void free_jinx(jinxptr)
    aJinx *jinxptr;
{
    free((char *) jinxptr);
}

void DeleteJinx(char *nick) {

    aJinx *jinxptr;
    aJinx *jinxptr2;

    if (JinxList == NULL)
        return;

    for (jinxptr = JinxList; jinxptr; jinxptr = jinxptr2) {
        jinxptr2 = jinxptr->next;
	if(!strcasecmp(jinxptr->nick, nick)) {
            if (jinxptr->prev)
                jinxptr->prev->next = jinxptr->next;
            else
                JinxList = jinxptr->next;
            if (jinxptr->next)
                jinxptr->next->prev = jinxptr->prev;
	    free_jinx(jinxptr);
	    return;
	}
    }

    return;
}

void DeleteHJinx(char *host) {

    aHJinx *hjinxptr;
    aHJinx *hjinxptr2;

    if (HJinxList == NULL)
        return;

    for (hjinxptr = HJinxList; hjinxptr; hjinxptr = hjinxptr2) {
        hjinxptr2 = hjinxptr->next;
	if(!strcasecmp(hjinxptr->host, host)) {
            if (hjinxptr->prev)
                hjinxptr->prev->next = hjinxptr->next;
            else
                HJinxList = hjinxptr->next;
            if (hjinxptr->next)
                hjinxptr->next->prev = hjinxptr->prev;
	    free_hjinx(hjinxptr);
	    return;
	}
    }

    return;
}

int CheckJinx(char *jinx) {

    aJinx *jinxptr;
    aJinx *jinxptr2;

    if (JinxList == NULL)
        return 0;

    for (jinxptr = JinxList; jinxptr; jinxptr = jinxptr2) {
        jinxptr2 = jinxptr->next;
	if(!strcasecmp(jinxptr->nick, jinx)) {
	    return 1;
	}
    }

    return 0;
}

void DeletePJinx(char *nick) {

    aPJinx *pjinxptr;
    aPJinx *pjinxptr2;

    if (PJinxList == NULL)
        return;

    for (pjinxptr = PJinxList; pjinxptr; pjinxptr = pjinxptr2) {
        pjinxptr2 = pjinxptr->next;
	if(!strcasecmp(pjinxptr->nick, nick)) {
            if (pjinxptr->prev)
                pjinxptr->prev->next = pjinxptr->next;
            else
                PJinxList = pjinxptr->next;
            if (pjinxptr->next)
                pjinxptr->next->prev = pjinxptr->prev;
	    free_pjinx(pjinxptr);
	    return;
	}
    }

    return;
}

int CheckPJinx(char *pjinx) {

    aPJinx *pjinxptr;
    aPJinx *pjinxptr2;

    if (PJinxList == NULL)
        return 0;

    for (pjinxptr = PJinxList; pjinxptr; pjinxptr = pjinxptr2) {
        pjinxptr2 = pjinxptr->next;
	if(!strcasecmp(pjinxptr->nick, pjinx)) {
	    return 1;
	}
    }

    return 0;
}

int CheckHJinx(char *hjinx) {

    aHJinx *hjinxptr;
    aHJinx *hjinxptr2;

    if (HJinxList == NULL)
        return 0;

    for (hjinxptr = HJinxList; hjinxptr; hjinxptr = hjinxptr2) {
        hjinxptr2 = hjinxptr->next;
	if(!strcasecmp(hjinxptr->host, hjinx)) {
	    return 1;
	}
    }

    return 0;
}

