/*
 * Copyright (c) 1999-2000 GZ
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include "services.h"

void indexMemoList(MemoBox *);

void initMemoBox(MemoBox *memobox) {
	memobox->memos = memobox->flags = memobox->sent = 0;
	memobox->lastsent = 0;
	memobox->max = MS_DEF_RCV_MAX;
	memobox->redirect[0] = 0;
	memobox->firstMemo = memobox->lastMemo = NULL;
}


void checkMemos(UserList *nick) {
	int unread, save, fwd, reply, i;
	MemoList *tmp;
	i=unread=save=fwd=reply=0;
	
	if(!nick->reg || !nick->reg->memos)
	  return;
	
	for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
		i++;
		if(tmp->flags & MEMO_UNREAD)
		  unread++;
		if(tmp->flags & MEMO_SAVE)
		  save++;
		if(tmp->flags & MEMO_FWD)
		  fwd++;
		if(tmp->flags & MEMO_REPLY)
		  reply++;
	}
	
	if(unread)
		uSend(MemoServ, nick, GetReply(RPL_YOUHAVEMEMOS, nick), unread, i);
}

void cleanMemos(UserList *nick) {
	MemoList *tmp;
	if(!nick->reg->memos)
	  return;
	if(nick->reg->memos->flags & MSELFCLEAN) {
		for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
			if(tmp->flags & MEMO_DELETE)
			  delMemo(nick->reg->memos, tmp);
		}
	}
}

void addMemo(MemoBox *mbox, MemoList *memo) {
	mostmemos++;
	if(!mbox->firstMemo) {
		mbox->firstMemo = memo;
		memo->previous = NULL;
	}
	else {
		mbox->lastMemo->next = memo;
		memo->previous = mbox->lastMemo;
	}
	mbox->lastMemo = memo;
	memo->next = NULL;
	indexMemoList(mbox);
}

void delMemo(MemoBox *mbox, MemoList *memo) {
	mostmemos--;
	if(memo->previous)
	  memo->previous->next = memo->next;
	else
	  mbox->firstMemo = memo->next;
	if(memo->next)
	  memo->next->previous = memo->previous;
	else
	  mbox->lastMemo = memo->previous;

	sfree(memo);
	indexMemoList(mbox);
}	

void indexMemoList(MemoBox *mbox) {
	MemoList *tmp;
	int i=1;
	for(tmp=mbox->firstMemo;tmp;tmp=tmp->next) {
		tmp->index = i;
		i++;
	}
	mbox->memos = (i-1);
}

/* yeah yeah, global variable, feh. */
char memos[20][NICKLEN];

void sendToMemoServ(UserList * nick, char **args, int numargs)
{
	char           *from = nick->nick;
	
        if(!strcmp(args[0], "\001PING")) {
                  if(numargs<3)
                       uSend(MemoServ, nick, GetReply(RPL_PING1, nick), args[1]);      
                  else
                       uSend(MemoServ, nick, GetReply(RPL_PING2, nick), args[1], args[2]);      
                  return;
        }

	if(!strcasecmp(args[0], "HELP") && numargs == 1) {
		sendHelp(from, "memoserv", NULL, 0);
		return;
	}
	else if(!strcasecmp(args[0], "HELP")) {
		sendHelp(from, "memoserv", args, numargs);
		return;
	}
	if(((nick->reg && nick->reg->memos) && (nick->reg->memos->flags & MSECURE) && nick->caccess < 3)) {
                uSend(MemoServ, nick, GetReply(ERR_ACCESSDENIED, nick));      
		return;
	}
	else if(nick->reg && !nick->reg->flags&NACTIVE) {
                uSend(MemoServ, nick, GetReply(ERR_NEEDACTNICK, nick));      
		return;
	}
	else if(nick->caccess < 2) {
                uSend(MemoServ, nick, GetReply(ERR_ACCESSDENIED, nick));      
		return;
	}
	else {
		char *blech = args[1];
		int i=0, x=0;

		if (!nick->reg->memos) {
			MemoBox *odd;
			odd = smalloc(sizeof(MemoBox));
			initMemoBox(odd);
			nick->reg->memos = odd;
		}

		if(blech) {
			bzero(memos, 20*NICKLEN);
			while((*blech) && (x < 20)) {
				if(*blech == ',') {
					memos[x][i] = 0;
					x++;
					i=0;
					blech++;
				}
				else if(!*blech) {
					memos[x][i] = 0;
					x++;
					memos[x][0] = 0;
					break;
				}
				else {
					memos[x][i] = *blech;
					i++;
					blech++;
				}
			}
		}
			
		msgtabCheck(MemoServ, nick, args, numargs); /* New messagetables to replace old code -GZ */
	
	}
}

void MreadCmd(UserList *nick, char **args, int numargs) {
	int i;
	MemoList *tmp;
	struct tm *time;
	char temp[30];

	if(!strcasecmp(args[1], "all")) {
		for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
			time=localtime(&tmp->sent);
			strftime(temp, 30, "%a %d/%m/%Y %T %Z", time);
	                uSend(MemoServ, nick, GetReply(RPL_MEMO1, nick), tmp->index, temp, 
				tmp->flags & MEMO_DELETE ? "Deleted" : tmp->flags & MEMO_UNREAD ? "New" : "Read");      
	                uSend(MemoServ, nick, GetReply(RPL_MEMO2, nick), tmp->from, tmp->memo);      
			tmp->flags &= ~MEMO_UNREAD;
		}
	}
	
	else {	
		int y;
		for(y=0;memos[y][0];y++) {
			i=atoi(memos[y]);
			if(i > nick->reg->memos->memos || i <= 0) {
		                uSend(MemoServ, nick, GetReply(ERR_NOSUCHMEMO, nick), i);      
				return;
			}
			for(tmp = nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
				if(tmp->index == i)
				  break;
			}

			time=localtime(&tmp->sent);
			strftime(temp, 30, "%a %d/%m/%Y %T %Z", time);
	                uSend(MemoServ, nick, GetReply(RPL_MEMO1, nick), tmp->index, temp, 
				tmp->flags & MEMO_DELETE ? "Deleted" : tmp->flags & MEMO_UNREAD ? "New" : "Read");      
	                uSend(MemoServ, nick, GetReply(RPL_MEMO2, nick), tmp->from, tmp->memo);      
			tmp->flags &= ~MEMO_UNREAD;
		}
	}
}

void MlistCmd(UserList *nick, char **args, int numargs) {
	MemoList *tmp;
	struct tm *time;
	char temp[30];
	char search[NICKLEN+2];
	if(numargs < 2)
		bzero(search, NICKLEN+2);
	else
		strcpy(search, args[1]);

	uSend(MemoServ, nick, GetReply(RPL_MLISTHEADER, nick), nick->nick);
	
	if(nick->reg->memos->firstMemo) {
		for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
			if(search[0]) {
				if(!match(tmp->from, search)) {
					time=localtime(&tmp->sent);
					strftime(temp, 30, "%a %d/%m/%Y %T %Z", time);
	                		uSend(MemoServ, nick, GetReply(RPL_MLIST1, nick), tmp->index, temp, tmp->from,
						tmp->flags & MEMO_DELETE ? "Deleted" : tmp->flags & MEMO_UNREAD ? "New" : "Read");
				}
			}
			else {
				time=localtime(&tmp->sent);
				strftime(temp, 30, "%a %d/%m/%Y %T %Z", time);
	                		uSend(MemoServ, nick, GetReply(RPL_MLIST1, nick), tmp->index, temp, tmp->from,
						tmp->flags & MEMO_DELETE ? "Deleted" : tmp->flags & MEMO_UNREAD ? "New" : "Read");
			}
		}
	}
	else
		uSend(MemoServ, nick, GetReply(ERR_MLISTEMPTY, nick));

	uSend(MemoServ, nick, GetReply(RPL_MLISTFOOTER, nick));

}
	
void MsendCmd(UserList *nick, char **args, int numargs) {			
	char *from = nick->nick;
	RegNickList *sendto;
	MemoList *send;
	UserList *unf;
	char memo[512]; /* sanity */
	int i;
	char sendnick[NICKLEN];

	sendto = getRegNickData(args[1]);
	
	if(!sendto || !sendto->flags&NACTIVE) {
		uSend(MemoServ, nick, GetReply(ERR_NICKNOTREGGED, nick), args[1]);
		return;
	}
	
	if(!sendto->memos) {
		sendto->memos = smalloc(sizeof(MemoBox));
		initMemoBox(sendto->memos);
	}	
	
	send = smalloc(sizeof(MemoList));
	strcpy(send->from, from);
	send->memo[0] = 0;
	send->sent = CTime;
	send->flags |= MEMO_UNREAD;
	strcpy(memo, args[2]);
	for(i=3;i<numargs;i++) {
		strcat(memo, " ");
		strcat(memo, args[i]);
	}

	if(strlen(memo) >= MEMOLEN) {
		uSend(MemoServ, nick, GetReply(ERR_MEMOTOOLONG, nick), MEMOLEN -1 );
		sfree(send);
		return;
	}

	strcpy(send->memo, memo);

	strcpy(sendnick,sendto->nick);

	if(sendto->memos->flags & MFORWARDED) {
		RegNickList *fwd_nick = getRegNickData(sendto->memos->redirect);
		if (!fwd_nick) {
			sendto->memos->flags &= ~MFORWARDED;
			sendto->memos->redirect[0] = 0;
		}
		else {
			if (!fwd_nick->memos) {
				fwd_nick->memos = smalloc(sizeof(MemoBox));
				initMemoBox(fwd_nick->memos);
			}

			if ((!(fwd_nick->memos->flags & MNOFORWARD)) && (!(fwd_nick->memos->flags & MNOMEMO))) {
				sendto = fwd_nick;
				send->flags |= MEMO_FWD;
			}
		}

	}

	if(sendto->memos->flags & MNOMEMO) {
		uSend(MemoServ, nick, GetReply(ERR_ISNOMEMO, nick), sendnick );
		sfree(send);
		return;
	}

	if(sendto->memos->memos >= sendto->memos->max) {
		uSend(MemoServ, nick, GetReply(ERR_MEMOBOXFULL, nick), sendnick );
		sfree(send);
		return;
	}
	if ((CTime - nick->reg->memos->lastsent) > 3600L) {
		nick->reg->memos->lastsent = CTime;
		nick->reg->memos->sent = 0;
	}

	if((nick->reg->memos->sent >= MS_SENT_MAX) && !isOper(nick)){
		uSend(MemoServ, nick, GetReply(ERR_TOOMANYMEMOS, nick));
		sfree(send);
		return;
	}

	addMemo(sendto->memos, send);
	unf=getNickData(sendto->nick);
	if(unf && unf->caccess >= 2)
		uSend(MemoServ, unf, GetReply(RPL_GOTNEWMEMO, unf), from, send->index);
	nick->reg->memos->sent++;
	uSend(MemoServ, nick, GetReply(RPL_SENTMEMO, nick), sendnick);
}

void MforwardCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	MemoList *send, *memo;
	RegNickList *sendto;
	UserList *unf;
	char sendnick[NICKLEN];
	char tempmemo[MEMOLEN];
	int sent_cnt = 0;

	if(!strcasecmp(args[2], from)) {
		uSend(MemoServ, nick, GetReply(ERR_NOFWDTOSELF, nick));
		return;
	}
	
	if(!nick->reg->memos->firstMemo) {
		uSend(MemoServ, nick, GetReply(ERR_MLISTEMPTY, nick));
		return;
	}

	sendto = getRegNickData(args[2]);

	if (!sendto || !sendto->flags&NACTIVE) {
		uSend(MemoServ, nick, GetReply(ERR_NICKNOTREGGED, nick), args[2]);
		return;
	}	

	if(!sendto->memos) {
		sendto->memos = smalloc(sizeof(MemoBox));
		initMemoBox(sendto->memos);
	}	

	strcpy(sendnick,sendto->nick);

	if(sendto->memos->flags & MFORWARDED) {
		RegNickList *fwd_nick = getRegNickData(sendto->memos->redirect);
		if (!fwd_nick) {
			sendto->memos->flags &= ~MFORWARDED;
			sendto->memos->redirect[0] = 0;
		}
		else {
			if (!fwd_nick->memos) {
				fwd_nick->memos = smalloc(sizeof(MemoBox));
				initMemoBox(fwd_nick->memos);
			}

			if ((!(fwd_nick->memos->flags & MNOFORWARD)) && (!(fwd_nick->memos->flags & MNOMEMO))) {
				sendto = fwd_nick;
			}
		}

	}

	if((sendto->memos->flags & MNOFORWARD) || (sendto->memos->flags & MNOMEMO)) {
		uSend(MemoServ, nick, GetReply(ERR_ISNOFORWARD, nick), sendnick);
		return;
	}

	if ((CTime - nick->reg->memos->lastsent) > 3600L) {
		nick->reg->memos->lastsent = CTime;
		nick->reg->memos->sent = 0;
	}
	
	if((nick->reg->memos->sent >= MS_SENT_MAX) && !isOper(nick)) {
		uSend(MemoServ, nick, GetReply(ERR_TOOMANYMEMOS, nick));
		return;
	}

	if(!strcasecmp(args[1], "all")) {
		for(memo=nick->reg->memos->firstMemo;memo;memo=memo->next) {

			if(sendto->memos->memos >= sendto->memos->max) {
				uSend(MemoServ, nick, GetReply(ERR_MEMOBOXFULL, nick), sendnick);
				break;
			}

			send = smalloc(sizeof(MemoList));
			strcpy(send->from, memo->from);
			send->sent = CTime;
			send->flags |= MEMO_UNREAD|MEMO_FWD;
			strcpy(tempmemo, memo->memo);
			tempmemo[MEMOLEN-NICKLEN-7] = 0;
			sprintf(send->memo,"FWD:%s > %s",from, tempmemo);
			addMemo(sendto->memos, send);
			sent_cnt++;
		}
	}
	else {	
		int i,y;
		for(y=0;memos[y][0];y++) {
			i=atoi(memos[y]);
			if(i > nick->reg->memos->memos || i <= 0) {
				uSend(MemoServ, nick, GetReply(ERR_NOSUCHMEMO, nick), i);
				return;
			}
			for(memo = nick->reg->memos->firstMemo;memo;memo=memo->next) {
				if(memo->index == i)
				  break;
			}
			
			if(sendto->memos->memos >= sendto->memos->max) {
				uSend(MemoServ, nick, GetReply(ERR_MEMOBOXFULL, nick), sendnick);
				break;
			}

			send = smalloc(sizeof(MemoList));
			strcpy(send->from, memo->from);
			send->sent = CTime;
			send->flags |= MEMO_UNREAD|MEMO_FWD;
			strcpy(tempmemo, memo->memo);
			tempmemo[MEMOLEN-NICKLEN-7] = 0;
			sprintf(send->memo,"FWD:%s > %s",from, tempmemo);
			addMemo(sendto->memos, send);
			sent_cnt++;
		}
	}
	if (sent_cnt) {
		nick->reg->memos->sent++;
		unf=getNickData(sendto->nick);
		if(unf && unf->caccess >= 2)
			uSend(MemoServ, unf, GetReply(RPL_GOTNEWMEMO, unf), from, send->index);
			uSend(MemoServ, nick, GetReply(RPL_MEMOFORWARDED, nick), sent_cnt, sent_cnt == 1 ? "" : "s", sendnick);
	}
}

void MdeleteCmd(UserList *nick, char **args, int numargs) {		
	MemoList *tmp;

	if(!strcasecmp(args[1], "all")) {
		for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
			tmp->flags |= MEMO_DELETE;
		}
		uSend(MemoServ, nick, GetReply(RPL_DELMEMO, nick), "ALL");
	}
	else if(!strcasecmp(args[1], "new")) {
		for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
			if(tmp->flags & MEMO_UNREAD) {
				tmp->flags |= MEMO_DELETE;
			}
		}
		uSend(MemoServ, nick, GetReply(RPL_DELMEMO, nick), "ALL UNREAD");
	}
	else if(!strcasecmp(args[1], "old") || !strcasecmp(args[1], "read")) {
		for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
			if(!(tmp->flags & MEMO_UNREAD))
			  tmp->flags |= MEMO_DELETE;
		}
		uSend(MemoServ, nick, GetReply(RPL_DELMEMO, nick), "ALL READ");
	}
	else {
		int y,i;
		for(y=0;memos[y][0];y++) {
			i = atoi(memos[y]);
			if(i<=0 || i > nick->reg->memos->memos)
				uSend(MemoServ, nick, GetReply(ERR_NOSUCHMEMO, nick), i);
			else {
				for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
					if(tmp->index == i) {
						tmp->flags |= MEMO_DELETE;
						break;
					}
				}
				uSend(MemoServ, nick, GetReply(RPL_DELMEMONR, nick), i);
			}
		}
	}

	McleanCmd(nick, args, numargs);
}

void McleanCmd(UserList *nick, char **args, int numargs) {
	MemoList *tmp;
	int i,x;
	
	i=x=0;
	
	for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next,x++) {
		if(tmp->flags & MEMO_DELETE) {
			delMemo(nick->reg->memos, tmp);
			i++;
		}
	}
}


void MsetCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	
	if (!strcasecmp(args[1], "secure")) {
		if(numargs<3)
			uSend(MemoServ, nick, GetReply(RPL_NSETIS, nick), "Secure Memos", nick->reg->memos->flags & MSECURE ? "ON" : "OFF");
		else if (!strcasecmp(args[2], "on")) {
			nick->reg->memos->flags |= MSECURE;
			uSend(MemoServ, nick, GetReply(RPL_NSETNOW, nick), "Secure Memos", "ON");
		}
		else if (!strcasecmp(args[2], "off")) {
			nick->reg->memos->flags &= ~MSECURE;
			uSend(MemoServ, nick, GetReply(RPL_NSETNOW, nick), "Secure Memos", "OFF");
		}
		else {
			uSend(MemoServ, nick, GetReply(ERR_BADSETTING, nick), args[2]);
			uSend(MemoServ, nick, GetReply(RPL_MOREHELPCMD, nick), MemoServ, "SET SECURE");
		}
	}
	else if (!strcasecmp(args[1], "nomemo")) {
		if(numargs<3)
			uSend(MemoServ, nick, GetReply(RPL_NSETIS, nick), "Memo Reception", nick->reg->memos->flags & MNOMEMO  ? "OFF" : "ON");
		else if (!strcasecmp(args[2], "on")) {
			nick->reg->memos->flags |= MNOMEMO;
			uSend(MemoServ, nick, GetReply(RPL_NSETNOW, nick), "Memo Reception", "OFF");
		}
		else if (!strcasecmp(args[2], "off")) {
			nick->reg->memos->flags &= ~MNOMEMO;
			uSend(MemoServ, nick, GetReply(RPL_NSETNOW, nick), "Memo Reception", "ON");
		}
		else {
			uSend(MemoServ, nick, GetReply(ERR_BADSETTING, nick), args[2]);
			uSend(MemoServ, nick, GetReply(RPL_MOREHELPCMD, nick), MemoServ, "SET NOMEMO");
		}
	}
	else if (!strcasecmp(args[1], "max")) {
		if(numargs<3)
			uSend(MemoServ, nick, GetReply(RPL_MMAX, nick), "Maximum Memos", nick->reg->memos->max);
		else {
			int i = atoi(args[2]);
			if(i <= 0 || i > MS_RCV_MAX_HIGHEST)
				uSend(MemoServ, nick, GetReply(ERR_MMAXSET, nick), MS_RCV_MAX_HIGHEST);
 			else {
				nick->reg->memos->max = i;
				uSend(MemoServ, nick, GetReply(RPL_MMAXNOW, nick), "Maximum Memos", nick->reg->memos->max);
			}
		}
	}
	else if (!strcasecmp(args[1],"forward")) {
		if (numargs<3) {
			if ((nick->reg->memos->flags & MFORWARDED) && (getRegNickData(nick->reg->memos->redirect)))
				uSend(MemoServ, nick, GetReply(RPL_MFORWARD, nick), nick->reg->memos->redirect);
			else
				uSend(MemoServ, nick, GetReply(RPL_MFORWARD, nick), "NOBODY");
		}
		else {
			if (!strcasecmp(args[2],from)) {
				nick->reg->memos->flags &= ~MFORWARDED;
				nick->reg->memos->redirect[0] = 0;
				uSend(MemoServ, nick, GetReply(RPL_MFORWARDNOW, nick), "NOBODY");
			}
			else {
				RegNickList *fwd_nick = getRegNickData(args[2]);
				if ((!fwd_nick) || (!(fwd_nick->flags & NACTIVE)))
					uSend(MemoServ, nick, GetReply(ERR_NICKNOTREGGED, nick), args[2]);
				else {
					if (!fwd_nick->memos) {
						fwd_nick->memos = smalloc(sizeof(MemoBox));
						initMemoBox(fwd_nick->memos);
					}
					if ((fwd_nick->memos->flags & MNOFORWARD) || (fwd_nick->memos->flags & MNOMEMO))
						uSend(MemoServ, nick, GetReply(ERR_NOFWDMEMOS, nick), fwd_nick->nick);
					else {	
						nick->reg->memos->flags |= MFORWARDED;
						strcpy(nick->reg->memos->redirect,fwd_nick->nick);
						uSend(MemoServ, nick, GetReply(RPL_MFORWARDNOW, nick), nick->reg->memos->redirect);
					}
				}
			}
		}
	}		
	else if (!strcasecmp(args[1],"nofwd")) {
		if (numargs < 3)
			uSend(MemoServ, nick, GetReply(RPL_NSETIS, nick), "Accept No Forwards", nick->reg->memos->flags & MNOFORWARD ? "ON" : "OFF");

		else if (!strcasecmp(args[2], "on")) {
			nick->reg->memos->flags |= MNOFORWARD;
			uSend(MemoServ, nick, GetReply(RPL_NSETNOW, nick), "Accept No Forwards", "ON");
		}
		else if (!strcasecmp(args[2], "off")) {
			nick->reg->memos->flags &= ~MNOFORWARD;
			uSend(MemoServ, nick, GetReply(RPL_NSETNOW, nick), "Accept No Forwards", "OFF");
		}
		else {
			uSend(MemoServ, nick, GetReply(ERR_BADSETTING, nick), args[2]);
			uSend(MemoServ, nick, GetReply(RPL_MOREHELPCMD, nick), MemoServ, "SET NOFWD");
		}
	}
	else {
			uSend(MemoServ, nick, GetReply(ERR_BADSETTING, nick), args[1]);
			uSend(MemoServ, nick, GetReply(RPL_MOREHELPCMD, nick), MemoServ, "SET");
	}
}

void MretrieveCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	char memobuffer[60000]; /* fill this up? not likely */
	char buffer[512], temp[31]; /* temporary, don't ask... */
	char sendfrom[USERLEN + HOSTLEN];
	MemoList *tmp;
	struct tm *time;
	int type=0, x;
	
	bzero(memobuffer, 60000);
	sprintf(sendfrom, "%s@%s", MemoServ, NETWORK);
	
	if(numargs<2)
	  type = 0;
	else {
		if(!strcasecmp(args[1], "all"))
		  type=0;
		if(!strcasecmp(args[1], "read"))
		  type = 1;
		else if(!strcasecmp(args[1], "save"))
		  type = 2;
		else if(!strcasecmp(args[1], "unread"))
		  type = 3;
	}
	
	sprintf(memobuffer, "\nMemos for %s:\n", from);
	for(tmp=nick->reg->memos->firstMemo;tmp;tmp=tmp->next) {
		x=0;
		if(type) {
			switch(type) {
			 case 1:
				if(!(tmp->flags & MEMO_UNREAD))
				  x=1;
				break;
			 case 2:
				if(tmp->flags & MEMO_SAVE)
				  x=1;
				break;
			 case 3:
				if(tmp->flags & MEMO_UNREAD)
				  x=1;
				break;
			}
		}
		else
		  x=1;
		
		sprintf(buffer, "Memo (%3i) from: %s\n", tmp->index, tmp->from);
		strcat(memobuffer, buffer);
		time=localtime(&tmp->sent);
		strftime(temp, 30, "%a %d/%m/%Y %T %Z", time);
		sprintf(buffer, "Sent at: %-30s  Flagged %c%c%c%c%c\n", temp, tmp->flags & MEMO_UNREAD ? '*' : '-', tmp->flags & MEMO_DELETE ? 'D' : ' ', tmp->flags & MEMO_SAVE ? 'S' : ' ', tmp->flags & MEMO_FWD ? 'f' : ' ', tmp->flags & MEMO_REPLY ? 'r' : ' ');
		strcat(memobuffer, buffer);
		sprintf(buffer, "%s\n\n", tmp->memo);
		strcat(memobuffer, buffer);
	}
	
	sendMail(nick->reg->email, sendfrom, "Retreived memos:", memobuffer);
	uSend(MemoServ, nick, GetReply(RPL_MEMOSRETRIEVED, nick), nick->reg->email);
}

void McsendCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	int i=0,x=0,lev=0;
	int mfirstword = 2;
	RegChanList *chan;
	cAccessList *tmp;
	MemoList *send;
	UserList *unf;
	RegNickList *snick;
	char memo[512];
	char sfrom[NICKLEN+CHANLEN];
	int fwd_flag;
	
	chan = getRegChanData(args[1]);
	
	if(!chan) {
		uSend(MemoServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}
	
	if(1 > getChanOp(chan, from)) {
		uSend(MemoServ, nick, GetReply(ERR_ACCESSDENIED, nick));
		return;
	}

	if ((args[2][0] >= '0') && (args[2][0] <= '9')) {
		lev = atoi(args[2]);
		if ((lev < 0) || (lev > FOUNDER)) {
			uSend(MemoServ, nick, GetReply(ERR_MNEEDLEVEL, nick));
			return;
		}
		mfirstword = 3;
	}
	
	sprintf(sfrom, "%s/%s", chan->name, from);
	sprintf(memo, "%i+ >> %s", lev, args[mfirstword]);
	for(i=mfirstword+1;i<numargs;i++) {
		strcat(memo, " ");
		strcat(memo, args[i]);
	}
	
	if(strlen(memo) >= (MEMOLEN-CHANLEN + 2)) {
		uSend(MemoServ, nick, GetReply(ERR_MEMOTOOLONG, nick), MEMOLEN - CHANLEN - 3 );
		return;
	}

	if ((CTime - nick->reg->memos->lastsent) > 3600L) {
		nick->reg->memos->sent = 0;
		nick->reg->memos->lastsent = CTime;
	}

	if((nick->reg->memos->sent >= MS_SENT_MAX) && !isOper(nick)) {
		uSend(MemoServ, nick, GetReply(ERR_TOOMANYMEMOS, nick));
		return;
	}

	i=0;
	for(tmp=chan->firstOp;tmp;tmp=tmp->next) {
		if(tmp->uflags >= lev && match(tmp->nick, from)) {
			i++;

			snick = getRegNickData(tmp->nick);
			if (!snick)
				continue;

			if (!snick->memos) {
				snick->memos = smalloc(sizeof(MemoBox));
				initMemoBox(snick->memos);
			}	

			fwd_flag = 0;
			if (snick->memos->flags & MFORWARDED) {
				RegNickList *fwd_nick = getRegNickData(snick->memos->redirect);
				if (!fwd_nick) {
					snick->memos->flags &= ~MFORWARDED;
					snick->memos->redirect[0] = 0;
				}
				else {
					if (!fwd_nick->memos) {
						fwd_nick->memos = smalloc(sizeof(MemoBox));
						initMemoBox(fwd_nick->memos);
					}
					if (!fwd_nick->memos->flags & (MNOFORWARD|MNOMEMO)) {
						snick = fwd_nick;
						fwd_flag = 1;
					}
				}
			}

			if ((snick->memos->memos >= snick->memos->max) || (snick->memos->flags & MNOMEMO))
				continue;

			x++;
			send = smalloc(sizeof(MemoList));
			strcpy(send->memo, memo);
			strcpy(send->from, sfrom);
			send->flags |= MEMO_UNREAD;
			if (fwd_flag)
				send->flags |= MEMO_FWD;
			send->sent = CTime;
			addMemo(snick->memos, send);
			unf=getNickData(snick->nick);
			if(unf && unf->caccess >= 2)
				uSend(MemoServ, unf, GetReply(RPL_GOTNEWMEMO, unf), from, send->index);
		}
	}

	if (x) {
		nick->reg->memos->sent++;
		uSend(MemoServ, nick, GetReply(RPL_SENTMEMO, nick), chan->name);
	}

	uSend(MemoServ, nick, GetReply(RPL_SENTMEMOCHAN, nick), x, i);
}
			
	
