/*
 * Copyright (c) 1999-2000 GZ
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include "services.h"

int  HJCheck(char *);

long getkilltime(char *s)
{
/* THIS FUNCTION RETURNS A TIME IN SECONDS BASED ON INPUT STRING
   FORMATS ALLOWED:
		<value>h = hours
		<value>d = days
		<value>w = weeks
   MISSING LETTER IS ASSUMED TO MEAN MINUTES (BACKWARDS COMPATIBLE)
   MISSING VALUE WILL BE ASSUMED AS 1.

   THIS FUNCTION IS USED BY Oautokill and Oignore FOR TIME COMPUTATION */

   long base_mult;
   char mult_char[2];
   int value_width;
   int base_value;

   base_mult = 60L; /* INITIALIZE TO CONVERT MINS TO SECS */
   strcpy(mult_char,"m");
   base_value = 1;

   if ((!strcasecmp(s,"add")) || (!(*s))) /* COMPATIBILITY FOR OLD FUNCTION */
      return 0;

   value_width = strlen(s) - 1;

   if (isalpha(s[value_width]))  /* IF THERE'S A LETTER ON THE END, PARSE IT */
   {
      mult_char[0] = s[value_width];

      if (!strcasecmp(mult_char,"h"))
          base_mult = 3600L;
      else if (!strcasecmp(mult_char,"d"))
          base_mult = 86400L;
      else if (!strcasecmp(mult_char,"w"))
          base_mult = 604800L;

      if (value_width > 0)
      {
         base_value = atoi(s);
         if (base_value < 1)
           base_value = 1;
      }
   }
   else
   {
      base_value = atoi(s);
      if (base_value < 0)
        base_value = 0;
   }

   return((long)(base_value) * base_mult);
}
   
char *expand_time(long t)
{
/* EXPANDS TIME IN SECONDS TO FULLY QUALIFIED STRING 
   USED BY AUTOKILL FUNCTIONS FOR DISPLAY */

   static char retstr[128];
   int temp1;

   *retstr = 0;

   if (t == 0)
   {
      strcpy(retstr,"forever");
   }
   else
   {
      if (t >= 604800L)
      {
        temp1 = (int)(t / 604800L);
        t = t % 604800L;
        sprintf(retstr,"%d week%s",temp1, (temp1 == 1 ? "":"s"));
      }
      if (t >= 86400L)
      {
        temp1 = (int)(t / 86400L);
        t = t % 86400L;
        if (*retstr)
           strcat(retstr,", ");
        sprintf(retstr,"%s%d day%s",retstr, temp1, (temp1 == 1 ? "":"s"));
      }
      if (t >= 3600L)
      {
        temp1 = (int)(t / 3600L);
        t = t % 3600L;
        if (*retstr)
          strcat(retstr,", ");
        sprintf(retstr,"%s%d hour%s",retstr, temp1, (temp1 == 1 ? "":"s"));
      }

      temp1 = (int)(t / 60L);

      if ((*retstr) && (temp1 > 0))
        strcat(retstr,", ");

      if ((!(*retstr)) && (temp1 == 0))
         temp1 = 1;

      if (temp1)    
         sprintf(retstr,"%s%d minute%s", retstr, temp1, (temp1 == 1 ? "":"s"));
   }

   return((char *)retstr);
}
   
void sendToOperServ(UserList * nick, char **args, int numargs)
{
	char           *from = nick->nick;
	
	if (!strcmp(args[0], "\001PING")) {
		if (numargs < 3)
			uSend(OperServ, nick, GetReply(RPL_PING1, nick), args[1]);
		else
			uSend(OperServ, nick, GetReply(RPL_PING2, nick), args[1], args[2]);
		return;
	}
	if (!isOper(nick)) {
		uSend(OperServ, nick, GetReply(ERR_NEEDOPER, nick));
		return;
	}

	if (!strcasecmp(args[0], "help") && numargs > 1) {
		sendHelp(from, OperServ, args, numargs);
		return;
	} else if (!strcasecmp(args[0], "help")) {
		sendHelp(from, OperServ, NULL, 0);
		return;
	}

	if(!match(args[0], "a*kill"))
		OautokillCmd(A_AKILL, nick, args, numargs);
	else if(!strcasecmp(args[0], "ignore"))
		OautokillCmd(A_IGNORE, nick, args, numargs);

	else
		msgtabCheck(OperServ, nick, args, numargs); /* New messagetables to replace old code -GZ */
}

void OautokillCmd(int type, UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	long t;
	int i;
	char akreason[80];
	char akills[9];

	if (type == A_AKILL)
		strcpy(akills, "Autokill");
	else
		strcpy(akills, "Ignore");
	
	if (numargs < 2) {
                listAkills(from, type, NULL);
		return;
	}
	if (!strcasecmp(args[1], "list")) {
		if(numargs < 3)
	                listAkills(from, type, NULL);
		else	
			listAkills(from, type, args[2]);
		return;
	}

	if (!isServop(nick) && !isRoot(nick)) {
		uSend(OperServ, nick, GetReply(ERR_NEEDSERVOP, nick));
		return;
	}

	if (numargs < 3) {
		uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, akills);
		return;
	}

	if (args[1][0] == '-' || !strcasecmp(args[1], "del") || !strcasecmp(args[1], "delete")) {
		if (!check_mask(args[2],0)) {
			uSend(OperServ, nick, GetReply(ERR_BADMASK, nick), args[2]);
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, akills);
			return;
		}
		removeakill(from, args[2]);
		return;
	}
	if (numargs < 4) {
		uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, akills);
		return;
	} 
	else {
		t = getkilltime(args[1]);
		if (!check_mask(args[2],0)) {
			uSend(OperServ, nick, GetReply(ERR_BADMASK, nick), args[2]);
			return;
		}

		if (findakill(args[2]))
		{
			uSend(OperServ, nick, GetReply(ERR_MASKEXISTS, nick), args[2], akills);
			return;
		}

		bzero(akreason,80);
		strncpy(akreason, args[3], 79);
		for (i = 4; i < numargs; i++)
		{
			if ((strlen(akreason) + strlen(args[i])) > 78)
				break;
			strcat(akreason, " ");
			strcat(akreason, args[i]);
		}

		uSend(OperServ, nick, GetReply(RPL_AKILLADDED, nick), akills, args[2], expand_time(t), akreason);
		logDump(operlog,"%s!%s@%s added %s for %s for %s [%s]", from, nick->user, nick->host, akills, args[2], expand_time(t), akreason);
		addakill(t, args[2], from, type, akreason);
		return;
	}
}
  
void OmodeCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;

	int on=1, onarg = 3, i=0, valid = 1;
	ChanList *tmp = getChanData(args[1]);
	cBanList *tmpban;
	cNickList *tmpnick;
	char string[512];
	  
	if(!tmp) {
		uSend(OperServ, nick, GetReply(ERR_NOSUCHCHAN, nick), args[1]);
		return;
	}

	if ((args[2][0] != '+') && (args[2][0] != '-')) {
		uSend(OperServ, nick, GetReply(ERR_INVALIDMODES, nick), args[2]);
		return;
	}
  
	for(i=0;args[2][i];i++)
	{
		if (!valid)
			break;

		switch(args[2][i]) {
			case '+':
				on=1;
				break;
			case '-':
				on=0;
				break;
			case 'b':
				if (numargs > onarg) {
					if(on) {
						tmpban = smalloc(sizeof(cBanList));
						strcpy(tmpban->ban, args[onarg]);
						addChanBan(tmp, tmpban);
#ifdef CDEBUG
						sSend(":%s GNOTICE :Added ban %s on %s", myname, args[onarg], tmp->name);
#endif
					} else {
						tmpban = getChanBan(tmp, args[onarg]);
						if (tmpban) {
#ifdef CDEBUG
							sSend(":%s GNOTICE :Removed ban %s on %s", myname, args[onarg], tmp->name);
#endif
							delChanBan(tmp, tmpban);
						}
					}
						onarg++;
				} else
					valid = 0;
				break;
					 
			case 'i':
				if(on) {
					tmp->modes |= PM_I;
#ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s +i", myname, tmp->name);
#endif
				}
				else {
					tmp->modes -= PM_I;
#ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s -i", myname, tmp->name);
#endif
				}
				break;

			case 'l':
				 if(on) {
					if (numargs > onarg) {
						int chan_limit = atoi(args[onarg]);
						if (chan_limit > 0) {
							tmp->modes |= PM_L;
							if (tmp->reg)
								tmp->reg->limit = chan_limit;
#ifdef CDEBUG
							sSend(":%s GNOTICE :Set %s +l %lu", myname, tmp->name, tmp->reg->limit);
#endif
						}
						else
							valid = 0;
						onarg++;
					}								
				 }
				 else {
					 tmp->modes -= PM_L;
					 if(tmp->reg)
						tmp->reg->limit = 0;
#ifdef CDEBUG
					 sSend(":%s GNOTICE :Set %s -l", myname, tmp->name, tmp->reg->limit);
#endif					 
				 }
				 break;

			case 'k':
				if (numargs > onarg) {
					if(on) {
						tmp->modes |= PM_K;
						if (tmp->reg)
							strcpy(tmp->reg->key,args[onarg]);
# ifdef CDEBUG
						sSend(":%s GNOTICE :Set %s +k %s", myname, tmp->name, tmp->reg->key);
# endif
					}
					else {
						tmp->modes -= PM_K;
						if (tmp->reg)
							tmp->reg->key[0] = '\0';
# ifdef CDEBUG
						sSend(":%s GNOTICE :Set %s -k %s", myname, tmp->reg->name, args[onarg]);
# endif
					}
					onarg++;
				}
				else
					valid = 0;
				break;

			case 'm':
				if(on) {
					tmp->modes |= PM_M;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s +m", myname, tmp->name);
# endif
				}
				else {
					tmp->modes -= PM_M;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s -m", myname, tmp->name);
# endif
				}
				break;

			case 'n':
				if(on) {
					tmp->modes |= PM_N;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s +n", myname, tmp->name);
# endif		 
				}
				else {
					tmp->modes -= PM_N;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s -m", myname, tmp->name);
# endif
				}
				break;

			case 'o':
				if(numargs > onarg) {
					if(on) {
						tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
						if (tmpnick) {
							tmpnick->op |= CHANOP;
# ifdef CDEBUG
							sSend(":%s GNOTICE :Oped %s in %s", myname, args[onarg], tmp->name);
# endif
						}
						else
							valid = 0;
					}
					else {
						tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
						if(tmpnick) {
							tmpnick->op &= ~CHANOP;
# ifdef CDEBUG
							sSend(":%s GNOTICE :DeOped %s in %s", myname, args[onarg], tmp->name);
# endif
						}
						else
							valid = 0;
					}
					onarg++;
				}
				else
					valid = 0;
				break;

			  case 'p':
				if(on) {
					tmp->modes |= PM_P;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s +p", myname, tmp->name);
# endif		 
				}
				else {
					tmp->modes -= PM_P;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s -p", myname, tmp->name);
# endif
				}
				break;

			case 's':
				if(on) {
					if ((tmp->reg) && isServop(nick))
						tmp->reg->flags |= CPRIVATE;
					tmp->modes |= PM_S;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s +s", myname, tmp->name);
# endif		 
				}
				else {
					if ((tmp->reg) && isServop(nick))
						tmp->reg->flags &= ~CPRIVATE;

					tmp->modes -= PM_S;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s -s", myname, tmp->name);
# endif
				}
				break;

			case 't':
				if(on) {
					tmp->modes |= PM_T;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s +t", myname, tmp->name);
# endif		 
				}
				else {
					tmp->modes -= PM_T;
# ifdef CDEBUG
					sSend(":%s GNOTICE :Set %s -t", myname, tmp->name);
# endif
				}
				break;

			case 'v':
				if(numargs > onarg) {
					if(on) {
						tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
						if (tmpnick) {
							tmpnick->op |= CHANVOICE;
# ifdef CDEBUG
							sSend(":%s GNOTICE :Voiced %s in %s", myname, args[onarg], tmp->name);
# endif
						}
						else
							valid = 0;
					}
					else {
						tmpnick = getChanUserData(tmp, getNickData(args[onarg]));
						if(tmpnick) {
							tmpnick->op &= ~CHANVOICE;
# ifdef CDEBUG
							sSend(":%s GNOTICE :DeVoiced %s in %s", myname, args[onarg], tmp->name);
# endif
						}
						else
							valid = 0;
					}
					onarg++;
				}
				else
					valid = 0;
				break;
		}	 
	}

	if (!valid) {
		uSend(OperServ, nick, GetReply(ERR_INVALIDMODES, nick), args[2]);
		return;
	}

	string[0] = 0;
	for(i=2;i<onarg && args[i];i++) {
		strcat(string, args[i]);
		strcat(string, " ");
	}
	string[strlen(string)-1] = 0;

	sSend(":%s MODE %s %s", OperServ, args[1], string);
	sSend(":%s GLOBOPS :%s MODE %s (%s)", OperServ, from, args[1], string);
	logDump(operlog,"%s!%s@%s MODE %s (%s)", from, nick->user, nick->host, args[1], string);
} 

void OrawCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	char            msg[512];
	int             i = 2;
	
	if(!args[1])
		return;
	strcpy(msg, args[1]);
	while (i < numargs) {
		strcat(msg, " ");
		strcat(msg, args[i]);
		i++;
	}
	sSend("%s", msg);
        logDump(rootlog,"%s!%s@%s RAW (%s)", from, nick->user, nick->host, msg);
	return;
} 

void OinjectCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	char            msg[512];
	int             i = 2;
	
	if(!args[1])
		return;
	strcpy(msg, args[1]);
	while (i < numargs) {
		strcat(msg, " ");
		strcat(msg, args[i]);
		i++;
	}
	parseLine(msg);
        logDump(rootlog,"%s!%s@%s INJECT (%s)", from, nick->user, nick->host, msg);
	return;
} 

void OcountCmd(UserList *nick, char **args, int numargs) {
        char *from = nick->nick;

        if (args[1] && isdigit(*args[1])) {
        mostusers = atol(args[1]);
        }
        sSend(":%s GLOBOPS :%s reset the counter to %ld", OperServ, from, mostusers);
	logDump(operlog,"%s!%s@%s reset the counter to %ld", from, nick->user, nick->host, mostusers);
}

void OrestartCmd(UserList *nick, char **args, int numargs) {
        char *from = nick->nick;
        char msg[256];

        *msg = 0;

        if (numargs > 1) {
                int     i = 2;

                strcpy(msg," (");
                strcat(msg,args[1]);

                while (i < numargs) {
                        strcat(msg, " ");
                        strcat(msg, args[i]);
                        i++;
                }
                strcat(msg,")");
        }

        sSend(":%s GLOBOPS :RESTART From %s%s", myname, from, msg);
        logDump(operlog,"%s!%s@%s RESTART%s", from, nick->user, nick->host, msg);

        sSend(":%s PART #Services", NickServ);
        sSend(":%s PART #Services", ChanServ);
        sSend(":%s PART #Services", MemoServ);
        sSend(":%s PART #Services", OperServ);
        sSend(":%s PART #Services", WebServ);
        sSend(":%s PART #Services", GlobalMessenger);
        sSend(":%s PART %s", WebServ, WEBTVCHAN);
	uSend(OperServ, nick, GetReply(RPL_RESTARTING, nick));
        srestart();
}

void OshutdownCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	char msg[256];

	*msg = 0;

	if (numargs > 1) {
		int	i = 2;

		strcpy(msg," (");
		strcat(msg,args[1]);

		while (i < numargs) {
			strcat(msg, " ");
			strcat(msg, args[i]);
			i++;
		}
		strcat(msg,")");
	}
	
	sSend(":%s GLOBOPS :SHUTDOWN From %s%s", myname, from, msg);
	logDump(operlog,"%s!%s@%s SHUTDOWN%s", from, nick->user, nick->host, msg);

        sSend(":%s PART #Services", NickServ);
        sSend(":%s PART #Services", ChanServ);
        sSend(":%s PART #Services", MemoServ);
        sSend(":%s PART #Services", OperServ);
        sSend(":%s PART #Services", WebServ);
        sSend(":%s PART #Services", GlobalMessenger);
        sSend(":%s PART %s", WebServ, WEBTVCHAN);
	uSend(OperServ, nick, GetReply(RPL_SHUTTINGDOWN, nick));
	sshutdown(0);
} 

void OjupeCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
 
	if(CheckServer(args[1])) {
		uSend(OperServ, nick, GetReply(ERR_THISEXISTS, nick), args[1], "server");
		return;
	}

	AddServer(args[1], "Jupitered");
	sSend("SERVER %s 2 JupiteredServer :Jupitered by %s %s", args[1], from, ctime(&(CTime)));
	sSend(":%s GLOBOPS :Server %s Jupitered by %s at %s", myname, args[1], from, ctime(&(CTime)));
	logDump(operlog,"%s!%s@%s Jupitered Server %s", from, nick->user, nick->host, args[1]);
	return;
}
 
void ObroadcastCmd(UserList *nick, char **args, int numargs) {
        char *from = nick->nick;

        int i;

        char broadcast[256];
	memset(broadcast, '\0', 255);

        for(i=1;i < numargs;i++) {
            strcat(broadcast, args[i]);
            strcat(broadcast, " ");
        }

#ifndef UNREAL
        sSend(":%s NOTICE $*." NETWORK " :\002[GLOBAL NOTICE]\002 %s (Please do not reply.)", GlobalMessenger, broadcast);
#else
        sSend(":%s NOTICE $* :\002[GLOBAL NOTICE]\002 %s (Please do not reply.)", GlobalMessenger, broadcast);
#endif
        sSend(":%s GLOBOPS :%s used global broadcast.", myname, from);
        logDump(operlog,"%s used global broadcast (%s).", from, broadcast);
        return;
}

void OsvsnoopCmd(UserList *nick, char **args, int numargs) {
        char *from = nick->nick;

	aServer *server;
        UserList *tmp;

	if (!strcasecmp(args[1],"ADD")) {

		server = GetServer(args[2]);
		if(!server) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHSERVER, nick), args[2]);
			return;
		}

		if(server->svsnoop == 1) {
			uSend(OperServ, nick, GetReply(ERR_ALREADYSVSNOOP, nick), args[2]);
			return;
		}		

		server->svsnoop = 1;
	        sSend(":%s GLOBOPS :%s activated SVSNOOP for %s.", myname, from, args[2]);
		sSend(":%s SVSNOOP %s +", myname, server->name);

	        for(tmp=firstUser;tmp;tmp=tmp->next)
        	{
                	if (isOper(tmp) && !strcasecmp(tmp->server, server->name)) {
				setMode(tmp->nick, "-NnAaOo");
                	}
	        }
	}

	else if (!strcasecmp(args[1],"DEL")) {

		server = GetServer(args[2]);
		if(!server) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHSERVER, nick), args[2]);
			return;
		}

		if(server->svsnoop == 0) {
			uSend(OperServ, nick, GetReply(ERR_NOTSVNOOP, nick), args[2]);
			return;
		}

		server->svsnoop = 0;
	        sSend(":%s GLOBOPS :%s deactivated SVSNOOP for %s.", myname, from, args[2]);
		sSend(":%s SVSNOOP %s -", myname, server->name);
	}

        logDump(operlog,"%s used SVSNOOP (%s).", from, args[2]);
        return;
}

void OekillCmd(UserList *nick, char **args, int numargs) {
        char *from = nick->nick;

   	anEmail *emailptr;
    	anEmail *emailptr2;

	char ereason[80];

	int i;

	if(numargs < 3 || !strcasecmp(args[1],"LIST")) {

		if (EmailList == NULL) {
			uSend(OperServ, nick, GetReply(RPL_EKILLEMPTY, nick));
        		return;
		}

		uSend(OperServ, nick, GetReply(RPL_EKILLHEAD, nick));
		for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
	        	emailptr2 = emailptr->next;
			uSend(OperServ, nick, GetReply(RPL_EKILLDATA, nick), emailptr->email, emailptr->by, emailptr->reason, ctime(&(emailptr->time)));
    		}
		uSend(OperServ, nick, GetReply(RPL_EKILLEND, nick));
	return;

	}

        if((!strchr(args[2], '@')) || (strlen(args[2])<5) || (strlen(args[2])>EMAILLEN)) {
		uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
		return;
	}
	
	if (numargs < 3) {
		uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
		return;
	}

	if(!strcasecmp(args[1],"DEL")) {
		if(!CheckEmail(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHEKILL, nick), args[2]);
			return;
		}

		uSend(OperServ, nick, GetReply(RPL_EKILLDELED, nick), args[2]);
                sSend(":%s GLOBOPS :%s removed dynamic EKILL for %s", OperServ, from, args[2]);
		DeleteEmail(args[2]);
		return;
	}

	if (numargs < 4) {
		uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
		return;
	}

	if(!strcasecmp(args[1],"ADD")) {

		if(CheckEmail(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_EKILLEXISTS, nick), args[2]);
			return;
		}

 		bzero(ereason,80);
		strncpy(ereason, args[3], 79);

		for (i = 4; i < numargs; i++)
		{
			if ((strlen(ereason) + strlen(args[i])) > 78)
				break;
			strcat(ereason, " ");
			strcat(ereason, args[i]);
		}

		uSend(OperServ, nick, GetReply(RPL_EKILLADDED, nick), args[2]);
                sSend(":%s GLOBOPS :%s added dynamic EKILL for %s (%s)", OperServ, from, args[2], ereason);
		AddEmail(args[2], from, ereason);
	}

        return;
}

void OlistCmd(UserList *nick, char **args, int numargs) {
	HostCloneList *clonelist;
	int i = 0;

	if(!strcasecmp(args[1],"PERMIGNORES")) {
		uSend(OperServ, nick, GetReply(RPL_PERMCLONEIGNHEAD, nick));
	        for (clonelist = firstPermClone; clonelist; clonelist = clonelist->next) {
			i++;
			uSend(OperServ, nick, GetReply(RPL_PERMCLONEIGNDATA, nick), i, clonelist->host);
        	}
		uSend(OperServ, nick, GetReply(RPL_PERMCLONEIGNEND, nick));
	}
        return;
}

void OuptimeCmd(UserList *nick, char **args, int numargs) {
	long            blah;
	int             days, hours, mins, seconds;
	
	blah = time(NULL);
	blah -= startup;
	days = (int) (blah / (24 * 3600));
	blah %= (24 * 3600);
	hours = (int) (blah / (3600));
	blah %= (3600);
	mins = (int) (blah / (60));
	blah %= (60);
	seconds = blah;
	uSend(OperServ, nick, GetReply(RPL_UPTIME1, nick));
	uSend(OperServ, nick, GetReply(RPL_UPTIME2, nick), days, hours, mins, seconds);
	return;
}

void OallocstatCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	FILE *fp;
	
	char psline[128];

	uSend(OperServ, nick, GetReply(RPL_ALLOCSTAT1, nick), AllocCalled);
	uSend(OperServ, nick, GetReply(RPL_ALLOCSTAT2, nick), FreeCalled);
	
	/* Ax0n: Added -w for wrap, incase we are invoked /very/long/path/services */
	sprintf(psline, "ps -uaxw | grep services");
	fp = popen(psline, "r");
	if(fp) {
		while(sfgets(psline, 128, fp)) {
			if(match(psline, "*grep*"))
				sSend(":%s NOTICE %s :%s", OperServ, from, psline);
		}
		fclose(fp);
	}
	return;
}

void OsyncCmd(UserList *nick, char **args, int numargs) {
	saveakills();
        strcpy(args[0], "save");
	sendToNickServ(nick, args, numargs);
	sendToChanServ(nick, args, numargs);
	sSend(":%s GLOBOPS :Synchronizing all remaining databases to disk...", OperServ);
                saveakills();
                savepermignores();
                saveHostJinxes();
                savePermJinxes();
                savedynamicekills();
}

void OtriggerCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;

	if(index(args[1], '@')) {
		char user[USERLEN];
		char host[HOSTLEN];
		HostClone *hc;
		UserClone *uc;
		sscanf(args[1], "%[^@]@%s", user, host);
		hc = getCloneData(host);
		if(hc)
			uc = getUserCloneData(hc, user);
		else
			uc = NULL;
		if(uc && hc) {
			if(numargs<3)
				uSend(OperServ, nick, GetReply(RPL_CLONESTATUS1, nick), user, host, uc->clones, uc->trigger, (uc->roleplay == 1) ? "ROLEPLAY ON!" : "");
			else {
				int i=atoi(args[2]);
				if(i>0) {
					uc->trigger = i;
					sSend(":%s GLOBOPS :%s@%s retriggered to %i by %s", OperServ, user, host, i, from);
				}
				else
					uSend(OperServ, nick, GetReply(ERR_INVALIDTRIGLEVEL, nick));
			}
		}
		else
			uSend(OperServ, nick, GetReply(ERR_NOSUCHTRIGGER, nick), user, host);
	}
	else {
		char host[HOSTLEN];
		HostClone *hc;
		strcpy(host, args[1]);
		hc = getCloneData(host);
		if(hc) {
			if(numargs<3)
				uSend(OperServ, nick, GetReply(RPL_CLONESTATUS2, nick), host, hc->clones, hc->trigger);
			else {
				int i = atoi(args[2]);
				if(i>0) {						
					hc->trigger = i;
					sSend(":%s GLOBOPS :%s retriggered to %i by %s", OperServ, host, i, from);
				}
				else
					uSend(OperServ, nick, GetReply(ERR_INVALIDTRIGLEVEL, nick));
			}
		}
		else {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHCHAN, nick), host);
		}
	}
}

void OclonesetCmd(UserList *nick, char **args, int numargs) {
	char *from = nick->nick;
	HostClone *hc;
	HostCloneList *permignore;
	int tempflags = 0;
	int goodhost = 0;

	hc = getCloneData(args[1]);

	if (!hc) {
		for (permignore = firstPermClone; permignore; permignore = permignore->next) {
			if (!strcasecmp(permignore->host,args[1])) {
				tempflags = CLONE_PERMIGNORE;
				break;
			}
		}
	}
	else 
		tempflags = hc->flags;

	if (numargs < 3) {
		if (tempflags)
			uSend(OperServ, nick, GetReply(RPL_HASCLONEFLAGS, nick), args[1], (tempflags & CLONE_KILLFLAG) ? " Auto Remove" : "", (tempflags & CLONE_IGNOREFLAG) ? " Ignore All" : "", (tempflags & CLONE_PERMIGNORE) ? " Perm Ignore" : "");
		else
			uSend(OperServ, nick, GetReply(RPL_NOCLONEFLAGS, nick), args[1]);
	}
	else {
		if (!strcasecmp(args[2],"kill")) {
			if (hc) {
				hc->flags |= CLONE_KILLFLAG;
				tempflags = hc->flags;
				goodhost = 1;
			}
		}
		else if(!strcasecmp(args[2],"-kill")) {
			if (hc) {
				hc->flags &= ~CLONE_KILLFLAG;
				tempflags = hc->flags;
				goodhost = 1;
			}
		}
		else if(!strcasecmp(args[2],"ignore")) {
			if (hc) {
				hc->flags |= CLONE_IGNOREFLAG;
				tempflags = hc->flags;
				goodhost = 1;
			}
		}
		else if(!strcasecmp(args[2],"-ignore")) {
			if (hc) {
				hc->flags &= ~CLONE_IGNOREFLAG;
				tempflags = hc->flags;
				goodhost = 1;
			}
		}
		else if(!strcasecmp(args[2],"permignore")) {
			
                         if(!hc && tempflags) {
					uSend(OperServ, nick, GetReply(ERR_PERMIGNEXISTS, nick), args[1]);
                                        goodhost = 2;
                         }

                         if (hc) {
				if (hc->flags & CLONE_PERMIGNORE) {
					uSend(OperServ, nick, GetReply(ERR_PERMIGNEXISTS, nick), args[1]);
                                        goodhost = 2;
                                }
				else {
					hc->flags |= CLONE_PERMIGNORE;
					permignore = smalloc(sizeof(struct hostcloneliststruct));
					strcpy(permignore->host,hc->host);
					strcpy(permignore->setby, from);
					permignore->setwhen = CTime;
					permignore->next = firstPermClone;
					permignore->prev = NULL;
					if (permignore->next)
						permignore->next->prev = permignore;
					firstPermClone = permignore;	
					tempflags = hc->flags;
					goodhost = 1;
				}
			}
		}
		else if(!strcasecmp(args[2],"-permignore")) {
			if (hc) {
				hc->flags &= ~CLONE_PERMIGNORE;
				tempflags = hc->flags;
			}

			for (permignore = firstPermClone; permignore; permignore = permignore->next) {
				if (!strcasecmp(args[1],permignore->host)) {
					if (permignore->prev)
						permignore->prev->next = permignore->next;
					if (permignore->next)
						permignore->next->prev = permignore->prev;
					if (permignore == firstPermClone)
						firstPermClone = permignore->next;
					sfree(permignore);
					goodhost = 1;
					break;
				}
			}

			if (!goodhost) {
				uSend(OperServ, nick, GetReply(ERR_NOSUCHPERMIGN, nick), args[1]);
				return;
			}
		}

		else {
			uSend(OperServ, nick, GetReply(ERR_UNKNOWNFLAG, nick), args[1]);
			return;
		}

		if (goodhost == 1) {
			if (tempflags)
				sSend(":%s GLOBOPS :%s changed flags for %s (now%s%s%s).", OperServ, from, args[1], (tempflags & CLONE_KILLFLAG) ? " Auto Remove" : "", (tempflags & CLONE_IGNOREFLAG) ? " Ignore All" : "", (tempflags & CLONE_PERMIGNORE) ? " Perm Ignore" : "");
			else
				sSend(":%s GLOBOPS :%s changed flags for %s (now NONE).", OperServ, from, args[1]);
		}
		else if (goodhost != 2) 
			uSend(OperServ, nick, GetReply(ERR_UNKNOWHOST, nick), args[1]);
	}
}

void OmkillCmd(UserList *nick, char **args, int numargs) {
	char *from=nick->nick;
	char *s1;
	int i=0;
	char user[USERLEN];
	char host[HOSTLEN];
	int goodmask = 0;

	if (index(args[1],'@'))
		sscanf(args[1], "%[^@]@%s", user, host);
	else {
		strcpy(host,args[1]);
		strcpy(user,"*");
	}

	if (!index(user,'!') || !index(host,'!'))
	{
		s1 = host;

		while(*s1) {
			if ((*s1 != '?') && (*s1 !='*') && (*s1 != '.'))
				goodmask = 1;
			s1++;
		}
	}

	if (!goodmask) {
		uSend(OperServ, nick, GetReply(ERR_BADMASK, nick), args[1]);
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, args[0]);
		return;
	}

	i = killMassClones(user,host);

	if (i)
	{
		sSend(":%s GLOBOPS :%s masskilled %s@%s (%i matches)", OperServ, from, user, host, i);
		logDump(operlog,"%s!%s@%s masskilled %s@%s (%i matches)", from, nick->user, nick->host, user, host, i);
	} else
		uSend(OperServ, nick, GetReply(ERR_NOMATCH, nick), user, host);
}

void OmapCmd(UserList *nick, char **args, int numargs) {

   	aServer *srvptr;
    	aServer *srvptr2;

//	printf("%s", 3408932); /* <- manual core if uncommented, useful for core-debugging -GZ */

	if (ServerList == NULL)
        	return;

	uSend(OperServ, nick, GetReply(RPL_MAPHEAD, nick));
	for (srvptr = ServerList; srvptr; srvptr = srvptr2) {
        	srvptr2 = srvptr->next;
		uSend(OperServ, nick, GetReply(RPL_MAPDATA, nick), srvptr->name, srvptr->version);
	}
	uSend(OperServ, nick, GetReply(RPL_MAPMYDATA, nick), sversion, RELEASE);
	uSend(OperServ, nick, GetReply(RPL_MAPEND, nick));
        return;
}

void OjinxCmd(UserList *nick, char **args, int numargs) {

        char *from = nick->nick;
	UserList *target;
	aJinx	*jinxptr, *jinxptr2;
	int i;
	char reason[1000];

	bzero(reason,1000);

	if (!strcasecmp(args[1], "add")) {

		if (numargs < 4) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JINX");
			return;
		}

	        if (!(target = getNickData(args[2]))) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHNICK, nick), args[2]);
                	return;
		}

		if(CheckJinx(target->nick)) {
			uSend(OperServ, nick, GetReply(ERR_JINXEXIST, nick), "JINX", target->nick);
			return;
		}

               if(isRoot(target)) {
	       		uSend(OperServ, nick, GetReply(ERR_CANTJINXSRA, nick));
          	        sSend(":%s GLOBOPS :\002Warning:\002 %s attempted to JINX \002%s\002.", OperServ, from, args[2]);
                   	return;
               }

	      	sprintf(reason, args[3]);
		for(i=4;i<numargs;i++) {
			strcat(reason, " ");
			strcat(reason, args[i]);
		}

		AddJinx(target->nick, from, reason);
		
		sSend(":%s GLOBOPS :%s added a JINX for \002%s\002 (%s)", OperServ, from, target->nick, reason);
		sSend(":%s JINX %s", myname, target->nick);
		uSend(OperServ, nick, GetReply(RPL_JINXADDED, nick), "JINX", target->nick);
           	logDump(operlog,"%s!%s@%s Added a JINX for %s", from, nick->user, nick->host, target->nick);

		return;
	}

	else if (!strcasecmp(args[1], "del")) {

		if (numargs < 3) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "JINX");
			return;
		}
		
		if(!CheckJinx(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHJINX, nick), "JINX", args[2]);
			return;
		}

		DeleteJinx(args[2]);
		sSend(":%s GLOBOPS :%s removed the JINX for \002%s\002.", OperServ, from, args[2]);
		uSend(OperServ, nick, GetReply(RPL_JINXREMOVED, nick), "JINX", args[2]);
		sSend(":%s UNJINX %s", myname, args[2]);
           	logDump(operlog,"%s!%s@%s removed a JINX for %s", from, nick->user, nick->host, args[2]);

		return;
	}

	else if (!strcasecmp(args[1], "list")) {

		if (JinxList == NULL) {
			uSend(OperServ, nick, GetReply(RPL_JINXLISTEMPTY, nick), "JINXES");
        		return;
		}

		uSend(OperServ, nick, GetReply(RPL_JINXLISTHEADER, nick), "JINXES");
		for (jinxptr = JinxList; jinxptr; jinxptr = jinxptr2) {
	        	jinxptr2 = jinxptr->next;
			uSend(OperServ, nick, GetReply(RPL_JINXLIST1, nick), jinxptr->nick, jinxptr->by, jinxptr->reason, ctime(&(jinxptr->time)));
    		}
		uSend(OperServ, nick, GetReply(RPL_JINXLISTFOOTER, nick), "JINXES");

	}
		
	return;
}

void OpermjinxCmd(UserList *nick, char **args, int numargs) {

	int i;
	char *from = nick->nick;
	UserList *target;
	aPJinx	*pjinxptr, *pjinxptr2;
	char reason[1000];

	bzero(reason,1000);

	if (!strcasecmp(args[1], "add")) {

		if (numargs < 4) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "PERMJINX");
			return;
		}

	        if ((target = getNickData(args[2]))) {
	               if(target->reg && target->reg->flags & NROOT) {
	       			uSend(OperServ, nick, GetReply(ERR_CANTJINXSRA, nick));
                	   	sSend(":%s GLOBOPS :\002Warning:\002 %s attempted to PERMJINX \002%s\002.", OperServ, from, args[2]);
	                   return;
        	       }
		}

		if(CheckPJinx(args[2])) {
	       		uSend(OperServ, nick, GetReply(ERR_JINXEXIST, nick), "PERMJINX", args[2]);
			return;
		}

		sprintf(reason, args[3]);
		for(i=4;i<numargs;i++) {
			strcat(reason, " ");
			strcat(reason, args[i]);
		}

		AddPJinx(args[2], from, reason);
		
		sSend(":%s GLOBOPS :%s added a PERMJINX for \002%s\002 (%s)", OperServ, from, args[2], reason);
		uSend(OperServ, nick, GetReply(RPL_JINXADDED, nick), "PERMJINX", args[2]);
		sSend(":%s JINX %s", myname, args[2]);

		AddJinx(args[2], from, reason);

           	logDump(operlog,"%s!%s@%s Added a PERMJINX for %s", from, nick->user, nick->host, args[2]);

		return;
	}

	else if (!strcasecmp(args[1], "del")) {

		if (numargs < 3) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "PERMJINX");
			return;
		}
		
		if(!CheckPJinx(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHJINX, nick), "PERMJINX", args[2]);
			return;
		}

		DeletePJinx(args[2]);
		sSend(":%s GLOBOPS :%s removed the PERMJINX for \002%s.\002", OperServ, from, args[2]);
		uSend(OperServ, nick, GetReply(RPL_JINXREMOVED, nick), "PERMJINX", args[2]);
		sSend(":%s UNJINX %s", myname, args[2]);

           	logDump(operlog,"%s!%s@%s removed a PERMJINX for %s", from, nick->user, nick->host, args[2]);

		return;
	}

	else if (!strcasecmp(args[1], "list")) {

		if (PJinxList == NULL) {
			uSend(OperServ, nick, GetReply(RPL_JINXLISTEMPTY, nick), "PERMJINXES");
        		return;
		}

                uSend(OperServ, nick, GetReply(RPL_JINXLISTHEADER, nick), "PERMJINXES");
		for (pjinxptr = PJinxList; pjinxptr; pjinxptr = pjinxptr2) {
	        	pjinxptr2 = pjinxptr->next;
			uSend(OperServ, nick, GetReply(RPL_JINXLIST1, nick), pjinxptr->nick, pjinxptr->by, pjinxptr->reason, ctime(&(pjinxptr->time)));
    		}
                uSend(OperServ, nick, GetReply(RPL_JINXLISTFOOTER, nick), "PERMJINXES");
	} else {
		uSend(OperServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[1]);
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "PERMJINX");
	}
		
	return;
}

void OhostjinxCmd(UserList *nick, char **args, int numargs) {

	int i;
	char *from = nick->nick;
	aHJinx	*hjinxptr, *hjinxptr2;
	char reason[1000];

	bzero(reason,1000);

	if (!strcasecmp(args[1], "add")) {

		if (numargs < 4) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "HOSTJINX");
			return;
		}
		if((!strchr(args[2], '@'))) {
			uSend(OperServ, nick, GetReply(ERR_NEEDUSERHOST, nick));
			return;
		}

		if(CheckHJinx(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_JINXEXIST, nick), "HOSTJINX", args[2]);
			return;
		}

		if(!HJCheck(args[2]) && !isRoot(nick)) {
			sSend(":%s GLOBOPS :%s attempted to HOSTJINX %s - Thank God for this check!", OperServ, from, args[2]);
			return;
		}

	      	sprintf(reason, args[3]);
		for(i=4;i<numargs;i++) {
			strcat(reason, " ");
			strcat(reason, args[i]);
		}

		AddHJinx(args[2], from, reason);
		
		sSend(":%s GLOBOPS :%s added a HOSTJINX for \002%s\002 (%s)", OperServ, from, args[2], reason);
		uSend(OperServ, nick, GetReply(RPL_JINXADDED, nick), "HOSTJINX", args[2]);
		sSend(":%s JINX %s", myname, args[2]);

           	logDump(operlog,"%s!%s@%s Added a HOSTJINX for %s", from, nick->user, nick->host, args[2]);

		return;
	}

	else if (!strcasecmp(args[1], "del")) {

		if (numargs < 3) {
			uSend(OperServ, nick, GetReply(ERR_BADPARAMS, nick));
			uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "HOSTJINX");
			return;
		}
		
		if(!CheckHJinx(args[2])) {
			uSend(OperServ, nick, GetReply(ERR_NOSUCHJINX, nick), "HOSTJINX", args[2]);
			return;
		}

		DeleteHJinx(args[2]);
		sSend(":%s GLOBOPS :%s removed the HOSTJINX for %s.", OperServ, from, args[2]);
		uSend(OperServ, nick, GetReply(RPL_JINXREMOVED, nick), "HOSTJINX", args[2]);
		sSend(":%s UNJINX %s", myname, args[2]);

           	logDump(operlog,"%s!%s@%s removed a HOSTJINX for %s", from, nick->user, nick->host, args[2]);

		return;
	}

	else if (!strcasecmp(args[1], "list")) {

		if (HJinxList == NULL) {
			uSend(OperServ, nick, GetReply(RPL_JINXLISTEMPTY, nick), "JINXES");
        		return;
		}

                uSend(OperServ, nick, GetReply(RPL_JINXLISTHEADER, nick), "HOSTJINXES");
		for (hjinxptr = HJinxList; hjinxptr; hjinxptr = hjinxptr2) {
	        	hjinxptr2 = hjinxptr->next;
			uSend(OperServ, nick, GetReply(RPL_JINXLIST1, nick), hjinxptr->host, hjinxptr->by, hjinxptr->reason, ctime(&(hjinxptr->time)));
    		}
                uSend(OperServ, nick, GetReply(RPL_JINXLISTFOOTER, nick), "PERMJINXES");

	} else {
		uSend(OperServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[1]);	
		uSend(OperServ, nick, GetReply(RPL_MOREHELPCMD, nick), OperServ, "HOSTJINX");
	}
		
	return;
}

int HJCheck(char *mask)
{
   	char *temp1;
	int foundflag = 0;

   	for (temp1 = mask; *temp1; temp1++)
   	{
      		if ((*temp1 != '.') && (*temp1 != '*') && (*temp1 != '?'))
        	foundflag++ ;
   	}

   	if (foundflag > 3) return 1;
	 else return 0;

}
		
void OenforceCmd(UserList *tmp, char **args, int numargs) {

	char *from = tmp->nick;
	RegNickList *reg;
	int i;

    	anEmail *emailptr;
    	anEmail *emailptr2;

	i=0;

    	if (EmailList == NULL)
        	return;

	sSend(":%s GLOBOPS :Enforcing EKILLS. (Requested by %s)", OperServ, from);

    	for (emailptr = EmailList; emailptr; emailptr = emailptr2) {
        	emailptr2 = emailptr->next;
		
		for (reg=firstRegNick; reg; reg=reg->next) {
			if (!match(reg->email,emailptr->email)) {
				i++;
		                logDump(nicklog, "%s DELETE %s (EKILL ENFORCE for %s)", from, reg->nick, reg->email);
		                delRegNick(reg);
			}
		}
	}

	logDump(nicklog, "EKILL ENFORCE requested by %s", from);
	sSend(":%s GLOBOPS :Deleted %i nicks from database. (EKILL ENFORCE by %s)", NickServ, i, from);
}
