/*
 * Copyright (c) 1999-2000 GZ
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "services.h"

#ifdef TELNET
tn_sock telnets[5];
int freetn = 0;

void breakTelnetData(tn_sock *, char *);
void parseTelnetData(tn_sock *);

void initTelnet(void) {
	int i;
	for(i=0;i<5;i++)
	  memset(&telnets[i], 0, sizeof(telnets[i]));
}

void checkTelnet(void) {
	fd_set blah;
	struct timeval tv;
	int i;
	char readin[1024];

	if(-1<freetn<5) { /* we will accept new conns */
		tv.tv_sec = 0;
		tv.tv_usec = 500; /* wait 500 msec for info */
		FD_ZERO(&blah);
		FD_SET(telnetsock, &blah);
		if(select(telnetsock+1, &blah, NULL, NULL, &tv)) {
			/* someone rang? ;) */
			telnets[freetn].lensa = sizeof(telnets[freetn].sa);
			if((telnets[freetn].socket = accept(telnetsock, (struct sockaddr *)&telnets[freetn].sa, &telnets[freetn].lensa))>0) {
				/* now we've connected, fill in the blanks... */
				addr2ascii(AF_INET, &telnets[freetn].sa.sin_addr, sizeof(telnets[freetn].sa.sin_addr), telnets[freetn].address);
				bzero(telnets[freetn].readbuf, 1024);
				telnets[freetn].connect = CTime;
				telnets[freetn].lastrcv = CTime;
				sSend(":%s GLOBOPS :Telnet connect from %s", OperServ, telnets[freetn].address);
				for(i=0;i<5;i++) {
					if(!telnets[i].socket)
					  break;
					else if(i==4 && telnets[i].socket)
					  freetn = -1;
				}
				freetn = i;
			}
		}
	}
	
	for(i=0;i<5;i++) {
		if(telnets[i].socket) {
			bzero(readin, 1024);
			switch(read(telnets[i].socket, readin, 1024)) {
			 case -1:
				break; /* no data.. */
			 case 0:
				close(telnets[i].socket);
				memset(&telnets[i], 0, sizeof(tn_sock));
				break; /* Closed the socket */
			 default:
				telnets[i].lastrcv = CTime;
				breakTelnetData(&telnets[i], readin);
				break;
			}
		}
	}
}

void breakTelnetData(tn_sock *from, char *data) {
	int i;
	if(from->readbuf[0])
	  i=strlen(from->readbuf);
	else
	  i=0;
	
	while(*data) {
		if(*data == '\r' || *data == '\n') {
			from->readbuf[i] = 0;
			parseTelnetData(from);
			while(*data == '\r' || *data == '\n')
			  data++;
			bzero(from->readbuf, 1024);
			i=0;
		}
		else {
			from->readbuf[i] = *data;
			data++;
			i++;
		}
	}
}

void parseTelnetData(tn_sock *from) {
	char realargs[256][256];
	char *args[256];
	int i, numargs;
	RegNickList *blah;
	
	for(i=0;i<256;i++)
	  args[i] = realargs[i];
	
	numargs = breakString(255, from->readbuf, args, " ");
	
	if(!strcmp(args[0], "login")) {
		blah=getRegNickData(args[1]);
		if(blah && (!strcmp(blah->password, args[2]))) {
			if(blah->flags & NROOT) {
				from->user = blah;
				from->loggedin = 1;
				sSend(":%s GLOBOPS :Telnet user login: %s@%s", OperServ, args[1], from->address);
				sockSend(from->socket, "Welcome %s", args[1]);
				sockSend(from->socket, "Logged in at %s", ctime(&CTime));
			}
			else {
				sSend(":%s GLOBOPS :Nonroot telnet user login: %s@%s", OperServ, args[1], from->address);
				sockSend(from->socket, "Sorry, you cannot use this.");
				close(from->socket);
				memset(from, 0, sizeof(from));
			}
		}
		else {
			sockSend(from->socket, "Incorrect login");
			sSend(":%s GLOBOPS :Incorrect login for %s from %s", OperServ, args[1], from->address);
		}
	}
	if(!from->loggedin)
	  sockSend(from->socket, "Login first.");
	else {
		if(!strcmp(args[0], "inject")) {
			parseLine(&from->readbuf[7]);
			sockSend(from->socket, "IJ: %s", &from->readbuf[7]);
		}
		else if(!strcmp(args[0], "raw")) {
			sSend(&from->readbuf[4]);
			sockSend(from->socket, "R : %s", &from->readbuf[4]);
		}
	}
}
	
#endif
